/*
  /-------------------------------------------------\
  |                                                 |
  |                                                 |
  |     /=====\     /=====\     |\          /|      |
  |    /       \   /       \    | \        / |      |
  |    |           |            |  \      /  |      |
  |    |           |            |   \    /   |      |
  |    \       /   \       /    |    \  /    |      |
  |     \=====/     \=====/     |     \/     |      |
  |                                                 |
  |                                                 |
  \-------------------------------------------------/

  ===================================================

  Name:      CCM (obtainExcitations.C)
  
  Copyright: Damian J. J. Farnell

  Updated:   Joerg Schulenburg 2008

  Purpose:   Determine CCM LSUBm excited-state equations.

  ===================================================

  ---------------------------------------------------

  MODULE NAME: OBTAIN.C 

  PURPOSE:     `Obtains' the CCM excited state equations for LSUBm

  ---------------------------------------------------

  Code amended for exicitations on Mon Jun 16 13:16:10 BST 2008, DJJF.
  Checked on Thu Jan 29 14:55:55 GMT 2009 DJJF
  
  ---------------------------------------------------

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// #include <iostream.h>
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#include "Constants.h"
#include "Externals.h"
#include "FunctionPrototypes.h"

#if (PARALLEL_FLAG>0)
#include <mpi.h>
#endif

int subsetExcitedClusters(int (*config)[5], int length); // 1 if ok, 0 if not
int subsetClusters(int (*config)[5], int length); // 1 if ok, 0 if not
//void findExcitedFundamentalSpinConfigurations(void);
// int findExcitedDescriptorIndex(int *descriptor); 
int findExcitedDescriptorOfCluster(int order,
                                   int (*configuration)[5],
                                   int *descriptor);
void collect_term(int indices[6],double dummy,int id);
int findClusterIndex(int order, int (*configuration)[5]); // obtain.C

// Jun08 Joerg new: find DescriptorIndex from Cluster (one call)
// replacing three lines by one (ToDo: call subset too?)
//   return <= 0 (not found), >=1 (found) 
int findExcitedClusterIndex(int order, int (*configuration)[5]) {
  static int descriptor[MAX_SPIN_LIST];
  findExcitedDescriptorOfCluster(order, configuration, descriptor);
  return findExcitedDescriptorIndex(descriptor);
}

int radiusExc=0; // only used within obtainExcitations.C

// defined in collect.C
extern unsigned long
 nterms_in1, nterms_in2,
 nterms_out1, nterms_out2;

// see at bottom for obtain(void)

/* test if site s1 and site s2 are neighbours k -> m */
static
int test_k_to_m_neighbour(struct hamiltonian_defn hh[],
                          int term, int s1[5], int s2[5]) {
  int are_they_neighbours=0;
  // struct hamiltonian_defn *hh=&Hamiltonian[bonds][map81[term]];
  int index=s1[4];
  int j;
  // neighbourhood of s2 to s1 (can be directional!)
  for (j=1;j<=hh[map81[term]].k_to_m_number[index];j++) {
	/* k ---> m                      uc   nn xyzc */
	if (s2[1] == s1[1] + hh[map81[term]].k_to_m[index][j][0] /* x */
	 && s2[2] == s1[2] + hh[map81[term]].k_to_m[index][j][1] /* y */
	 && s2[3] == s1[3] + hh[map81[term]].k_to_m[index][j][2] /* z */
	 && s2[4] ==         hh[map81[term]].k_to_m[index][j][3] /* uc */ ) {
             are_they_neighbours=j; break; }
  }
  return are_they_neighbours;
}	

/* test if site s1 and site s2 are neighbours k -> m */
static
int test_m_to_k_neighbour(struct hamiltonian_defn hh[],
                          int term, int s1[5], int s2[5]) {
  int are_they_neighbours=0;
  // struct hamiltonian_defn *hh=&Hamiltonian[bonds][map81[term]];
  int index=s1[4];
  int j;
  // neighbourhood of s2 to s1 (can be directional!)
  for (j=1;j<=hh[map81[term]].m_to_k_number[index];j++) {
	/* k ---> m                      uc   nn xyzc */
	if (s2[1] == s1[1] + hh[map81[term]].m_to_k[index][j][0] /* x */
	 && s2[2] == s1[2] + hh[map81[term]].m_to_k[index][j][1] /* y */
	 && s2[3] == s1[3] + hh[map81[term]].m_to_k[index][j][2] /* z */
	 && s2[4] ==         hh[map81[term]].m_to_k[index][j][3] /* uc */ ) {
             are_they_neighbours=j; break; }
  }
  return are_they_neighbours;
}	

// s^z s^z : all s 


double term_exc_1(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{
  
  /* 

     TERM 1: 
     
     P(i) . S(+)(i) . S(z)(j)

  */
  
  int i,j,k;
  // double multiplier;
  int (*configuration)[5];
  int neighbour;
  int index;
  int new_index;
  double sum;
  sum=0.0000;

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  for(k=1;k<=length;k++) {
    
    index=configuration[k][4];      

    for(neighbour=1;
        neighbour<=hh[map81[term]].k_to_m_number[index];
        neighbour++) {
      
      /* k ---> m */
      
      new_index=hh[map81[term]].k_to_m[index][neighbour][3];
                  
      if (Control.file) {
	int indices[7]={0};
	double dummy;
	indices[0]=config_index;
        indices[1]=term; // Jun08: origin was 1..81, now using mapped term
	indices[2]=config_index; // this index ALWAYS contains the excited-state coefficient
        indices[3]=0;
	indices[4]=0;
        indices[5]=0;
	dummy=(-0.5*UnitCell.s2[new_index])*
	  hh[map81[term]].coefficient[index][neighbour];
        collect_term(indices,dummy,3);
      }
    }
  }
  free(configuration);
  return sum;
}

double term_exc_2(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 2: 

     P(i) . F(j) . S(+)(i) . S(+)(j).
     
  */

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int i, j, k, l, m;
  int k1;
  int k2;
  //  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  // double multiplier;
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  double sum;
  int (*configuration)[5];
  int nn[5];

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  sum=0.0000;

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) conf1[1][m]=configuration[k1][m]; //p
	for(m=1;m<5;m++) conf2[1][m]=configuration[k2][m]; //f
	
	// loop over nearest neighbours
	
		are_they_neighbours=0;
		index=conf1[1][4];      
	  
		for(j=1;j<=hh[map81[term]].k_to_m_number[index];j++) {
	    
			/* k ---> m */
	    
			nn[1]=conf1[1][1]+hh[map81[term]].k_to_m[index][j][0];
			nn[2]=conf1[1][2]+hh[map81[term]].k_to_m[index][j][1];
			nn[3]=conf1[1][3]+hh[map81[term]].k_to_m[index][j][2];
			nn[4]=hh[map81[term]].k_to_m[index][j][3];
	    
			if(conf2[1][1] == nn[1] &&
				conf2[1][2] == nn[2] &&
				conf2[1][3] == nn[3] &&
				conf2[1][4] == nn[4])
			are_they_neighbours=j;
      
			if (are_they_neighbours) break;
		}
	
	// conf1[1] and conf2[1] need to be nearest-neighbours 
	
	if (are_they_neighbours) {
	  
	  counter=0;
	  for(l=1;l<=length;l++) {
	    if (l != k1 && l != k2 ) {
	      counter++;
	      tempArray[counter]=l;
	    }
	  }	  
	  
	  for(l=0;l<binary_max[counter];l++) {
	    
	    binary(l,counter,spin);
	    conf1_len=1;
	    conf2_len=1;
	    
	    for (j=1;j<=counter;j++) {
	      if (spin[j]) {
		conf1_len++;
		temp=tempArray[j];
		for (m=1;m<5;m++)
		  conf1[conf1_len][m]=configuration[temp][m];
	      } else {
		conf2_len++;
		temp=tempArray[j];
		for (m=1;m<5;m++) 
		  conf2[conf2_len][m]=configuration[temp][m];
	      }
	    }	      
	    
	    descriptorIndex1=findExcitedClusterIndex(conf1_len, conf1);
	    if (descriptorIndex1 <= 0) continue;

	    descriptorIndex2=findClusterIndex(conf2_len,conf2);
            if (descriptorIndex2 <= 0) continue;
	    
	    // printf("%d : %d %d\n",config_index,descriptorIndex1,descriptorIndex2);
		
	    if (Control.file) {
	      if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)) {
		int indices[7]={0};
		double dummy;
		indices[0]=config_index;
		indices[1]=term; // Jun08: origin was 1..81, now using mapped term
		indices[2]=descriptorIndex1; // this index ALWAYS contains the excited-state coefficient
		indices[3]=descriptorIndex2;
		indices[4]=0;
		indices[5]=0;
  dummy=hh[map81[term]].coefficient[index][are_they_neighbours];
                collect_term(indices,dummy,3);
	      }
	    }
	    
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}

double term_exc_3(int config_index, int length, int (*string)[5],
             struct hamiltonian_defn hh[], int term)
{
   /* 
     TERM 3: 
     
     P(j) . S(+)(j) . S(z)(i)
     
  */
  
  int i,j,k;
//  double multiplier;
  int (*configuration)[5];
  int neighbour;
  int index;
  int new_index;
  double sum;
  sum=0.0000;
  
  if (!hh[map81[term]].evaluate) return 0; /* skip that term */

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  
  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];
  
  for(k=1;k<=length;k++) {
    
    index = configuration[k][4]; // for every site ...
    
    for(neighbour=1;
        neighbour<=hh[map81[term]].m_to_k_number[index];
	neighbour++) {
      
      /* m ---> k */
      
      new_index = hh[map81[term]].m_to_k[index][neighbour][3];
      
      if (Control.file) {
	int indices[7]={0};
	double dummy;
	indices[0]=config_index;
        indices[1]=term; // Jun08: origin was 1..81, now using mapped term
	indices[2]=config_index; // this index ALWAYS contains the excited-state coefficient
        indices[3]=0;
	indices[4]=0;
        indices[5]=0;
	dummy=(-0.5*UnitCell.s2[new_index])*
	  hh[map81[term]].coefficient[index][neighbour];
        collect_term(indices,dummy,3);
      }
      
    }
  }
  free(configuration);
  return sum;
}

double term_exc_4(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 4: 

     P(j) . F(i) . S(+)(i) . S(+)(j).
     
  */

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int i, j, k, l, m;
  int k1;
  int k2;
  //  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  // double multiplier;
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  double sum;
  int (*configuration)[5];
  int nn[5];

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  sum=0.0000;

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) conf1[1][m]=configuration[k1][m]; 
	for(m=1;m<5;m++) conf2[1][m]=configuration[k2][m];
	
	// loop over nearest neighbours
	
		are_they_neighbours=0;
		index=conf1[1][4];      
	  
		for(j=1;j<=hh[map81[term]].m_to_k_number[index];j++) {
	    
			/* m ---> k */
	    
			nn[1]=conf1[1][1]+hh[map81[term]].m_to_k[index][j][0];
			nn[2]=conf1[1][2]+hh[map81[term]].m_to_k[index][j][1];
			nn[3]=conf1[1][3]+hh[map81[term]].m_to_k[index][j][2];
			nn[4]=hh[map81[term]].m_to_k[index][j][3];
	    
			if(conf2[1][1] == nn[1] &&
				conf2[1][2] == nn[2] &&
				conf2[1][3] == nn[3] &&
				conf2[1][4] == nn[4])
			are_they_neighbours=j;
      
			if (are_they_neighbours) break;
		}

	// conf1[1] and conf2[1] need to be nearest-neighbours 
	
	if (are_they_neighbours) {
	  
	  counter=0;
	  for(l=1;l<=length;l++) {
	    if (l != k1 && l != k2 ) {
	      counter++;
	      tempArray[counter]=l;
	    }
	  }	  
	  
	  for(l=0;l<binary_max[counter];l++) {
	    
	    binary(l,counter,spin);
	    conf1_len=1;
	    conf2_len=1;
	    
	    for (j=1;j<=counter;j++) {
	      if (spin[j]) {
		conf1_len++;
		temp=tempArray[j];
		for (m=1;m<5;m++)
		  conf1[conf1_len][m]=configuration[temp][m];
	      } else {
		conf2_len++;
		temp=tempArray[j];
		for (m=1;m<5;m++) 
		  conf2[conf2_len][m]=configuration[temp][m];
	      }
	    }	      
	    
	    descriptorIndex1=findExcitedClusterIndex(conf1_len, conf1);
	    if (descriptorIndex1 <= 0) continue;

	    descriptorIndex2=findClusterIndex(conf2_len, conf2);
            if (descriptorIndex2 <= 0) continue;
	    
	    if (Control.file) {
	      if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)) {
		int indices[7]={0};
		double dummy;
		indices[0]=config_index;
		indices[1]=term; // Jun08: origin was 1..81, now using mapped term
		indices[2]=descriptorIndex1; // this index ALWAYS contains the excited-state coefficient
		indices[3]=descriptorIndex2;
		indices[4]=0;
		indices[5]=0;
  dummy=hh[map81[term]].coefficient[index][are_they_neighbours];
                collect_term(indices,dummy,3);
	      }
	    }
	    
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}

double term_exc_5(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 5: 

     Q(i,j) . S(+)(i) . S(+)(j).

  */

  int conf[4][6];
  int i, j, k, l;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  //  double multiplier;
  int (*configuration)[5];
  double sum;

  sum=0.0000;

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  for(k=1;k<3;k++) {
    for(l=1;l<5;l++) {
      conf[k][l]=0;
    }
  } 
  

  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      
      if (k1 != k2) {  // ToDo: better k1 < k2 and factor 2 ??? Joerg
	
	for(l=1;l<5;l++) { conf[1][l]=configuration[k1][l]; }
	for(l=1;l<5;l++) { conf[2][l]=configuration[k2][l]; }
	
	// loop over nearest neighbours
	
		are_they_neighbours=0;
		index=conf[1][4];      
	  
		for(j=1;j<=hh[map81[term]].k_to_m_number[index];j++) {
	    
			/* k ---> m */
	    
			nn[1]=conf[1][1]+hh[map81[term]].k_to_m[index][j][0];
			nn[2]=conf[1][2]+hh[map81[term]].k_to_m[index][j][1];
			nn[3]=conf[1][3]+hh[map81[term]].k_to_m[index][j][2];
			nn[4]=hh[map81[term]].k_to_m[index][j][3];
	    
			if(conf[2][1] == nn[1] &&
				conf[2][2] == nn[2] &&
				conf[2][3] == nn[3] &&
				conf[2][4] == nn[4])
			are_they_neighbours=j;
      
			if (are_they_neighbours) break;
		}
		
	// conf[1] and conf[2] need to be nearest-neighbours 
	
	if (are_they_neighbours) {
	      
	  if (Control.file) {
	    int indices[7]={0};
	    double dummy;
	    indices[0]=config_index;
            indices[1]=term; // Jun08: origin was 1..81, now using mapped term
	    indices[2]=config_index; // contains the excited-state coeff
            indices[3]=0;
	    indices[4]=0;
            indices[5]=0;
	    dummy=hh[map81[term]].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,3);
	  }
	  
	}
      }
    }
  }
  free(configuration);
  return sum;
}

// now do s^2 s^+

double term_exc_6(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 6: 
     
     P(i) . S(+)(i) . S(+)(m).
     
  */
  
  int i, j, l;
  int k1;
  int k2;
  // int nn[5];
  int are_they_neighbours;
  int index;
  int descriptorIndex;
  double multiplier;
  int (*configuration)[5];
  int (*conf)[5];
  int tempConfig[MAX_LSUBM+2];
  int counter;
  double sum;
  sum=0.0000;

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf=         (int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++) {
      configuration[i][j]=string[i][j];
      conf[i][j]=0;
    }
  
  multiplier=1.0;
    
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      
      if (k1 != k2) {  // ToDo: better k1 < k2 and factor 2 ??? Joerg
	
	for(l=1;l<5;l++) { conf[1][l]=configuration[k1][l]; }
	for(l=1;l<5;l++) { tempConfig[l]=configuration[k2][l]; }
	
	// loop over nearest neighbours
	
        index = conf[1][4];
        are_they_neighbours=
          test_k_to_m_neighbour(hh, term, conf[1], tempConfig);
	
	// conf[1][] and tempConfig[] need to be nearest-neighbours 
	
	// now load the rest into conf
	
	counter=1;
	for(i=1;i<=length;i++) {
	  if ((i != k2) && (i != k1)) {
	    counter++;
	    for(l=1;l<5;l++)
	      conf[counter][l]=configuration[i][l];
	  }
	}
	
	if (counter != (length-1)) {
	  printf("TERM 7: Error counter != (length-1)...\nExiting\n\n");
	  exit(1);
	}
	
	if (are_they_neighbours) {
	  
	  descriptorIndex=findExcitedClusterIndex(length-1,conf);
	  
	  // if (descriptorIndex > 0)
	  //  sum+=multiplier*CCMcoeffs[descriptorIndex]*
	  //    hh[map81[term]].coefficient[index][are_they_neighbours];
	  
	  if (Control.file) {
	    if (descriptorIndex > 0) {
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=term; // Jun08: origin was 1..81, now using mapped term
	      indices[2]=descriptorIndex; // P(i)
	      indices[3]=0;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=hh[map81[term]].coefficient[index][are_they_neighbours]; // factor of +1 from H.
              collect_term(indices,dummy,3);
	    }
	  }
	  
	}
      }
    }
  }
  free(configuration);
  free(conf);
  return sum;
}

// s^z s^-

double term_exc_7(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 7:
     
     -2 * P(i) . F(j) . S(+)(i) . S(z)(j)
     
  */

  int i,j,k,l,m;
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int (*configuration)[5];
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int numberFlips;
  int (*conf1)[5], (*conf2)[5];
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
  int temp_number_conf2_sites;
  double sum;
  sum=0.0000;

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 

  for(k=1;k<=length;k++) {
        
    for(m=1;m<5;m++)
      conf1[1][m]=configuration[k][m]; // P(i)

    counter=1;
    for(m=1;m<=length;m++) {
      if (m != k) {
	tempArray[counter]=m;
	counter++;
      }
    }
    
    counter--;
    
    for(l=0;l<binary_max[counter];l++) {
      
      binary(l,counter,spin);
      conf1_len=1;
      conf2_len=0;
      
      for (j=1;j<=counter;j++) { 
	if (spin[j]) {
	  // P(i)
	  conf1_len++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf1[conf1_len][m]=configuration[temp][m];
	} else {
	  // F(j)
	  conf2_len++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf2[conf2_len][m]=configuration[temp][m];
	}
      }
      
      index=conf1[1][4];
      
      for(neighbour=1;neighbour<=hh[map81[term]].k_to_m_number[index];
	  neighbour++) {
	
	// nearest neighbour is at x,y,z, cell put information
	// into configuration and exit.
	
	/* k ---> m */
	
	nn[1]=conf1[1][1]+hh[map81[term]].k_to_m[index][neighbour][0];
	nn[2]=conf1[1][2]+hh[map81[term]].k_to_m[index][neighbour][1];
	nn[3]=conf1[1][3]+hh[map81[term]].k_to_m[index][neighbour][2];
	nn[4]=hh[map81[term]].k_to_m[index][neighbour][3];
	
	numberFlips=1;
	
	for(m=1;m<=conf2_len;m++)
	  if (nn[1]==conf2[m][1] &&
	      nn[2]==conf2[m][2] &&
	      nn[3]==conf2[m][3] &&
	      nn[4]==conf2[m][4])
	    numberFlips++;
	
	maxNumberFlips=UnitCell.s2[nn[4]];
	
	// the number of spin flips that you can have at a given
	// lattice site is depedent on the quantum spin number
	// of the site in the UNIT CELL index `length'
	
	if (numberFlips <= maxNumberFlips) {
	  
	  temp_number_conf2_sites=conf2_len+1;
	  
	  conf2[temp_number_conf2_sites][1]=nn[1];
	  conf2[temp_number_conf2_sites][2]=nn[2];
	  conf2[temp_number_conf2_sites][3]=nn[3];
	  conf2[temp_number_conf2_sites][4]=nn[4];
	  
	  descriptorIndex1=findExcitedClusterIndex(conf1_len, conf1);
	  if (descriptorIndex1 <= 0) continue;
	  
          descriptorIndex2=findClusterIndex(temp_number_conf2_sites, conf2);
	  if (descriptorIndex2 <= 0) continue;

	  multiplier=UnitCell.s2[nn[4]]; // from H -2s
	  
	  if (Control.file) {
	    if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)){
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=term; // Jun08: origin was 1..81, now using mapped term
	      indices[2]=descriptorIndex1; // P(i)
	      indices[3]=descriptorIndex2; // F(j)
	      indices[4]=0;
	      indices[5]=0;
	      dummy=multiplier*hh[map81[term]].coefficient[index][neighbour];
            collect_term(indices,dummy,3);
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}  

double term_exc_8(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 8: 
     
     - P(i) . G(j)(j) . S(+)(i) . S(+)(j)
     
    */

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int maxNumberFlips1;
  int maxNumberFlips;
  int numberFlips1;
  double sum;
  sum=0.0000;

  int (*configuration)[5];

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {

	maxNumberFlips=UnitCell.s2[configuration[k1][4]];
	
	if (maxNumberFlips > 1) {
	  
	  for(m=1;m<5;m++) {
	    conf1[1][m]=configuration[k1][m]; // G(j,j) m--1st one
	    conf1[2][m]=configuration[k1][m]; // G(j,j) m--2nd one
	    conf2[1][m]=configuration[k2][m]; // P(i)   <- note different order than normal here!
	  }
	  
	  // loop over nearest neighbours
	  
	  are_they_neighbours=0;
	  index=conf1[1][4];      
	  
	  for(j=1;j<=hh[map81[term]].m_to_k_number[index];j++) {
	    
	    // nearest neighbour is at x,y,z, temp put information
	    // into configuration and exit.
	    
	    /* m ---> k */
	    
	    nn[1]=conf1[1][1]+hh[map81[term]].m_to_k[index][j][0];
	    nn[2]=conf1[1][2]+hh[map81[term]].m_to_k[index][j][1];
	    nn[3]=conf1[1][3]+hh[map81[term]].m_to_k[index][j][2];
	    nn[4]=hh[map81[term]].m_to_k[index][j][3];
	    
	    if(conf2[1][1] == nn[1] &&
	       conf2[1][2] == nn[2] &&
	       conf2[1][3] == nn[3] &&
	       conf2[1][4] == nn[4])
	      are_they_neighbours=j;
  
	    if (are_they_neighbours) break;    
	  }
	  
	  // conf1[1] and conf2[1] need to be nearest-neighbours 
	  
	  if (are_they_neighbours) {
	    
	    counter=1;
	    for(l=1;l<=length;l++) {
	      if (l != k1 && l != k2 ) {
		tempArray[counter]=l;
		counter++;
	      }
	    }
	    
	    counter--;
	    
	    for(l=0;l<binary_max[counter];l++) {
	      
	      binary(l,counter,spin);
	      conf1_len=2;
	      conf2_len=1;
	      
	      for (j=1;j<=counter;j++) { 
		if (spin[j]) {
		  // G
		  conf1_len++;
		  temp=tempArray[j];
		  for (m=1;m<5;m++) 
		    conf1[conf1_len][m]=configuration[temp][m];
		} else {
		  // P
		  conf2_len++;
		  temp=tempArray[j];
		  for (m=1;m<5;m++) 
		    conf2[conf2_len][m]=configuration[temp][m];
		}
	      }	      
	      
	      // make sure that we don't have too many `k' sites.
	      
	      numberFlips1=0;
	      
	      for(m=1;m<=conf1_len;m++) {
		if (configuration[k1][1]==conf1[m][1] &&
		    configuration[k1][2]==conf1[m][2] &&
		    configuration[k1][3]==conf1[m][3] &&
		    configuration[k1][4]==conf1[m][4])
		  numberFlips1++;
	      }
	      maxNumberFlips1=UnitCell.s2[configuration[k1][4]];
	      
	      if (numberFlips1 <= maxNumberFlips1) {
		
		descriptorIndex1=findClusterIndex(conf1_len, conf1);
		if (descriptorIndex1 <= 0) continue;
		
		descriptorIndex2=findExcitedClusterIndex(conf2_len, conf2);
		if (descriptorIndex2 <= 0) continue;

		multiplier=-1.0; // -1 factro fom H.
		
		if (Control.file) {
		  if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)) {
		    int indices[7]={0};
		    double dummy;
		    indices[0]=config_index;
		    indices[1]=term; // Jun08: origin was 1..81, now using mapped term
		    indices[2]=descriptorIndex2; // p(i) -- again the excited-state coeff is always first!
		    indices[3]=descriptorIndex1; // g(j,j)
		    indices[4]=0;
		    indices[5]=0;
		    dummy=multiplier*
		      hh[map81[term]].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,3);
		  }
		}

	      }
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}

double term_exc_9(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 9: 
     
     - P(i) . F(j) . F(j) . S(+)(i) . S(+)(j)
   
    */

  int spin[MAX_LSUBM+1];
  int counter;
  int mm,nnn,temp1,temp2,temp3;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int (*conf3)[5];
  int i, j, k, l, m;
  int k1;
  int k2;
  // int nn[5];
  int are_they_neighbours;
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  int descriptorIndex3;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int conf3_len;
  int tempArray[MAX_LSUBM+1];
  double sum;
  sum=0.0000;

  int (*configuration)[5];

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf3=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
      conf3[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) conf1[1][m]=configuration[k1][m]; // P(k)
	for(m=1;m<5;m++) conf2[1][m]=configuration[k2][m]; // F(m)
	for(m=1;m<5;m++) conf3[1][m]=configuration[k2][m]; // F(m)
	
	// loop over nearest neighbours
	
        index = conf1[1][4];
        are_they_neighbours=
          test_k_to_m_neighbour(hh, term, conf1[1], conf2[1]);
	
	// conf1[1] and conf2[1] need to be nearest-neighbours 
	
	if (are_they_neighbours) {
	  
	  counter=1;
	  for(l=1;l<=length;l++) {
	    if (l != k1 && l != k2 ) {
	      tempArray[counter]=l;
	      counter++;
	    }
	  }
	  
	  counter--;
	  
	  if (counter > 30) {
	    printf("TERM 9: counter > 30...\nExiting!!\n\n");
	    exit(1);
	  }
          temp = counter; // split into summands not greater than 10 each
	  temp3 = ((temp>10) ? 10 : temp); temp -= temp3;  // 0..10
	  temp2 = ((temp>10) ? 10 : temp); temp -= temp2;  // 0..10
	  temp1 = ((temp>10) ? 10 : temp); temp -= temp1;  // 0..10
	  
          // trimary_max[i] = 3^i  , try all 3^counter codes, counter<SUBn
	  for(l=0;l<trimary_max[temp1];l++) {
	    for(mm=0;mm<trimary_max[temp2];mm++) {
	      for(nnn=0;nnn<trimary_max[temp3];nnn++) {
		
                // decode counter to spin[i] trimary code
		trimary(l,mm,nnn,counter,spin);
                // now we have next trimary code 000 -> 001 -> 002 -> 010 ...

		conf1_len=1;
		conf2_len=1;
		conf3_len=1;
		
		for (j=1;j<=counter;j++) { 
		  if (spin[j]==0) {         // P(k)
		    conf1_len++;
		    temp=tempArray[j];
		    for (m=1;m<5;m++) 
		      conf1[conf1_len][m]=configuration[temp][m];
		  } else if (spin[j]==1) {  // F(m)
		    conf2_len++;
		    temp=tempArray[j];
		    for (m=1;m<5;m++) 
		      conf2[conf2_len][m]=configuration[temp][m];
		  } else {                   // F(m)
		    conf3_len++;
		    temp=tempArray[j];
		    for (m=1;m<5;m++) 
		      conf3[conf3_len][m]=configuration[temp][m];
		  }
		}
		
		descriptorIndex1=findExcitedClusterIndex(conf1_len, conf1);
		if (descriptorIndex1 <= 0) continue;
		
		descriptorIndex2=findClusterIndex(conf2_len, conf2);
		if (descriptorIndex2 <= 0) continue;
		
		descriptorIndex3=findClusterIndex(conf3_len, conf3);	      
		if (descriptorIndex3 <= 0) continue;
						
		multiplier=-1.0; // factor from -1 in H.
				
		if (Control.file) {
		  if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0) 
		      && (descriptorIndex3 > 0)){
		    int indices[7]={0};
		    double dummy;
		    indices[0]=config_index;
		    indices[1]=term; // Jun08: origin was 1..81, now using mapped term
		    indices[2]=descriptorIndex1; // p(i)
		    indices[3]=descriptorIndex2;
		    indices[4]=descriptorIndex3;
		    indices[5]=0;
		    dummy=multiplier*hh[map81[term]].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,3);
		  }
		}
		
	      }
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(conf3);
  free(configuration);
  return sum;
}

double term_exc_10(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 10: 
     
     -2 * P(j) . S(z)(i) . S(z)(j)

     one of the most time-consuming terms because you've got to sweep
     through the index `k'

     */

  int i, j, l;
  int descriptorIndex;
//  double multiplier;
  double temp_multiplier;
  double sum;
  sum=0.0000;

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int minX, minY, minZ;
  int maxX, maxY, maxZ;

  int x,y,z,cell;
  int index;

  int numberFlips, maxNumberFlips;

  int (*configuration)[5];
  int nn[5];

  if (length+1+1 > Params.n) return 0;

  configuration=(int (*)[5])malloc((MAX_LSUBM+4)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  int maxPointX;
  int maxPointY;
  int maxPointZ;

  maxPointX=0;
  maxPointY=0;
  maxPointZ=0;
  for(j=1;j<=length;j++) {
    if (maxPointX < abs(string[j][1])) maxPointX=abs(string[j][1]);
    if (maxPointY < abs(string[j][2])) maxPointY=abs(string[j][2]);
    if (maxPointZ < abs(string[j][3])) maxPointZ=abs(string[j][3]);
  }

  if (Params.dim == 1) {
    minX=radiusExc;           minY=0;                minZ=0;
    maxX=radiusExc+maxPointX; maxY=0;                maxZ=0;
  } else if (Params.dim == 2) {
    minX=radiusExc;           minY=radiusExc;           minZ=0;
    maxX=radiusExc+maxPointX; maxY=radiusExc+maxPointY; maxZ=0;
  } else if (Params.dim == 3) {
    minX=radiusExc;           minY=radiusExc;           minZ=radiusExc;
    maxX=radiusExc+maxPointX; maxY=radiusExc+maxPointY; maxZ=radiusExc+maxPointZ;
  } else {
    printf("Really weird error. Dimension (%d)"
     " not recognised...\nExiting||\n\n",Params.dim);
    exit(1);
  }


  for(x=-minX;x<=maxX;x++) { // ToDo: better (faster) x = maxX-m ... m ???
    for(y=-minY;y<=maxY;y++) {
      for(z=-minZ;z<=maxZ;z++) {
    for(cell=1;cell<=UnitCell.number;cell++) {
		
	      configuration[length+1][1]=x;
	      configuration[length+1][2]=y;
	      configuration[length+1][3]=z;
	      configuration[length+1][4]=cell;
	      
	      numberFlips=1;
	      
	      for(l=1;l<=length;l++)
		if (configuration[length+1][1]==configuration[l][1] &&
		    configuration[length+1][2]==configuration[l][2] &&
		    configuration[length+1][3]==configuration[l][3] &&
		    configuration[length+1][4]==configuration[l][4])
		  numberFlips++;
	      
	      maxNumberFlips=UnitCell.s2[configuration[length+1][4]];

	      index=cell;

	      if (numberFlips <= maxNumberFlips) {

				for(j=1;j<=hh[map81[term]].m_to_k_number[index];j++) {
	      
					nn[1]=configuration[length+1][1]
						+hh[map81[term]].m_to_k[index][j][0];
					nn[2]=configuration[length+1][2]
						+hh[map81[term]].m_to_k[index][j][1];
					nn[3]=configuration[length+1][3]
						+hh[map81[term]].m_to_k[index][j][2];
					nn[4]=
						hh[map81[term]].m_to_k[index][j][3];
			
					temp_multiplier=-0.5*UnitCell.s2[cell]*UnitCell.s2[nn[4]]; // -2*s*s factor from H.
		
				descriptorIndex=findExcitedClusterIndex((length+1), configuration);
                // may be invalid if outside bounding box
		
				if (Control.file) {
					if (descriptorIndex > 0){
					int indices[7]={0};
					double dummy;
					indices[0]=config_index;
					indices[1]=term; // Jun08: origin was 1..81, now using mapped term
					indices[2]=descriptorIndex; // p(j)
					indices[3]=0;
					indices[4]=0;
					indices[5]=0;
					dummy=temp_multiplier*
					hh[map81[term]].coefficient[index][j];
					collect_term(indices,dummy,3);
				}
			  }
			}
		 }
	    }
      }
    }
  }
  return sum;
}

double term_exc_11(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 11:
     
     -2 * P(j) . F(j) . S(+)(j) . S(z)(i)
     
  */

  int i,j,k,l,m;
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int (*configuration)[5];
  int index;
  int nn[5];
  int neighbour;
//  int maxNumberFlips;
//  int numberFlips;
  int (*conf1)[5], (*conf2)[5];
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
//  int temp_number_conf2_sites;
  double sum;
  sum=0.0000;

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 

  for(k=1;k<=length;k++) {
        
    for(m=1;m<5;m++)
      conf1[1][m]=configuration[k][m]; // P(j)
    
	for(m=1;m<5;m++)
      conf2[1][m]=configuration[k][m]; // F(j)

    counter=1;
    for(m=1;m<=length;m++) {
      if (m != k) {
	tempArray[counter]=m;
	counter++;
      }
    }
    
    counter--;
    
    for(l=0;l<binary_max[counter];l++) {
      
      binary(l,counter,spin);
      conf1_len=1;
      conf2_len=1;
      
      for (j=1;j<=counter;j++) { 
	if (spin[j]) {
	  // P(i)
	  conf1_len++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf1[conf1_len][m]=configuration[temp][m];
	} else {
	  // F(j)
	  conf2_len++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf2[conf2_len][m]=configuration[temp][m];
	}
      }
      
      index=conf1[1][4];
      
      for(neighbour=1;neighbour<=hh[map81[term]].m_to_k_number[index];
	  neighbour++) {
	
	// nearest neighbour is at x,y,z, cell put information
	// into configuration and exit.
	
	/* m ---> k */
	
	nn[1]=conf1[1][1]+hh[map81[term]].m_to_k[index][neighbour][0];
	nn[2]=conf1[1][2]+hh[map81[term]].m_to_k[index][neighbour][1];
	nn[3]=conf1[1][3]+hh[map81[term]].m_to_k[index][neighbour][2];
	nn[4]=hh[map81[term]].m_to_k[index][neighbour][3];
				  
	  descriptorIndex1=findExcitedClusterIndex(conf1_len, conf1);
	  if (descriptorIndex1 <= 0) continue;
	  
	  descriptorIndex2=findClusterIndex(conf2_len, conf2);
	  if (descriptorIndex2 <= 0) continue;
	  
	  multiplier=2*0.5*UnitCell.s2[nn[4]]; // factor -2s
	  
	  if (Control.file) {
	    if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)){
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=term; // Jun08: origin was 1..81, now using mapped term
	      indices[2]=descriptorIndex1; // P(i)
	      indices[3]=descriptorIndex2; // F(j)
	      indices[4]=0;
	      indices[5]=0;
	      dummy=multiplier*hh[map81[term]].coefficient[index][neighbour];
            collect_term(indices,dummy,3);
	    }
	  }
	}
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}  

double term_exc_12(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 12:
     
     -2 * P(j) . F(i) . S(+)(i) . S(z)(j)
     
  */

  int i,j,k,l,m;
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int (*configuration)[5];
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int numberFlips;
  int (*conf1)[5], (*conf2)[5];
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
  int temp_number_conf2_sites;
  double sum;
  sum=0.0000;

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 

  for(k=1;k<=length;k++) {
        
    for(m=1;m<5;m++)
      conf1[1][m]=configuration[k][m]; // F(i)

    counter=1;
    for(m=1;m<=length;m++) {
      if (m != k) {
	tempArray[counter]=m;
	counter++;
      }
    }
    
    counter--;
    
    for(l=0;l<binary_max[counter];l++) {
      
      binary(l,counter,spin);
      conf1_len=1;
      conf2_len=0;
      
      for (j=1;j<=counter;j++) { 
	if (spin[j]) {
	  // F(i)
	  conf1_len++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf1[conf1_len][m]=configuration[temp][m];
	} else {
	  // P(j)
	  conf2_len++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf2[conf2_len][m]=configuration[temp][m];
	}
      }
      
      index=conf1[1][4];
      
      for(neighbour=1;neighbour<=hh[map81[term]].k_to_m_number[index];
	  neighbour++) {
	
	// nearest neighbour is at x,y,z, cell put information
	// into configuration and exit.
	
	/* k ---> m */
	
	nn[1]=conf1[1][1]+hh[map81[term]].k_to_m[index][neighbour][0];
	nn[2]=conf1[1][2]+hh[map81[term]].k_to_m[index][neighbour][1];
	nn[3]=conf1[1][3]+hh[map81[term]].k_to_m[index][neighbour][2];
	nn[4]=hh[map81[term]].k_to_m[index][neighbour][3];
	
	numberFlips=1;
	
	for(m=1;m<=conf2_len;m++)
	  if (nn[1]==conf2[m][1] &&
	      nn[2]==conf2[m][2] &&
	      nn[3]==conf2[m][3] &&
	      nn[4]==conf2[m][4])
	    numberFlips++;
	
	maxNumberFlips=UnitCell.s2[nn[4]];
	
	// the number of spin flips that you can have at a given
	// lattice site is depedent on the quantum spin number
	// of the site in the UNIT CELL index `length'
	
	if (numberFlips <= maxNumberFlips) {
	  
	  temp_number_conf2_sites=conf2_len+1;
	  
	  conf2[temp_number_conf2_sites][1]=nn[1];
	  conf2[temp_number_conf2_sites][2]=nn[2];
	  conf2[temp_number_conf2_sites][3]=nn[3];
	  conf2[temp_number_conf2_sites][4]=nn[4];
	  
	  descriptorIndex1=findClusterIndex(conf1_len, conf1);
	  if (descriptorIndex1 <= 0) continue;
	  
	  descriptorIndex2=findExcitedClusterIndex(temp_number_conf2_sites, conf2);
	  if (descriptorIndex2 <= 0) continue;
	  
	  multiplier=UnitCell.s2[nn[4]]; // -2s from H.
	    
	  if (Control.file) {
	    if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)){
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=term; // Jun08: origin was 1..81, now using mapped term
	      indices[2]=descriptorIndex2; // P(j)
	      indices[3]=descriptorIndex1; // F(i)
	      indices[4]=0;
	      indices[5]=0;
	      dummy=multiplier*hh[map81[term]].coefficient[index][neighbour];
            collect_term(indices,dummy,3);
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}  

double term_exc_13(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 13: 
     
     -2 * P(j) * G(i)(j) . S(+)(i) . S(+)(j)
     
    */

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int i, j, k, l, m;
  int k1;
  int k2;
  //  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  double sum;
  sum=0.0000;

  int (*configuration)[5];

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) 
	  conf1[1][m]=configuration[k1][m]; // G(i,j) <-- i here

	  for(m=1;m<5;m++)
	    conf1[2][m]=configuration[k2][m]; // G(i,j) <- j here

	 for(m=1;m<5;m++) 
	    conf2[1][m]=configuration[k2][m]; // P(j)
	
	
	// loop over nearest neighbours
	
        index = conf1[1][4];
        are_they_neighbours=
          test_k_to_m_neighbour(hh, term, conf1[1], conf2[1]);
	
	// conf1[1] and conf2[1] need to be nearest-neighbours 
	
	if (are_they_neighbours) {
	  
	  
	  counter=1;
	  for(l=1;l<=length;l++) {
	    if (l != k1 && l != k2 ) {
	      tempArray[counter]=l;
	      counter++;
	    }
	  }
	  
	  counter--;
	  
	  for(l=0;l<binary_max[counter];l++) {
	    
	    binary(l,counter,spin);
	    conf1_len=2;
	    conf2_len=1;
	    
	    for (j=1;j<=counter;j++) { 
	      if (spin[j]) {
		// G(i,j)
		conf1_len++;
		temp=tempArray[j];
		for (m=1;m<5;m++) 
		  conf1[conf1_len][m]=configuration[temp][m];
	      } else {
		// P(j)
		conf2_len++;
		temp=tempArray[j];
		for (m=1;m<5;m++) 
		  conf2[conf2_len][m]=configuration[temp][m];
	      }
	    }	      
	    
	    descriptorIndex1=findClusterIndex(conf1_len, conf1);
	    if (descriptorIndex1 <= 0) continue;
	    
	    descriptorIndex2=findExcitedClusterIndex(conf2_len, conf2);
	    if (descriptorIndex2 <= 0) continue;
	    
	    multiplier=-2.0; // from H.
	        
	    if (Control.file) {
	      if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)){
		int indices[7]={0};
		double dummy;
		indices[0]=config_index;
		indices[1]=term; // Jun08: origin was 1..81, now using mapped term
		indices[2]=descriptorIndex2; // p(j)
		indices[3]=descriptorIndex1; // g(i,j)
		indices[4]=0;
		indices[5]=0;
		dummy=multiplier*hh[map81[term]].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,3);
	      }
	    }
	    
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}


double term_exc_14(int config_index, int length, int (*string)[5],
                   struct hamiltonian_defn hh[], int term)
{

  /* TERM 14: 
     
     - 2 . P(j) . F(i) . F(j) . S(+)(i) . S(+)(j)
   
    */

  int spin[MAX_LSUBM+1];
  int counter;
  int mm,nnn,temp1,temp2,temp3;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int (*conf3)[5];
  int i, j, k, l, m;
  int k1;
  int k2;
  // int nn[5];
  int are_they_neighbours;
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  int descriptorIndex3;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int conf3_len;
  int tempArray[MAX_LSUBM+1];
  double sum;
  sum=0.0000;

  int (*configuration)[5];

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf3=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
      conf3[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) conf1[1][m]=configuration[k1][m]; // F(i)
	for(m=1;m<5;m++) conf2[1][m]=configuration[k2][m]; // P(j)
	for(m=1;m<5;m++) conf3[1][m]=configuration[k2][m]; // F(j)
	
	// loop over nearest neighbours
	
        index = conf1[1][4];
        are_they_neighbours=
          test_k_to_m_neighbour(hh, term, conf1[1], conf2[1]);
	
	// conf1[1] and conf2[1] need to be nearest-neighbours 
	
	if (are_they_neighbours) {
	  
	  counter=1;
	  for(l=1;l<=length;l++) {
	    if (l != k1 && l != k2 ) {
	      tempArray[counter]=l;
	      counter++;
	    }
	  }
	  
	  counter--;
	  
          // Joerg: I dont know, why splitting here
          //        but it makes program slow for counter > 10 ???
          //   trimary_max[i] = 3^i, 30*ln3=47.55*ln2 48bit=6*8bit
          // ToDo: simplify spin[i]=000 001 002 010 011 012 020 021 ...
	  if (counter > 30) {
	    printf("TERM 13: counter > 30...\nExiting!!\n\n");
	    exit(1);
	  }
          temp = counter; // split into summands not greater than 10 each
	  temp3 = ((temp>10) ? 10 : temp); temp -= temp3;  // 0..10
	  temp2 = ((temp>10) ? 10 : temp); temp -= temp2;  // 0..10
	  temp1 = ((temp>10) ? 10 : temp); temp -= temp1;  // 0..10
	  
          // trimary_max[i] = 3^i  , try all 3^counter codes, counter<SUBn
	  for(l=0;l<trimary_max[temp1];l++) {
	    for(mm=0;mm<trimary_max[temp2];mm++) {
	      for(nnn=0;nnn<trimary_max[temp3];nnn++) {
		
                // decode counter to spin[i] trimary code
		trimary(l,mm,nnn,counter,spin);
                // now we have next trimary code 000 -> 001 -> 002 -> 010 ...

		conf1_len=1;
		conf2_len=1;
		conf3_len=1;
		
		for (j=1;j<=counter;j++) { 
		  if (spin[j]==0) {         // F(i)
		    conf1_len++;
		    temp=tempArray[j];
		    for (m=1;m<5;m++) 
		      conf1[conf1_len][m]=configuration[temp][m];
		  } else if (spin[j]==1) {  // P(j)
		    conf2_len++;
		    temp=tempArray[j];
		    for (m=1;m<5;m++) 
		      conf2[conf2_len][m]=configuration[temp][m];
		  } else {                   // F(j)
		    conf3_len++;
		    temp=tempArray[j];
		    for (m=1;m<5;m++) 
		      conf3[conf3_len][m]=configuration[temp][m];
		  }
		}
		
		descriptorIndex1=findClusterIndex(conf1_len, conf1);
		if (descriptorIndex1 <= 0) continue;
		
		descriptorIndex2=findExcitedClusterIndex(conf2_len, conf2);
		if (descriptorIndex2 <= 0) continue;
		
		descriptorIndex3=findClusterIndex(conf3_len, conf3);	      
		if (descriptorIndex3 <= 0) continue;
		
		multiplier=-2.0; // from H.
			
		if (Control.file) {
		  if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0) 
		      && (descriptorIndex3 > 0)){
		    int indices[7]={0};
		    double dummy;
		    indices[0]=config_index;
		    indices[1]=term; // Jun08: origin was 1..81, now using mapped term
		    indices[2]=descriptorIndex2; // p(j)
		    indices[3]=descriptorIndex1; // f(i)
		    indices[4]=descriptorIndex3; // f(j)
		    indices[5]=0;
		    dummy=multiplier*hh[map81[term]].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,3);
		  }
		}
		
	      }
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(conf3);
  free(configuration);
  return sum;
}


double term_exc_15(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 15:
     
     -2 * Q(i)(j) . S(+)(i) . S(z)(j)
     
  */

  int i,j,k,l;
  int descriptorIndex;
//  double multiplier;
  double temp_multiplier;
  int (*configuration)[5];
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int numberFlips;
  double sum;
  sum=0.0000;

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  for(k=1;k<=length;k++) {
    
    index=configuration[k][4];
    
    for(neighbour=1;neighbour<=hh[map81[term]].k_to_m_number[index];
	neighbour++) {
      
      // nearest neighbour is at x,y,z, cell put information
      // into configuration and exit.
	
      /* k ---> m */
	
      nn[1]=configuration[k][1]+
	hh[map81[term]].k_to_m[index][neighbour][0];
      nn[2]=configuration[k][2]+
	hh[map81[term]].k_to_m[index][neighbour][1];
      nn[3]=configuration[k][3]+
	hh[map81[term]].k_to_m[index][neighbour][2];
      nn[4]=hh[map81[term]].k_to_m[index][neighbour][3];
	
      numberFlips=1;
	
      for(l=1;l<=length;l++)
	if (nn[1]==configuration[l][1] &&
	    nn[2]==configuration[l][2] &&
	    nn[3]==configuration[l][3] &&
	    nn[4]==configuration[l][4])
	  numberFlips++;
	
      maxNumberFlips=UnitCell.s2[nn[4]];
	
      // the number of spin flips that you can have at a given
      // lattice site is depedent on the quantum spin number
      // of the site in the UNIT CELL index `length'
	
      if (numberFlips <= maxNumberFlips) {
	  
	configuration[length+1][1]=nn[1];
	configuration[length+1][2]=nn[2];
	configuration[length+1][3]=nn[3];
	configuration[length+1][4]=nn[4];
	  
	descriptorIndex=findExcitedClusterIndex((length+1),configuration); 

	temp_multiplier=UnitCell.s2[nn[4]]; // -2s from H.
	  
	if (Control.file) {
	  if (descriptorIndex > 0) {
	    int indices[7]={0};
	    double dummy;
	    indices[0]=config_index;
	    indices[1]=term; // Jun08: origin was 1..81, now using mapped term
	    indices[2]=descriptorIndex; // Q(i,j)
	    indices[3]=0;
	    indices[4]=0;
	    indices[5]=0;
	    dummy=temp_multiplier*hh[map81[term]].coefficient[index][neighbour];
            collect_term(indices,dummy,3);
	  }
	}
	  
      }
    }
  }
  
  free(configuration);
  return sum;
}  


double term_exc_16(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 16: 
     
     -2 * Q(i)(j) . F(j) . S(+)(i) . S(+)(j)
     
    */

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int i, j, k, l, m;
  int k1;
  int k2;
  //  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  double sum;
  sum=0.0000;

  int (*configuration)[5];

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) 
	  conf1[1][m]=configuration[k1][m]; // Q(i,j) <-- i here (k1)
	
	for(m=1;m<5;m++)  
	  conf2[1][m]=configuration[k2][m]; // F(j) <- j here (k2)
	
	// loop over nearest neighbours
	
        index = conf1[1][4];
        are_they_neighbours=
          test_k_to_m_neighbour(hh, term, conf1[1], conf2[1]);
	
	// conf1[1] and conf2[1] need to be nearest-neighbours 
	
	if (are_they_neighbours) {
	  
	  for(m=1;m<5;m++)
	    conf1[2][m]=configuration[k2][m];
	  
	  counter=1;
	  for(l=1;l<=length;l++) {
	    if (l != k1 && l != k2 ) {
	      tempArray[counter]=l;
	      counter++;
	    }
	  }
	  
	  counter--;
	  
	  for(l=0;l<binary_max[counter];l++) {
	    
	    binary(l,counter,spin);
	    conf1_len=2;
	    conf2_len=1;
	    
	    for (j=1;j<=counter;j++) { 
	      if (spin[j]) {
		// Q(i,j)
		conf1_len++;
		temp=tempArray[j];
		for (m=1;m<5;m++) 
		  conf1[conf1_len][m]=configuration[temp][m];
	      } else {
		// F(j)
		conf2_len++;
		temp=tempArray[j];
		for (m=1;m<5;m++) 
		  conf2[conf2_len][m]=configuration[temp][m];
	      }
	    }	      
	    
	    descriptorIndex1=findExcitedClusterIndex(conf1_len, conf1);
	    if (descriptorIndex1 <= 0) continue;
	    
	    descriptorIndex2=findClusterIndex(conf2_len, conf2);
	    if (descriptorIndex2 <= 0) continue;
	    
	    multiplier=-2.0; // from H.
	        
	    if (Control.file) {
	      if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)){
		int indices[7]={0};
		double dummy;
		indices[0]=config_index;
		indices[1]=term; // Jun08: origin was 1..81, now using mapped term
		indices[2]=descriptorIndex1; // q(i,j)
		indices[3]=descriptorIndex2; // f(j)
		indices[4]=0;
		indices[5]=0;
		dummy=multiplier*hh[map81[term]].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,3);
	      }
	    }
	    
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}

double term_exc_17(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 17:
     
     - Q(j,j) . S(+)(j) . S(z)(i)
     
  */

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i,j,k,l,m;
  int descriptorIndex;
  double multiplier;
  int (*configuration)[5];
  int (*conf)[5];
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int maxNumberFlips2;
  int numberFlips2;
  int temp_length;
  double temp_multiplier;
  double sum;
  sum=0.0000;

  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));
  conf=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++) 
      configuration[i][j]=string[i][j];
  
  for(i=1;i<=MAX_LSUBM;i++)
    for(j=1;j<5;j++) 
      conf[i][j]=0;

  multiplier=1.0;

  for(k=1;k<=length;k++) {
    
    index=configuration[k][4];
    maxNumberFlips=UnitCell.s2[index];

    if (maxNumberFlips > 1) {
      
      for(j=1;j<=length;j++)
	for(m=1;m<5;m++)
	  conf[j][m]=configuration[j][m]; // Q(j,j) <- 1st m
      
      temp_length=length+1;
      
      // we repeat the `j' 
      
      for(m=1;m<5;m++)
	conf[temp_length][m]=configuration[k][m]; // H(m,m) <- 2nd m

      index=configuration[k][4];
      
      for(neighbour=1;neighbour<=
	    hh[map81[term]].m_to_k_number[index];
	  neighbour++) {
		
	/* m ---> k */
	
	nn[1]=configuration[k][1]+
	  hh[map81[term]].m_to_k[index][neighbour][0];
	nn[2]=configuration[k][2]+
	  hh[map81[term]].m_to_k[index][neighbour][1];
	nn[3]=configuration[k][3]+
	  hh[map81[term]].m_to_k[index][neighbour][2];
	nn[4]=hh[map81[term]].m_to_k[index][neighbour][3];
	
	numberFlips2=0;
	
	for(l=1;l<=temp_length;l++)
	  if (configuration[k][1]==conf[l][1] &&
	      configuration[k][2]==conf[l][2] &&
	      configuration[k][3]==conf[l][3] &&
	      configuration[k][4]==conf[l][4])
	    numberFlips2++;
	
	maxNumberFlips2=UnitCell.s2[index];

	if (numberFlips2 <= maxNumberFlips2)  {
	  
	  descriptorIndex=findExcitedClusterIndex(temp_length,conf);
	  
	  temp_multiplier=0.5*UnitCell.s2[nn[4]]; // facot -1s from H.
	  
	  if (Control.file) {
	    if (descriptorIndex > 0) {
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=term; // Jun08: origin was 1..81, now using mapped term
	      indices[2]=descriptorIndex; // Q(j,j)
	      indices[3]=0;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=temp_multiplier*
		hh[map81[term]].coefficient[index][neighbour];
            collect_term(indices,dummy,3);
	    }
	  }
	  
	}
      }
    }
  }
  free(configuration);
  free(conf);
  return sum;
}


double term_exc_18(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 18: 
     
     - Q(j,j) . F(i) . S(+)(i) . S(+)(j)
     
    */

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int maxNumberFlips1;
  int maxNumberFlips;
  int numberFlips1;
  double sum;
  sum=0.0000;

  int (*configuration)[5];

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {

	maxNumberFlips=UnitCell.s2[configuration[k1][4]];
	
	if (maxNumberFlips > 1) {
	  
	  for(m=1;m<5;m++) {
	    conf1[1][m]=configuration[k1][m]; // Q(m,m) m--1st one
	    conf1[2][m]=configuration[k1][m]; // Q(m,m) m--2nd one
	    conf2[1][m]=configuration[k2][m]; // F(i)   <- k
	  }
	  
	  // loop over nearest neighbours
	  
	  are_they_neighbours=0;
	  index=conf1[1][4];      
	  
	  for(j=1;j<=hh[map81[term]].m_to_k_number[index];j++) {
	    
	    // nearest neighbour is at x,y,z, temp put information
	    // into configuration and exit.
	    
	    /* m ---> k */
	    
	    nn[1]=conf1[1][1]+hh[map81[term]].m_to_k[index][j][0];
	    nn[2]=conf1[1][2]+hh[map81[term]].m_to_k[index][j][1];
	    nn[3]=conf1[1][3]+hh[map81[term]].m_to_k[index][j][2];
	    nn[4]=hh[map81[term]].m_to_k[index][j][3];
	    
	    if(conf2[1][1] == nn[1] &&
	       conf2[1][2] == nn[2] &&
	       conf2[1][3] == nn[3] &&
	       conf2[1][4] == nn[4])
	      are_they_neighbours=j;
  
	    if (are_they_neighbours) break;    
	  }
	  
	  // conf1[1] and conf2[1] need to be nearest-neighbours 
	  
	  if (are_they_neighbours) {
	    
	    counter=1;
	    for(l=1;l<=length;l++) {
	      if (l != k1 && l != k2 ) {
		tempArray[counter]=l;
		counter++;
	      }
	    }
	    
	    counter--;
	    
	    for(l=0;l<binary_max[counter];l++) {
	      
	      binary(l,counter,spin);
	      conf1_len=2;
	      conf2_len=1;
	      
	      for (j=1;j<=counter;j++) { 
		if (spin[j]) {
		  // H
		  conf1_len++;
		  temp=tempArray[j];
		  for (m=1;m<5;m++) 
		    conf1[conf1_len][m]=configuration[temp][m];
		} else {
		  // H
		  conf2_len++;
		  temp=tempArray[j];
		  for (m=1;m<5;m++) 
		    conf2[conf2_len][m]=configuration[temp][m];
		}
	      }	      
	      
	      // make sure that we don't have too many `k' sites.
	      
	      numberFlips1=0;
	      
	      for(m=1;m<=conf1_len;m++) {
		if (configuration[k1][1]==conf1[m][1] &&
		    configuration[k1][2]==conf1[m][2] &&
		    configuration[k1][3]==conf1[m][3] &&
		    configuration[k1][4]==conf1[m][4])
		  numberFlips1++;
	      }
	      maxNumberFlips1=UnitCell.s2[configuration[k1][4]];
	      
	      if (numberFlips1 <= maxNumberFlips1) {
		
		descriptorIndex1=findExcitedClusterIndex(conf1_len, conf1);
		if (descriptorIndex1 <= 0) continue;
		
		descriptorIndex2=findClusterIndex(conf2_len, conf2);
		if (descriptorIndex2 <= 0) continue;

		multiplier=-1.0; // from H.
		
		if (Control.file) {
		  if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)) {
		    int indices[7]={0};
		    double dummy;
		    indices[0]=config_index;
		    indices[1]=term; // Jun08: origin was 1..81, now using mapped term
		    indices[2]=descriptorIndex1; // Q(j,j)
		    indices[3]=descriptorIndex2; // F(i)
		    indices[4]=0;
		    indices[5]=0;
		    dummy=multiplier*
		      hh[map81[term]].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,3);
		  }
		}

	      }
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}

double term_exc_19(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 19: 

     - L(i,j,j) . s(+)(i) . S(+)(j).

  */

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index, index1;
  int maxNumberFlips1;
  int numberFlips1;
  int descriptorIndex;
  // double multiplier;
  double temp_multiplier;
  int (*configuration)[5];
  int (*conf)[5];
  int temp_length;
  double sum;
  sum=0.0000;

  configuration=(int (*)[5])malloc((MAX_LSUBM+3)*5*sizeof(int));
  conf=(int (*)[5])malloc((MAX_LSUBM+3)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  for(k=1;k<MAX_LSUBM+3;k++) 
    for(l=1;l<5;l++) 
      conf[k][l]=0;
    
   
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      
      if (k1 != k2) {

	for(i=1;i<=length;i++)
	  for(j=1;j<5;j++)
	    conf[i][j]=configuration[i][j];
	
	// a check here...
	
	// now add in extra `j' site for L(i,j,j)
	
	temp_length=length+1;
	
	for(m=1;m<5;m++)
	  conf[temp_length][m]=configuration[k1][m]; // `m'
	
	numberFlips1=0;
	
	for(l=1;l<=temp_length;l++) {
	  if (configuration[k1][1]==conf[l][1] &&
	      configuration[k1][2]==conf[l][2] &&
	      configuration[k1][3]==conf[l][3] &&
	      configuration[k1][4]==conf[l][4])
	    numberFlips1++;
	
	}

	index1=configuration[k1][4];
	maxNumberFlips1=UnitCell.s2[index1];

	are_they_neighbours=0;
	index=configuration[k1][4]; 
	
	for(j=1;j<=hh[map81[term]].m_to_k_number[index];j++) {
	  
	  // nearest neighbour is at x,y,z, cell put information
	  // into configuration and exit.
	  
	  // m -- > k

	  nn[1]=configuration[k1][1]+
	    hh[map81[term]].m_to_k[index][j][0];
	  nn[2]=configuration[k1][2]+
	    hh[map81[term]].m_to_k[index][j][1];
	  nn[3]=configuration[k1][3]+
	    hh[map81[term]].m_to_k[index][j][2];
	  nn[4]=hh[map81[term]].m_to_k[index][j][3];
	  
	  if(configuration[k2][1] == nn[1] &&
	     configuration[k2][2] == nn[2] &&
	     configuration[k2][3] == nn[3] &&
	     configuration[k2][4] == nn[4])
	    are_they_neighbours=j;      

	  if (are_they_neighbours) break;
	}
	
	if ((maxNumberFlips1 > 1) && 
	    (numberFlips1 <= maxNumberFlips1) &&
	    (are_they_neighbours)) {
	  
	  descriptorIndex=findExcitedClusterIndex(temp_length, conf);
	  
	  temp_multiplier=-1.0; // from H.

	  
	  if (Control.file) {
	    if (descriptorIndex > 0) {
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=term; // Jun08: origin was 1..81, now using mapped term
	      indices[2]=descriptorIndex; // L(i,j,j)
	      indices[3]=0;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=temp_multiplier*
		hh[map81[term]].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,3);
	    }
	  }

	}
      }
    }
  }
  free(configuration);
  free(conf);
  return sum;
}

// s+ s^z

double term_exc_20(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 20: 
     
     P(j) . S(+)(i) . S(+)(j).
     
  */
  
  int i, j, l;
  int k1;
  int k2;
  // int nn[5];
  int are_they_neighbours;
  int index;
  int descriptorIndex;
  double multiplier;
  int (*configuration)[5];
  int (*conf)[5];
  int tempConfig[MAX_LSUBM+2];
  int counter;
  double sum;
  sum=0.0000;

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf=         (int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++) {
      configuration[i][j]=string[i][j];
      conf[i][j]=0;
    }
  
  multiplier=1.0;
    
  for(k1=1;k1<=length;k1++) { // m here
    for(k2=1;k2<=length;k2++) { // k here
      
      if (k1 != k2) {  // ToDo: better k1 < k2 and factor 2 ??? Joerg
	
	for(l=1;l<5;l++) { conf[1][l]=configuration[k1][l]; }
	for(l=1;l<5;l++) { tempConfig[l]=configuration[k2][l]; }
	
	// loop over nearest neighbours
	
        index = conf[1][4];
        are_they_neighbours=
          test_m_to_k_neighbour(hh, term, conf[1], tempConfig);
	
	// conf[1][] and tempConfig[] need to be nearest-neighbours 
	
	// now load the rest into conf
	
	counter=1;
	for(i=1;i<=length;i++) {
	  if ((i != k2) && (i != k1)) {
	    counter++;
	    for(l=1;l<5;l++)
	      conf[counter][l]=configuration[i][l];
	  }
	}
	
	if (counter != (length-1)) {
	  printf("TERM 20: Error counter != (length-1)...\nExiting\n\n");
	  exit(1);
	}
	
	if (are_they_neighbours) {
	  
	  descriptorIndex=findExcitedClusterIndex(length-1,conf);
	  
	  // if (descriptorIndex > 0)
	  //  sum+=multiplier*CCMcoeffs[descriptorIndex]*
	  //    hh[map81[term]].coefficient[index][are_they_neighbours];
	  
	  if (Control.file) {
	    if (descriptorIndex > 0) {
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=term; // Jun08: origin was 1..81, now using mapped term
	      indices[2]=descriptorIndex; // P(j)
	      indices[3]=0;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=hh[map81[term]].coefficient[index][are_they_neighbours];
              collect_term(indices,dummy,3);
	    }
	  }
	  
	}
      }
    }
  }
  free(configuration);
  free(conf);
  return sum;
}

// s^- s^z

double term_exc_21(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 21:
     
     -2 * P(j) . F(i) . S(+)(j) . S(z)(i)
     
  */

  int i,j,k,l,m;
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int (*configuration)[5];
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int numberFlips;
  int (*conf1)[5], (*conf2)[5];
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
  int temp_number_conf2_sites;
  double sum;
  sum=0.0000;

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 

  for(k=1;k<=length;k++) {
        
    for(m=1;m<5;m++)
      conf1[1][m]=configuration[k][m]; // P(j)

    counter=1;
    for(m=1;m<=length;m++) {
      if (m != k) {
	tempArray[counter]=m;
	counter++;
      }
    }
    
    counter--;
    
    for(l=0;l<binary_max[counter];l++) {
      
      binary(l,counter,spin);
      conf1_len=1;
      conf2_len=0;
      
      for (j=1;j<=counter;j++) { 
	if (spin[j]) {
	  // P(j)
	  conf1_len++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf1[conf1_len][m]=configuration[temp][m];
	} else {
	  // F(i)
	  conf2_len++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf2[conf2_len][m]=configuration[temp][m];
	}
      }
      
      index=conf1[1][4];
      
      for(neighbour=1;neighbour<=hh[map81[term]].m_to_k_number[index];
	  neighbour++) {
	
	// nearest neighbour is at x,y,z, cell put information
	// into configuration and exit.
	
	/* m ---> k */
	
	nn[1]=conf1[1][1]+hh[map81[term]].m_to_k[index][neighbour][0];
	nn[2]=conf1[1][2]+hh[map81[term]].m_to_k[index][neighbour][1];
	nn[3]=conf1[1][3]+hh[map81[term]].m_to_k[index][neighbour][2];
	nn[4]=hh[map81[term]].m_to_k[index][neighbour][3];
	
	numberFlips=1;
	
	for(m=1;m<=conf2_len;m++)
	  if (nn[1]==conf2[m][1] &&
	      nn[2]==conf2[m][2] &&
	      nn[3]==conf2[m][3] &&
	      nn[4]==conf2[m][4])
	    numberFlips++;
	
	maxNumberFlips=UnitCell.s2[nn[4]];
	
	// the number of spin flips that you can have at a given
	// lattice site is depedent on the quantum spin number
	// of the site in the UNIT CELL index `length'
	
	if (numberFlips <= maxNumberFlips) {
	  
	  temp_number_conf2_sites=conf2_len+1;
	  
	  conf2[temp_number_conf2_sites][1]=nn[1];
	  conf2[temp_number_conf2_sites][2]=nn[2];
	  conf2[temp_number_conf2_sites][3]=nn[3];
	  conf2[temp_number_conf2_sites][4]=nn[4];
	  
	  descriptorIndex1=findExcitedClusterIndex(conf1_len, conf1);
	  if (descriptorIndex1 <= 0) continue;
	  
	  descriptorIndex2=findClusterIndex(temp_number_conf2_sites, conf2);
	  if (descriptorIndex2 <= 0) continue;
	  
	  multiplier=UnitCell.s2[nn[4]]; // factor -2s
	    
	  if (Control.file) {
	    if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)){
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=term; // Jun08: origin was 1..81, now using mapped term
	      indices[2]=descriptorIndex1; // P(j)
	      indices[3]=descriptorIndex2; // F(i)
	      indices[4]=0;
	      indices[5]=0;
	      dummy=multiplier*hh[map81[term]].coefficient[index][neighbour];
            collect_term(indices,dummy,3);
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}  

double term_exc_22(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 22: 
     
     - P(j) . G(i)(i) . S(+)(i) . S(+)(j)
     
    */

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int maxNumberFlips1;
  int maxNumberFlips;
  int numberFlips1;
  double sum;
  sum=0.0000;

  int (*configuration)[5];

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {

	maxNumberFlips=UnitCell.s2[configuration[k1][4]];
	
	if (maxNumberFlips > 1) {
	  
	  for(m=1;m<5;m++) {
	    conf1[1][m]=configuration[k1][m]; // G(i,i) i--1st one
	    conf1[2][m]=configuration[k1][m]; // G(i,i) i--2nd one
	    conf2[1][m]=configuration[k2][m]; // P(j)   <- note different order than normal here!
	  }
	  
	  // loop over nearest neighbours
	  
	  are_they_neighbours=0;
	  index=conf1[1][4];      
	  
	  for(j=1;j<=hh[map81[term]].k_to_m_number[index];j++) {
	    
	    // nearest neighbour is at x,y,z, temp put information
	    // into configuration and exit.
	    
	    /* k ---> m */
	    
	    nn[1]=conf1[1][1]+hh[map81[term]].k_to_m[index][j][0];
	    nn[2]=conf1[1][2]+hh[map81[term]].k_to_m[index][j][1];
	    nn[3]=conf1[1][3]+hh[map81[term]].k_to_m[index][j][2];
	    nn[4]=hh[map81[term]].k_to_m[index][j][3];
	    
	    if(conf2[1][1] == nn[1] &&
	       conf2[1][2] == nn[2] &&
	       conf2[1][3] == nn[3] &&
	       conf2[1][4] == nn[4])
	      are_they_neighbours=j;
  
	    if (are_they_neighbours) break;    
	  }
	  
	  // conf1[1] and conf2[1] need to be nearest-neighbours 
	  
	  if (are_they_neighbours) {
	    
	    counter=1;
	    for(l=1;l<=length;l++) {
	      if (l != k1 && l != k2 ) {
		tempArray[counter]=l;
		counter++;
	      }
	    }
	    
	    counter--;
	    
	    for(l=0;l<binary_max[counter];l++) {
	      
	      binary(l,counter,spin);
	      conf1_len=2;
	      conf2_len=1;
	      
	      for (j=1;j<=counter;j++) { 
		if (spin[j]) {
		  // G
		  conf1_len++;
		  temp=tempArray[j];
		  for (m=1;m<5;m++) 
		    conf1[conf1_len][m]=configuration[temp][m];
		} else {
		  // P
		  conf2_len++;
		  temp=tempArray[j];
		  for (m=1;m<5;m++) 
		    conf2[conf2_len][m]=configuration[temp][m];
		}
	      }	      
	      
	      // make sure that we don't have too many `k' sites.
	      
	      numberFlips1=0;
	      
	      for(m=1;m<=conf1_len;m++) {
		if (configuration[k1][1]==conf1[m][1] &&
		    configuration[k1][2]==conf1[m][2] &&
		    configuration[k1][3]==conf1[m][3] &&
		    configuration[k1][4]==conf1[m][4])
		  numberFlips1++;
	      }
	      maxNumberFlips1=UnitCell.s2[configuration[k1][4]];
	      
	      if (numberFlips1 <= maxNumberFlips1) {
		
		descriptorIndex1=findClusterIndex(conf1_len, conf1);
		if (descriptorIndex1 <= 0) continue;
		
		descriptorIndex2=findExcitedClusterIndex(conf2_len, conf2);
		if (descriptorIndex2 <= 0) continue;

		multiplier=-1.0; // from H
		
		if (Control.file) {
		  if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)) {
		    int indices[7]={0};
		    double dummy;
		    indices[0]=config_index;
		    indices[1]=term; // Jun08: origin was 1..81, now using mapped term
		    indices[2]=descriptorIndex2; // p(j) -- again the excited-state coeff is always first!
		    indices[3]=descriptorIndex1; // g(i,i)
		    indices[4]=0;
		    indices[5]=0;
		    dummy=multiplier*
		      hh[map81[term]].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,3);
		  }
		}

	      }
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}

double term_exc_23(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 23: 
     
     - P(j) . F(i) . F(i) . S(+)(i) . S(+)(j)
   
    */

  int spin[MAX_LSUBM+1];
  int counter;
  int mm,nnn,temp1,temp2,temp3;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int (*conf3)[5];
  int i, j, k, l, m;
  int k1;
  int k2;
  // int nn[5];
  int are_they_neighbours;
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  int descriptorIndex3;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int conf3_len;
  int tempArray[MAX_LSUBM+1];
  double sum;
  sum=0.0000;

  int (*configuration)[5];

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf3=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
      conf3[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) conf1[1][m]=configuration[k1][m]; // P(j)
	for(m=1;m<5;m++) conf2[1][m]=configuration[k2][m]; // F(i)
	for(m=1;m<5;m++) conf3[1][m]=configuration[k2][m]; // F(i)
	
	// loop over nearest neighbours
	
        index = conf1[1][4];
        are_they_neighbours=
          test_m_to_k_neighbour(hh, term, conf1[1], conf2[1]);
	
	// conf1[1] and conf2[1] need to be nearest-neighbours 
	
	if (are_they_neighbours) {
	  
	  counter=1;
	  for(l=1;l<=length;l++) {
	    if (l != k1 && l != k2 ) {
	      tempArray[counter]=l;
	      counter++;
	    }
	  }
	  
	  counter--;
	  
	  if (counter > 30) {
	    printf("TERM 9: counter > 30...\nExiting!!\n\n");
	    exit(1);
	  }
          temp = counter; // split into summands not greater than 10 each
	  temp3 = ((temp>10) ? 10 : temp); temp -= temp3;  // 0..10
	  temp2 = ((temp>10) ? 10 : temp); temp -= temp2;  // 0..10
	  temp1 = ((temp>10) ? 10 : temp); temp -= temp1;  // 0..10
	  
          // trimary_max[i] = 3^i  , try all 3^counter codes, counter<SUBn
	  for(l=0;l<trimary_max[temp1];l++) {
	    for(mm=0;mm<trimary_max[temp2];mm++) {
	      for(nnn=0;nnn<trimary_max[temp3];nnn++) {
		
                // decode counter to spin[i] trimary code
		trimary(l,mm,nnn,counter,spin);
                // now we have next trimary code 000 -> 001 -> 002 -> 010 ...

		conf1_len=1;
		conf2_len=1;
		conf3_len=1;
		
		for (j=1;j<=counter;j++) { 
		  if (spin[j]==0) {         // P(j)
		    conf1_len++;
		    temp=tempArray[j];
		    for (m=1;m<5;m++) 
		      conf1[conf1_len][m]=configuration[temp][m];
		  } else if (spin[j]==1) {  // F(i)
		    conf2_len++;
		    temp=tempArray[j];
		    for (m=1;m<5;m++) 
		      conf2[conf2_len][m]=configuration[temp][m];
		  } else {                   // F(i)
		    conf3_len++;
		    temp=tempArray[j];
		    for (m=1;m<5;m++) 
		      conf3[conf3_len][m]=configuration[temp][m];
		  }
		}
		
		descriptorIndex1=findExcitedClusterIndex(conf1_len, conf1);
		if (descriptorIndex1 <= 0) continue;
		
		descriptorIndex2=findClusterIndex(conf2_len, conf2);
		if (descriptorIndex2 <= 0) continue;
		
		descriptorIndex3=findClusterIndex(conf3_len, conf3);	      
		if (descriptorIndex3 <= 0) continue;
		
		multiplier=-1.0; // from H.
				
		if (Control.file) {
		  if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0) 
		      && (descriptorIndex3 > 0)){
		    int indices[7]={0};
		    double dummy;
		    indices[0]=config_index;
		    indices[1]=term; // Jun08: origin was 1..81, now using mapped term
		    indices[2]=descriptorIndex1; // p(j)
		    indices[3]=descriptorIndex2;
		    indices[4]=descriptorIndex3;
		    indices[5]=0;
		    dummy=multiplier*hh[map81[term]].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,3);
		  }
		}
		
	      }
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(conf3);
  free(configuration);
  return sum;
}

double term_exc_24(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 24: 
     
     -2 * P(i) . S(z)(i) . S(z)(j)

     one of the most time-consuming terms because you've got to sweep
     through the index `k'

     */

  int i, j, l;
  int descriptorIndex;
//  double multiplier;
  double temp_multiplier;
  double sum;
  sum=0.0000;

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int minX, minY, minZ;
  int maxX, maxY, maxZ;

  int x,y,z,cell;
  int index;

  int numberFlips, maxNumberFlips;

  int (*configuration)[5];
  int nn[5];

  if (length+1+1 > Params.n) return 0;

  configuration=(int (*)[5])malloc((MAX_LSUBM+4)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  int maxPointX;
  int maxPointY;
  int maxPointZ;

  maxPointX=0;
  maxPointY=0;
  maxPointZ=0;
  for(j=1;j<=length;j++) {
    if (maxPointX < abs(string[j][1])) maxPointX=abs(string[j][1]);
    if (maxPointY < abs(string[j][2])) maxPointY=abs(string[j][2]);
    if (maxPointZ < abs(string[j][3])) maxPointZ=abs(string[j][3]);
  }

  if (Params.dim == 1) {
    minX=radiusExc;           minY=0;                minZ=0;
    maxX=radiusExc+maxPointX; maxY=0;                maxZ=0;
  } else if (Params.dim == 2) {
    minX=radiusExc;           minY=radiusExc;           minZ=0;
    maxX=radiusExc+maxPointX; maxY=radiusExc+maxPointY; maxZ=0;
  } else if (Params.dim == 3) {
    minX=radiusExc;           minY=radiusExc;           minZ=radiusExc;
    maxX=radiusExc+maxPointX; maxY=radiusExc+maxPointY; maxZ=radiusExc+maxPointZ;
  } else {
    printf("Really weird error. Dimension (%d)"
     " not recognised...\nExiting||\n\n",Params.dim);
    exit(1);
  }


  for(x=-minX;x<=maxX;x++) { // ToDo: better (faster) x = maxX-m ... m ???
    for(y=-minY;y<=maxY;y++) {
      for(z=-minZ;z<=maxZ;z++) {
    for(cell=1;cell<=UnitCell.number;cell++) {
		
	      configuration[length+1][1]=x;
	      configuration[length+1][2]=y;
	      configuration[length+1][3]=z;
	      configuration[length+1][4]=cell;
	      
	      numberFlips=1;
	      
	      for(l=1;l<=length;l++)
		if (configuration[length+1][1]==configuration[l][1] &&
		    configuration[length+1][2]==configuration[l][2] &&
		    configuration[length+1][3]==configuration[l][3] &&
		    configuration[length+1][4]==configuration[l][4])
		  numberFlips++;
	      
	      maxNumberFlips=UnitCell.s2[configuration[length+1][4]];

	      index=cell;

	      if (numberFlips <= maxNumberFlips) {

				for(j=1;j<=hh[map81[term]].k_to_m_number[index];j++) {
	      
					nn[1]=configuration[length+1][1]
						+hh[map81[term]].k_to_m[index][j][0];
					nn[2]=configuration[length+1][2]
						+hh[map81[term]].k_to_m[index][j][1];
					nn[3]=configuration[length+1][3]
						+hh[map81[term]].k_to_m[index][j][2];
					nn[4]=
						hh[map81[term]].k_to_m[index][j][3];
			
					temp_multiplier=-0.5*UnitCell.s2[cell]*UnitCell.s2[nn[4]]; // -2ss from H.
		
				descriptorIndex=findExcitedClusterIndex((length+1), configuration);
                // may be invalid if outside bounding box
		
				if (Control.file) {
					if (descriptorIndex > 0){
					int indices[7]={0};
					double dummy;
					indices[0]=config_index;
					indices[1]=term; // Jun08: origin was 1..81, now using mapped term
					indices[2]=descriptorIndex; // p(i)
					indices[3]=0;
					indices[4]=0;
					indices[5]=0;
					dummy=temp_multiplier*
					hh[map81[term]].coefficient[index][j];
					collect_term(indices,dummy,3);
				}
			  }
			}
		 }
	    }
      }
    }
  }
  return sum;
}

double term_exc_25(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 25:
     
     -2 * P(i) . F(j) . S(+)(j) . S(z)(i)
     
  */

  int i,j,k,l,m;
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int (*configuration)[5];
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int numberFlips;
  int (*conf1)[5], (*conf2)[5];
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
  int temp_number_conf2_sites;
  double sum;
  sum=0.0000;

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 

  for(k=1;k<=length;k++) {
        
    for(m=1;m<5;m++)
      conf1[1][m]=configuration[k][m]; // F(j)

    counter=1;
    for(m=1;m<=length;m++) {
      if (m != k) {
	tempArray[counter]=m;
	counter++;
      }
    }
    
    counter--;
    
    for(l=0;l<binary_max[counter];l++) {
      
      binary(l,counter,spin);
      conf1_len=1;
      conf2_len=0;
      
      for (j=1;j<=counter;j++) { 
	if (spin[j]) {
	  // F(j)
	  conf1_len++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf1[conf1_len][m]=configuration[temp][m];
	} else {
	  // P(i)
	  conf2_len++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf2[conf2_len][m]=configuration[temp][m];
	}
      }
      
      index=conf1[1][4];
      
      for(neighbour=1;neighbour<=hh[map81[term]].m_to_k_number[index];
	  neighbour++) {
	
	// nearest neighbour is at x,y,z, cell put information
	// into configuration and exit.
	
	/* m ---> k */
	
	nn[1]=conf1[1][1]+hh[map81[term]].m_to_k[index][neighbour][0];
	nn[2]=conf1[1][2]+hh[map81[term]].m_to_k[index][neighbour][1];
	nn[3]=conf1[1][3]+hh[map81[term]].m_to_k[index][neighbour][2];
	nn[4]=hh[map81[term]].m_to_k[index][neighbour][3];
	
	numberFlips=1;
	
	for(m=1;m<=conf2_len;m++)
	  if (nn[1]==conf2[m][1] &&
	      nn[2]==conf2[m][2] &&
	      nn[3]==conf2[m][3] &&
	      nn[4]==conf2[m][4])
	    numberFlips++;
	
	maxNumberFlips=UnitCell.s2[nn[4]];
	
	// the number of spin flips that you can have at a given
	// lattice site is depedent on the quantum spin number
	// of the site in the UNIT CELL index `length'
	
	if (numberFlips <= maxNumberFlips) {
	  
	  temp_number_conf2_sites=conf2_len+1;
	  
	  conf2[temp_number_conf2_sites][1]=nn[1];
	  conf2[temp_number_conf2_sites][2]=nn[2];
	  conf2[temp_number_conf2_sites][3]=nn[3];
	  conf2[temp_number_conf2_sites][4]=nn[4];
	  
	  descriptorIndex1=findClusterIndex(conf1_len, conf1);
	  if (descriptorIndex1 <= 0) continue;
	  
	  descriptorIndex2=findExcitedClusterIndex(temp_number_conf2_sites, conf2);
	  if (descriptorIndex2 <= 0) continue;
	  
	  multiplier=UnitCell.s2[nn[4]]; // -2s
	    
	  if (Control.file) {
	    if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)){
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=term; // Jun08: origin was 1..81, now using mapped term
	      indices[2]=descriptorIndex2; // P(i)
	      indices[3]=descriptorIndex1; // F(j)
	      indices[4]=0;
	      indices[5]=0;
	      dummy=multiplier*hh[map81[term]].coefficient[index][neighbour];
            collect_term(indices,dummy,3);
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}  

double term_exc_26(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 26:
     
     -2 * P(i) . F(i) . S(+)(i) . S(z)(j)
     
  */

  int i,j,k,l,m;
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int (*configuration)[5];
  int index;
  int nn[5];
  int neighbour;
//  int maxNumberFlips;
//  int numberFlips;
  int (*conf1)[5], (*conf2)[5];
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
//  int temp_number_conf2_sites;
  double sum;
  sum=0.0000;

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 

  for(k=1;k<=length;k++) {
        
    for(m=1;m<5;m++)
      conf1[1][m]=configuration[k][m]; // P(i)
    
	for(m=1;m<5;m++)
      conf2[1][m]=configuration[k][m]; // F(i)

    counter=1;
    for(m=1;m<=length;m++) {
      if (m != k) {
	tempArray[counter]=m;
	counter++;
      }
    }
    
    counter--;
    
    for(l=0;l<binary_max[counter];l++) {
      
      binary(l,counter,spin);
      conf1_len=1;
      conf2_len=1;
      
      for (j=1;j<=counter;j++) { 
	if (spin[j]) {
	  // P(i)
	  conf1_len++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf1[conf1_len][m]=configuration[temp][m];
	} else {
	  // F(i)
	  conf2_len++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf2[conf2_len][m]=configuration[temp][m];
	}
      }
      
      index=conf1[1][4];
      
      for(neighbour=1;neighbour<=hh[map81[term]].k_to_m_number[index];
	  neighbour++) {
	
	// nearest neighbour is at x,y,z, cell put information
	// into configuration and exit.
	
	/* k ---> m */
	
	nn[1]=conf1[1][1]+hh[map81[term]].k_to_m[index][neighbour][0];
	nn[2]=conf1[1][2]+hh[map81[term]].k_to_m[index][neighbour][1];
	nn[3]=conf1[1][3]+hh[map81[term]].k_to_m[index][neighbour][2];
	nn[4]=hh[map81[term]].k_to_m[index][neighbour][3];
				  
	  descriptorIndex1=findExcitedClusterIndex(conf1_len, conf1);
	  if (descriptorIndex1 <= 0) continue;
	  
	  descriptorIndex2=findClusterIndex(conf2_len, conf2);
	  if (descriptorIndex2 <= 0) continue;
	  
	  multiplier=UnitCell.s2[nn[4]]; // -2s
	  
	  if (Control.file) {
	    if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)){
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=term; // Jun08: origin was 1..81, now using mapped term
	      indices[2]=descriptorIndex1; // P(i)
	      indices[3]=descriptorIndex2; // F(i)
	      indices[4]=0;
	      indices[5]=0;
	      dummy=multiplier*hh[map81[term]].coefficient[index][neighbour];
            collect_term(indices,dummy,3);
	    }
	  }
	}
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}  


double term_exc_27(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 27: 
     
     -2 * P(i) * G(i)(j) . S(+)(i) . S(+)(j)
     
    */

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int i, j, k, l, m;
  int k1;
  int k2;
  //  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  double sum;
  sum=0.0000;

  int (*configuration)[5];

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) 
	  conf1[1][m]=configuration[k1][m]; // G(i,j) <-- j here
    for(m=1;m<5;m++)
	  conf1[2][m]=configuration[k2][m]; // G(i,j) <- i here.
	 for(m=1;m<5;m++) 
	  conf2[1][m]=configuration[k2][m]; // P(i)
	  
	// loop over nearest neighbours
	
        index = conf1[1][4];
        are_they_neighbours=
          test_m_to_k_neighbour(hh, term, conf1[1], conf2[1]);
	
	// conf1[1] and conf2[1] need to be nearest-neighbours 
	
	if (are_they_neighbours) {
	  
	  counter=1;
	  for(l=1;l<=length;l++) {
	    if (l != k1 && l != k2 ) {
	      tempArray[counter]=l;
	      counter++;
	    }
	  }
	  
	  counter--;
	  
	  for(l=0;l<binary_max[counter];l++) {
	    
	    binary(l,counter,spin);
	    conf1_len=2;
	    conf2_len=1;
	    
	    for (j=1;j<=counter;j++) { 
	      if (spin[j]) {
		// G(i,j)
		conf1_len++;
		temp=tempArray[j];
		for (m=1;m<5;m++) 
		  conf1[conf1_len][m]=configuration[temp][m];
	      } else {
		// P(i)
		conf2_len++;
		temp=tempArray[j];
		for (m=1;m<5;m++) 
		  conf2[conf2_len][m]=configuration[temp][m];
	      }
	    }	      
	    
	    descriptorIndex1=findClusterIndex(conf1_len, conf1);
	    if (descriptorIndex1 <= 0) continue;
	    
	    descriptorIndex2=findExcitedClusterIndex(conf2_len, conf2);
	    if (descriptorIndex2 <= 0) continue;
	    
	    multiplier=-2.0;
	        
	    if (Control.file) {
	      if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)){
		int indices[7]={0};
		double dummy;
		indices[0]=config_index;
		indices[1]=term; // Jun08: origin was 1..81, now using mapped term
		indices[2]=descriptorIndex2; // p(i)
		indices[3]=descriptorIndex1; // g(i,j)
		indices[4]=0;
		indices[5]=0;
		dummy=multiplier*hh[map81[term]].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,3);
	      }
	    }
	    
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}

double term_exc_28(int config_index, int length, int (*string)[5],
                   struct hamiltonian_defn hh[], int term)
{

  /* TERM 28: 
     
     - 2 . P(i) . F(i) . F(j) . S(+)(i) . S(+)(j)
   
    */

  int spin[MAX_LSUBM+1];
  int counter;
  int mm,nnn,temp1,temp2,temp3;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int (*conf3)[5];
  int i, j, k, l, m;
  int k1;
  int k2;
  // int nn[5];
  int are_they_neighbours;
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  int descriptorIndex3;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int conf3_len;
  int tempArray[MAX_LSUBM+1];
  double sum;
  sum=0.0000;

  int (*configuration)[5];

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf3=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
      conf3[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) conf1[1][m]=configuration[k1][m]; // F(j)
	for(m=1;m<5;m++) conf2[1][m]=configuration[k2][m]; // P(i)
	for(m=1;m<5;m++) conf3[1][m]=configuration[k2][m]; // F(i)
	
	// loop over nearest neighbours
	
        index = conf1[1][4];
        are_they_neighbours=
          test_m_to_k_neighbour(hh, term, conf1[1], conf2[1]);
	
	// conf1[1] and conf2[1] need to be nearest-neighbours 
	
	if (are_they_neighbours) {
	  
	  counter=1;
	  for(l=1;l<=length;l++) {
	    if (l != k1 && l != k2 ) {
	      tempArray[counter]=l;
	      counter++;
	    }
	  }
	  
	  counter--;
	  
          // Joerg: I dont know, why splitting here
          //        but it makes program slow for counter > 10 ???
          //   trimary_max[i] = 3^i, 30*ln3=47.55*ln2 48bit=6*8bit
          // ToDo: simplify spin[i]=000 001 002 010 011 012 020 021 ...
	  if (counter > 30) {
	    printf("TERM 28: counter > 30...\nExiting!!\n\n");
	    exit(1);
	  }
          temp = counter; // split into summands not greater than 10 each
	  temp3 = ((temp>10) ? 10 : temp); temp -= temp3;  // 0..10
	  temp2 = ((temp>10) ? 10 : temp); temp -= temp2;  // 0..10
	  temp1 = ((temp>10) ? 10 : temp); temp -= temp1;  // 0..10
	  
          // trimary_max[i] = 3^i  , try all 3^counter codes, counter<SUBn
	  for(l=0;l<trimary_max[temp1];l++) {
	    for(mm=0;mm<trimary_max[temp2];mm++) {
	      for(nnn=0;nnn<trimary_max[temp3];nnn++) {
		
                // decode counter to spin[i] trimary code
		trimary(l,mm,nnn,counter,spin);
                // now we have next trimary code 000 -> 001 -> 002 -> 010 ...

		conf1_len=1;
		conf2_len=1;
		conf3_len=1;
		
		for (j=1;j<=counter;j++) { 
		  if (spin[j]==0) {         // F(j)
		    conf1_len++;
		    temp=tempArray[j];
		    for (m=1;m<5;m++) 
		      conf1[conf1_len][m]=configuration[temp][m];
		  } else if (spin[j]==1) {  // P(i)
		    conf2_len++;
		    temp=tempArray[j];
		    for (m=1;m<5;m++) 
		      conf2[conf2_len][m]=configuration[temp][m];
		  } else {                   // F(i)
		    conf3_len++;
		    temp=tempArray[j];
		    for (m=1;m<5;m++) 
		      conf3[conf3_len][m]=configuration[temp][m];
		  }
		}
		
		descriptorIndex1=findClusterIndex(conf1_len, conf1);
		if (descriptorIndex1 <= 0) continue;
		
		descriptorIndex2=findExcitedClusterIndex(conf2_len, conf2);
		if (descriptorIndex2 <= 0) continue;
		
		descriptorIndex3=findClusterIndex(conf3_len, conf3);	      
		if (descriptorIndex3 <= 0) continue;
		
		multiplier=-2.0;
			
		if (Control.file) {
		  if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0) 
		      && (descriptorIndex3 > 0)){
		    int indices[7]={0};
		    double dummy;
		    indices[0]=config_index;
		    indices[1]=term; // Jun08: origin was 1..81, now using mapped term
		    indices[2]=descriptorIndex2; // p(i)
		    indices[3]=descriptorIndex1; // f(j)
		    indices[4]=descriptorIndex3; // f(i)
		    indices[5]=0;
		    dummy=multiplier*hh[map81[term]].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,3);
		  }
		}
		
	      }
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(conf3);
  free(configuration);
  return sum;
}

double term_exc_29(int config_index, int length, int (*string)[5],
                   struct hamiltonian_defn hh[], int term)
{

  /* TERM 29:
     
     -2 * Q(i)(j) . S(+)(j) . S(z)(i)
     
  */

  int i,j,k,l;
  int descriptorIndex;
//  double multiplier;
  double temp_multiplier;
  int (*configuration)[5];
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int numberFlips;
  double sum;
  sum=0.0000;

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  for(k=1;k<=length;k++) {
    
    index=configuration[k][4];
    
    for(neighbour=1;neighbour<=hh[map81[term]].m_to_k_number[index];
	neighbour++) {
      
      // nearest neighbour is at x,y,z, cell put information
      // into configuration and exit.
	
      /* m ---> k */
	
      nn[1]=configuration[k][1]+
	hh[map81[term]].m_to_k[index][neighbour][0];
      nn[2]=configuration[k][2]+
	hh[map81[term]].m_to_k[index][neighbour][1];
      nn[3]=configuration[k][3]+
	hh[map81[term]].m_to_k[index][neighbour][2];
      nn[4]=hh[map81[term]].m_to_k[index][neighbour][3];
	
      numberFlips=1;
	
      for(l=1;l<=length;l++)
	if (nn[1]==configuration[l][1] &&
	    nn[2]==configuration[l][2] &&
	    nn[3]==configuration[l][3] &&
	    nn[4]==configuration[l][4])
	  numberFlips++;
	
      maxNumberFlips=UnitCell.s2[nn[4]];
	
      // the number of spin flips that you can have at a given
      // lattice site is depedent on the quantum spin number
      // of the site in the UNIT CELL index `length'
	
      if (numberFlips <= maxNumberFlips) {
	  
	configuration[length+1][1]=nn[1];
	configuration[length+1][2]=nn[2];
	configuration[length+1][3]=nn[3];
	configuration[length+1][4]=nn[4];
	  
	descriptorIndex=findExcitedClusterIndex((length+1),configuration);

	temp_multiplier=UnitCell.s2[nn[4]]; // -2s
	  
	if (Control.file) {
	  if (descriptorIndex > 0) {
	    int indices[7]={0};
	    double dummy;
	    indices[0]=config_index;
	    indices[1]=term; // Jun08: origin was 1..81, now using mapped term
	    indices[2]=descriptorIndex; // Q(i,j)
	    indices[3]=0;
	    indices[4]=0;
	    indices[5]=0;
	    dummy=temp_multiplier*hh[map81[term]].coefficient[index][neighbour];
            collect_term(indices,dummy,3);
	  }
	}
	  
      }
    }
  }
  
  free(configuration);
  return sum;
}  

double term_exc_30(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 30: 
     
     -2 * Q(i)(j) . F(i) . S(+)(i) . S(+)(j)
     
    */

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int i, j, k, l, m;
  int k1;
  int k2;
  //  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  double sum;
  sum=0.0000;

  int (*configuration)[5];

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) 
	  conf1[1][m]=configuration[k1][m]; // Q(i,j) <-- j here
	
	for(m=1;m<5;m++) 
	  conf2[1][m]=configuration[k2][m]; // F(i)
	
	// loop over nearest neighbours
	
        index = conf1[1][4];
        are_they_neighbours=
          test_m_to_k_neighbour(hh, term, conf1[1], conf2[1]);
	
	// conf1[1] and conf2[1] need to be nearest-neighbours 
	
	if (are_they_neighbours) {
	  
	  for(m=1;m<5;m++)
	    conf1[2][m]=configuration[k2][m]; // Q(i,j) <- i now!
	  
	  counter=1;
	  for(l=1;l<=length;l++) {
	    if (l != k1 && l != k2 ) {
	      tempArray[counter]=l;
	      counter++;
	    }
	  }
	  
	  counter--;
	  
	  for(l=0;l<binary_max[counter];l++) {
	    
	    binary(l,counter,spin);
	    conf1_len=2;
	    conf2_len=1;
	    
	    for (j=1;j<=counter;j++) { 
	      if (spin[j]) {
		// Q(i,j)
		conf1_len++;
		temp=tempArray[j];
		for (m=1;m<5;m++) 
		  conf1[conf1_len][m]=configuration[temp][m];
	      } else {
		// F(j)
		conf2_len++;
		temp=tempArray[j];
		for (m=1;m<5;m++) 
		  conf2[conf2_len][m]=configuration[temp][m];
	      }
	    }	      
	    
	    descriptorIndex1=findExcitedClusterIndex(conf1_len, conf1);
	    if (descriptorIndex1 <= 0) continue;
	    
	    descriptorIndex2=findClusterIndex(conf2_len, conf2);
	    if (descriptorIndex2 <= 0) continue;
	    
	    multiplier=-2.0;
	        
	    if (Control.file) {
	      if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)){
		int indices[7]={0};
		double dummy;
		indices[0]=config_index;
		indices[1]=term; // Jun08: origin was 1..81, now using mapped term
		indices[2]=descriptorIndex1; // q(i,j)
		indices[3]=descriptorIndex2; // f(i)
		indices[4]=0;
		indices[5]=0;
		dummy=multiplier*hh[map81[term]].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,3);
	      }
	    }
	    
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}

double term_exc_31(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 31:
     
     - Q(i,i) . S(+)(i) . S(z)(j)
     
  */

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i,j,k,l,m;
  int descriptorIndex;
  double multiplier;
  int (*configuration)[5];
  int (*conf)[5];
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int maxNumberFlips2;
  int numberFlips2;
  int temp_length;
  double temp_multiplier;
  double sum;
  sum=0.0000;

  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));
  conf=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++) 
      configuration[i][j]=string[i][j];
  
  for(i=1;i<=MAX_LSUBM;i++)
    for(j=1;j<5;j++) 
      conf[i][j]=0;

  multiplier=1.0;

  for(k=1;k<=length;k++) {
    
    index=configuration[k][4];
    maxNumberFlips=UnitCell.s2[index];

    if (maxNumberFlips > 1) {
      
      for(j=1;j<=length;j++)
	for(m=1;m<5;m++)
	  conf[j][m]=configuration[j][m]; // Q(i,i) <- 1st i
      
      temp_length=length+1;
      
      // we repeat the `j' 
      
      for(m=1;m<5;m++)
	conf[temp_length][m]=configuration[k][m]; // Q(i,i) <- 2nd i

      index=configuration[k][4];
      
      for(neighbour=1;neighbour<=
	    hh[map81[term]].k_to_m_number[index];
	  neighbour++) {
		
	/* k ---> m */
	
	nn[1]=configuration[k][1]+
	  hh[map81[term]].k_to_m[index][neighbour][0];
	nn[2]=configuration[k][2]+
	  hh[map81[term]].k_to_m[index][neighbour][1];
	nn[3]=configuration[k][3]+
	  hh[map81[term]].k_to_m[index][neighbour][2];
	nn[4]=hh[map81[term]].k_to_m[index][neighbour][3];
	
	numberFlips2=0;
	
	for(l=1;l<=temp_length;l++)
	  if (configuration[k][1]==conf[l][1] &&
	      configuration[k][2]==conf[l][2] &&
	      configuration[k][3]==conf[l][3] &&
	      configuration[k][4]==conf[l][4])
	    numberFlips2++;
	
	maxNumberFlips2=UnitCell.s2[index];

	if (numberFlips2 <= maxNumberFlips2)  {
	  
	  descriptorIndex=findExcitedClusterIndex(temp_length,conf);
	  
	  temp_multiplier=0.5*UnitCell.s2[nn[4]]; // -s
	  
	  if (Control.file) {
	    if (descriptorIndex > 0) {
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=term; // Jun08: origin was 1..81, now using mapped term
	      indices[2]=descriptorIndex; // Q(i,i)
	      indices[3]=0;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=temp_multiplier*
		hh[map81[term]].coefficient[index][neighbour];
            collect_term(indices,dummy,3);
	    }
	  }
	  
	}
      }
    }
  }
  free(configuration);
  free(conf);
  return sum;
}


double term_exc_32(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 32: 
     
     - Q(i,i) . F(j) . S(+)(i) . S(+)(j)
     
    */

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int maxNumberFlips1;
  int maxNumberFlips;
  int numberFlips1;
  double sum;
  sum=0.0000;

  int (*configuration)[5];

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {

	maxNumberFlips=UnitCell.s2[configuration[k1][4]];
	
	if (maxNumberFlips > 1) {
	  
	  for(m=1;m<5;m++) {
	    conf1[1][m]=configuration[k1][m]; // Q(i,i) i--1st one
	    conf1[2][m]=configuration[k1][m]; // Q(i,i) i--2nd one
	    conf2[1][m]=configuration[k2][m]; // F(j)   <- j
	  }
	  
	  // loop over nearest neighbours
	  
	  are_they_neighbours=0;
	  index=conf1[1][4];      
	  
	  for(j=1;j<=hh[map81[term]].k_to_m_number[index];j++) {
	    
	    // nearest neighbour is at x,y,z, temp put information
	    // into configuration and exit.
	    
	    /* k ---> m */
	    
	    nn[1]=conf1[1][1]+hh[map81[term]].k_to_m[index][j][0];
	    nn[2]=conf1[1][2]+hh[map81[term]].k_to_m[index][j][1];
	    nn[3]=conf1[1][3]+hh[map81[term]].k_to_m[index][j][2];
	    nn[4]=hh[map81[term]].k_to_m[index][j][3];
	    
	    if(conf2[1][1] == nn[1] &&
	       conf2[1][2] == nn[2] &&
	       conf2[1][3] == nn[3] &&
	       conf2[1][4] == nn[4])
	      are_they_neighbours=j;
  
	    if (are_they_neighbours) break;    
	  }
	  
	  // conf1[1] and conf2[1] need to be nearest-neighbours 
	  
	  if (are_they_neighbours) {
	    
	    counter=1;
	    for(l=1;l<=length;l++) {
	      if (l != k1 && l != k2 ) {
		tempArray[counter]=l;
		counter++;
	      }
	    }
	    
	    counter--;
	    
	    for(l=0;l<binary_max[counter];l++) {
	      
	      binary(l,counter,spin);
	      conf1_len=2;
	      conf2_len=1;
	      
	      for (j=1;j<=counter;j++) { 
		if (spin[j]) {
		  // Q(i,i)
		  conf1_len++;
		  temp=tempArray[j];
		  for (m=1;m<5;m++) 
		    conf1[conf1_len][m]=configuration[temp][m];
		} else {
		  // F(j)
		  conf2_len++;
		  temp=tempArray[j];
		  for (m=1;m<5;m++) 
		    conf2[conf2_len][m]=configuration[temp][m];
		}
	      }	      
	      
	      // make sure that we don't have too many `k' sites.
	      
	      numberFlips1=0;
	      
	      for(m=1;m<=conf1_len;m++) {
		if (configuration[k1][1]==conf1[m][1] &&
		    configuration[k1][2]==conf1[m][2] &&
		    configuration[k1][3]==conf1[m][3] &&
		    configuration[k1][4]==conf1[m][4])
		  numberFlips1++;
	      }
	      maxNumberFlips1=UnitCell.s2[configuration[k1][4]];
	      
	      if (numberFlips1 <= maxNumberFlips1) {
		
		descriptorIndex1=findExcitedClusterIndex(conf1_len, conf1);
		if (descriptorIndex1 <= 0) continue;
		
		descriptorIndex2=findClusterIndex(conf2_len, conf2);
		if (descriptorIndex2 <= 0) continue;

		multiplier=-1.0; // from H.
		
		if (Control.file) {
		  if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)) {
		    int indices[7]={0};
		    double dummy;
		    indices[0]=config_index;
		    indices[1]=term; // Jun08: origin was 1..81, now using mapped term
		    indices[2]=descriptorIndex1; // Q(i,i)
		    indices[3]=descriptorIndex2; // F(j)
		    indices[4]=0;
		    indices[5]=0;
		    dummy=multiplier*
		      hh[map81[term]].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,3);
		  }
		}

	      }
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}

double term_exc_33(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 33: 

     - L(i,i,j) . s(+)(i) . S(+)(j).

  */

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index, index1;
  int maxNumberFlips1;
  int numberFlips1;
  int descriptorIndex;
  // double multiplier;
  double temp_multiplier;
  int (*configuration)[5];
  int (*conf)[5];
  int temp_length;
  double sum;
  sum=0.0000;

  configuration=(int (*)[5])malloc((MAX_LSUBM+3)*5*sizeof(int));
  conf=(int (*)[5])malloc((MAX_LSUBM+3)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  for(k=1;k<MAX_LSUBM+3;k++) 
    for(l=1;l<5;l++) 
      conf[k][l]=0;
    
   
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      
      if (k1 != k2) {

	for(i=1;i<=length;i++)
	  for(j=1;j<5;j++)
	    conf[i][j]=configuration[i][j];
	
	// a check here...
	
	// now add in extra `i' site for L(i,i,j)
	
	temp_length=length+1;
	
	for(m=1;m<5;m++)
	  conf[temp_length][m]=configuration[k1][m]; // `m'
	
	numberFlips1=0;
	
	for(l=1;l<=temp_length;l++) {
	  if (configuration[k1][1]==conf[l][1] &&
	      configuration[k1][2]==conf[l][2] &&
	      configuration[k1][3]==conf[l][3] &&
	      configuration[k1][4]==conf[l][4])
	    numberFlips1++;
	
	}

	index1=configuration[k1][4];
	maxNumberFlips1=UnitCell.s2[index1];

	are_they_neighbours=0;
	index=configuration[k1][4]; 
	
	for(j=1;j<=hh[map81[term]].k_to_m_number[index];j++) {
	  
	  // nearest neighbour is at x,y,z, cell put information
	  // into configuration and exit.
	  
	  // k -- > m

	  nn[1]=configuration[k1][1]+
	    hh[map81[term]].k_to_m[index][j][0];
	  nn[2]=configuration[k1][2]+
	    hh[map81[term]].k_to_m[index][j][1];
	  nn[3]=configuration[k1][3]+
	    hh[map81[term]].k_to_m[index][j][2];
	  nn[4]=hh[map81[term]].k_to_m[index][j][3];
	  
	  if(configuration[k2][1] == nn[1] &&
	     configuration[k2][2] == nn[2] &&
	     configuration[k2][3] == nn[3] &&
	     configuration[k2][4] == nn[4])
	    are_they_neighbours=j;      

	  if (are_they_neighbours) break;
	}
	
	if ((maxNumberFlips1 > 1) && 
	    (numberFlips1 <= maxNumberFlips1) &&
	    (are_they_neighbours)) {
	  
	  descriptorIndex=findExcitedClusterIndex(temp_length, conf);
	  
	  temp_multiplier=-1.0; // from H.

	  
	  if (Control.file) {
	    if (descriptorIndex > 0) {
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=term; // Jun08: origin was 1..81, now using mapped term
	      indices[2]=descriptorIndex; // L(i,i,j)
	      indices[3]=0;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=temp_multiplier*
		hh[map81[term]].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,3);
	    }
	  }

	}
      }
    }
  }
  free(configuration);
  free(conf);
  return sum;
}

// s^+ s^- : terms 34 to 36

double term_exc_34(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 34:
     
     - 2 . P(j) . S(+)(i) . S(z)(j)
     
  */

  int i,j,k,l,m;
  int descriptorIndex;
  // double multiplier;
  double temp_multiplier;
  int (*configuration)[5];
  int (*conf)[5];
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int numberFlips;
  int counter;
  double sum;
  sum=0.0000;

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));
  conf=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++) 
      configuration[i][j]=string[i][j];
      
  for(i=1;i<=MAX_LSUBM+1;i++)
    for(j=1;j<5;j++) 
      conf[i][j]=0;
      
  for(k=1;k<=length;k++) {
    
    index=configuration[k][4];
    
    counter=0;
    for(l=1;l<=length;l++) {
      if (l != k) {
	counter++;
	for(m=1;m<5;m++)
	  conf[counter][m]=configuration[l][m];
      } 
    }
    for(neighbour=1;neighbour<=hh[map81[term]].k_to_m_number[index];
	neighbour++) {
      
      counter=length-1;
            
      /* k ---> m */
      
      nn[1]=configuration[k][1]+
	hh[map81[term]].k_to_m[index][neighbour][0];
      nn[2]=configuration[k][2]+
	hh[map81[term]].k_to_m[index][neighbour][1];
      nn[3]=configuration[k][3]+
	hh[map81[term]].k_to_m[index][neighbour][2];
      nn[4]=hh[map81[term]].k_to_m[index][neighbour][3];
      
      numberFlips=1;
      
      for(l=1;l<=(length-1);l++)
	if (nn[1]==conf[l][1] &&
	    nn[2]==conf[l][2] &&
	    nn[3]==conf[l][3] &&
	    nn[4]==conf[l][4])
	  numberFlips++;
      
      maxNumberFlips=UnitCell.s2[nn[4]];
            
      if (numberFlips <= maxNumberFlips) {
	
	counter++;
	
	conf[counter][1]=nn[1];
	conf[counter][2]=nn[2];
	conf[counter][3]=nn[3];
	conf[counter][4]=nn[4];
	
	if (counter != length) {
	  printf("TERM 37: Error counter != length...\nExiting!!\n\n");
	  exit(1);
	}
	
	descriptorIndex=findExcitedClusterIndex(counter,conf);
	
	temp_multiplier=UnitCell.s2[nn[4]]; // -2 * s factor from H.
		
	if (Control.file) {
	  if (descriptorIndex > 0){
	    int indices[7]={0};
	    double dummy;
	    indices[0]=config_index;
	    indices[1]=term; // Jun08: origin was 1..81, now using mapped term
	    indices[2]=descriptorIndex; // P(j)
	    indices[3]=0;
	    indices[4]=0;
            indices[5]=0;
	    dummy=temp_multiplier*hh[map81[term]].coefficient[index][neighbour];
            collect_term(indices,dummy,3);
	  }
	}
	
      }
    }
  }
  free(configuration);
  free(conf);
  return sum;
}  

double term_exc_35(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{
  /* TERM 35: 

     - 2 . P(j) . F(j) . S(+)(i) . S(+)(j).
     
  */

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int tempString[5];
  double sum;
  sum=0.0000;

  int (*configuration)[5];

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) 
	  conf1[1][m]=configuration[k1][m]; // P(j)
	
	for(m=1;m<5;m++) 
	  conf2[1][m]=configuration[k1][m]; // F(j)
	
	for(m=1;m<5;m++) 
	  tempString[m]=configuration[k2][m]; // s(+)(i)
	
	// loop over nearest neighbours
	
	are_they_neighbours=0;
	index=conf1[1][4];      
	
	for(j=1;j<=hh[map81[term]].m_to_k_number[index];j++) {
	  
	  // nearest neighbour is at x,y,z, temp put information
	  // into configuration and exit.
	  
	  /* m ---> k */
	  
	  nn[1]=conf1[1][1]+hh[map81[term]].m_to_k[index][j][0];
	  nn[2]=conf1[1][2]+hh[map81[term]].m_to_k[index][j][1];
	  nn[3]=conf1[1][3]+hh[map81[term]].m_to_k[index][j][2];
	  nn[4]=hh[map81[term]].m_to_k[index][j][3];
	    
	  if(tempString[1] == nn[1] &&
	     tempString[2] == nn[2] &&
	     tempString[3] == nn[3] &&
	     tempString[4] == nn[4])
	    are_they_neighbours=j;      

	  if (are_they_neighbours) break;
	}

	// conf1[1] and conf2[1] need to be nearest-neighbours 
	
	if (are_they_neighbours) {
	  
	  counter=1;
	  for(l=1;l<=length;l++) {
	    if (l != k1 && l != k2 ) {
	      tempArray[counter]=l;
	      counter++;
	    }
	  }
	  
	  counter--;
	  
	  for(l=0;l<binary_max[counter];l++) {
	    
	    binary(l,counter,spin);
	    conf1_len=1;
	    conf2_len=1;
	    
	    for (j=1;j<=counter;j++) { 
	      if (spin[j]) {
		conf1_len++;
		temp=tempArray[j];
		for (m=1;m<5;m++) 
		  conf1[conf1_len][m]=configuration[temp][m];
	      } else {
		conf2_len++;
		temp=tempArray[j];
		for (m=1;m<5;m++) 
		  conf2[conf2_len][m]=configuration[temp][m];
	      }
	    }	      
	    
	    descriptorIndex1=findExcitedClusterIndex(conf1_len, conf1);
	    if (descriptorIndex1 <= 0) continue;
	    
	    descriptorIndex2=findClusterIndex(conf2_len, conf2);
	    if (descriptorIndex2 <= 0) continue;
	    
	    multiplier=-2.0; // -2 Factor from H.
	        
	    if (Control.file) {
	      if ((descriptorIndex1 > 0)&&(descriptorIndex2 > 0)){
		int indices[7]={0};
		double dummy;
		indices[0]=config_index;
		indices[1]=term; // Jun08: origin was 1..81, now using mapped term
		indices[2]=descriptorIndex1; // P(j)
		indices[3]=descriptorIndex2; // F(j)
		indices[4]=0;
		indices[5]=0;
		dummy=multiplier*hh[map81[term]].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,3);
	      }
	    }
	    
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}

double term_exc_36(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 36: 

     - Q(j,j) . S(+)(i) . S(+)(j).

  */

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i, j, k, l, m;
  int k1;
  int k2;
  int counter;
  int nn[5];
  int are_they_neighbours;
  int index;
  int maxNumberFlips1;
  int numberFlips1;
  int descriptorIndex;
  double multiplier;
  int (*configuration)[5];
  int (*conf)[5];
  int temp_length;
  double sum;
  sum=0.0000;

  configuration=(int (*)[5])malloc((MAX_LSUBM+3)*5*sizeof(int));
  conf=(int (*)[5])malloc((MAX_LSUBM+3)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  for(k=1;k<MAX_LSUBM+3;k++) 
    for(l=1;l<5;l++) 
      conf[k][l]=0;
      
  for(k1=1;k1<=length;k1++) { // j
    for(k2=1;k2<=length;k2++) { // i
      
      if (k1 != k2) {

	counter=1;
	for(i=1;i<=length;i++) {
	  if (i != k2) { 
	    for(j=1;j<5;j++)
	      conf[counter][j]=configuration[i][j]; // Q(j,j)
	    counter++;
	  }
	}
	
	if (counter > length) {
	  printf("Error in term 53...\nExiting!!\n\n");
	  exit(1);
	}
	
	temp_length=length;
	
	for(m=1;m<5;m++)
	  conf[temp_length][m]=configuration[k1][m]; // `i'
	
	numberFlips1=0;
	
	for(l=1;l<=temp_length;l++) {
	  if (configuration[k1][1]==conf[l][1] &&
	      configuration[k1][2]==conf[l][2] &&
	      configuration[k1][3]==conf[l][3] &&
	      configuration[k1][4]==conf[l][4])
	    numberFlips1++;
	}

	are_they_neighbours=0;
	index=configuration[k1][4]; 
	maxNumberFlips1=UnitCell.s2[index];

	for(j=1;j<=hh[map81[term]].m_to_k_number[index];j++) {
	  
	  // nearest neighbour is at x,y,z, cell put information
	  // into configuration and exit.
	  
	  // j --> i or m --> k

	  nn[1]=configuration[k1][1]+
	    hh[map81[term]].m_to_k[index][j][0];
	  nn[2]=configuration[k1][2]+
	    hh[map81[term]].m_to_k[index][j][1];
	  nn[3]=configuration[k1][3]+
	    hh[map81[term]].m_to_k[index][j][2];
	  nn[4]=hh[map81[term]].m_to_k[index][j][3];
	  
	  if(configuration[k2][1] == nn[1] &&
	     configuration[k2][2] == nn[2] &&
	     configuration[k2][3] == nn[3] &&
	     configuration[k2][4] == nn[4])
	    are_they_neighbours=j;      

	  if (are_they_neighbours) break;
	}

	
	if ((maxNumberFlips1 > 1) && 
	    (numberFlips1 <= maxNumberFlips1) &&
	    (are_they_neighbours)) {
	  
	  descriptorIndex=findExcitedClusterIndex(temp_length, conf);
	    
      multiplier=-1.0; // from H. 

	  if (Control.file) {
	    if (descriptorIndex > 0) {
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
              indices[1]=term; // Jun08: origin was 1..81, now using mapped term
	      indices[2]=descriptorIndex; // Q(j,j)
              indices[3]=0;
	      indices[4]=0;
              indices[5]=0;
	      dummy=multiplier*
		hh[map81[term]].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,3);
	    }
	  }

	}
      }
    }
  }
  free(configuration);
  free(conf);
  return sum;
}

// s^- s^+ : terms 37 to 39

double term_exc_37(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 37:
     
     - 2 . P(i) . S(+)(j) . S(z)(i)
     
  */

  int i,j,k,l,m;
  int descriptorIndex;
  // double multiplier;
  double temp_multiplier;
  int (*configuration)[5];
  int (*conf)[5];
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int numberFlips;
  int counter;
  double sum;
  sum=0.0000;

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));
  conf=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++) 
      configuration[i][j]=string[i][j];
      
  for(i=1;i<=MAX_LSUBM+1;i++)
    for(j=1;j<5;j++) 
      conf[i][j]=0;
      
  for(k=1;k<=length;k++) {
    
    index=configuration[k][4];
    
    counter=0;
    for(l=1;l<=length;l++) {
      if (l != k) {
	counter++;
	for(m=1;m<5;m++)
	  conf[counter][m]=configuration[l][m];
      } 
    }
    for(neighbour=1;neighbour<=hh[map81[term]].m_to_k_number[index];
	neighbour++) {
      
      counter=length-1;
            
      /* m ---> k */
      
      nn[1]=configuration[k][1]+
	hh[map81[term]].m_to_k[index][neighbour][0];
      nn[2]=configuration[k][2]+
	hh[map81[term]].m_to_k[index][neighbour][1];
      nn[3]=configuration[k][3]+
	hh[map81[term]].m_to_k[index][neighbour][2];
      nn[4]=hh[map81[term]].m_to_k[index][neighbour][3];
      
      numberFlips=1;
      
      for(l=1;l<=(length-1);l++)
	if (nn[1]==conf[l][1] &&
	    nn[2]==conf[l][2] &&
	    nn[3]==conf[l][3] &&
	    nn[4]==conf[l][4])
	  numberFlips++;
      
      maxNumberFlips=UnitCell.s2[nn[4]];
            
      if (numberFlips <= maxNumberFlips) {
	
	counter++;
	
	conf[counter][1]=nn[1];
	conf[counter][2]=nn[2];
	conf[counter][3]=nn[3];
	conf[counter][4]=nn[4];
	
	if (counter != length) {
	  printf("TERM 34: Error counter != length...\nExiting!!\n\n");
	  exit(1);
	}
	
	descriptorIndex=findExcitedClusterIndex(counter,conf);
	
	temp_multiplier=UnitCell.s2[nn[4]]; // factor of -2 s from H.
		
	if (Control.file) {
	  if (descriptorIndex > 0){
	    int indices[7]={0};
	    double dummy;
	    indices[0]=config_index;
	    indices[1]=term; // Jun08: origin was 1..81, now using mapped term
	    indices[2]=descriptorIndex; // P(i)
	    indices[3]=0;
	    indices[4]=0;
            indices[5]=0;
	    dummy=temp_multiplier*hh[map81[term]].coefficient[index][neighbour];
            collect_term(indices,dummy,3);
	  }
	}
	
      }
    }
  }
  free(configuration);
  free(conf);
  return sum;
}  

double term_exc_38(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{
  /* TERM 38: 

     - 2 . P(i) . F(i) . S(+)(i) . S(+)(j).
     
  */

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int tempString[5];
  double sum;
  sum=0.0000;

  int (*configuration)[5];

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) 
	  conf1[1][m]=configuration[k1][m]; // P(i)
	
	for(m=1;m<5;m++) 
	  conf2[1][m]=configuration[k1][m]; // F(i)
	
	for(m=1;m<5;m++) 
	  tempString[m]=configuration[k2][m]; // s(+)(j)
	
	// loop over nearest neighbours
	
	are_they_neighbours=0;
	index=conf1[1][4];      
	
	for(j=1;j<=hh[map81[term]].k_to_m_number[index];j++) {
	  
	  // nearest neighbour is at x,y,z, temp put information
	  // into configuration and exit.
	  
	  /* k ---> m */
	  
	  nn[1]=conf1[1][1]+hh[map81[term]].k_to_m[index][j][0];
	  nn[2]=conf1[1][2]+hh[map81[term]].k_to_m[index][j][1];
	  nn[3]=conf1[1][3]+hh[map81[term]].k_to_m[index][j][2];
	  nn[4]=hh[map81[term]].k_to_m[index][j][3];
	    
	  if(tempString[1] == nn[1] &&
	     tempString[2] == nn[2] &&
	     tempString[3] == nn[3] &&
	     tempString[4] == nn[4])
	    are_they_neighbours=j;      

	  if (are_they_neighbours) break;
	}

	// conf1[1] and conf2[1] need to be nearest-neighbours 
	
	if (are_they_neighbours) {
	  
	  counter=1;
	  for(l=1;l<=length;l++) {
	    if (l != k1 && l != k2 ) {
	      tempArray[counter]=l;
	      counter++;
	    }
	  }
	  
	  counter--;
	  
	  for(l=0;l<binary_max[counter];l++) {
	    
	    binary(l,counter,spin);
	    conf1_len=1;
	    conf2_len=1;
	    
	    for (j=1;j<=counter;j++) { 
	      if (spin[j]) {
		conf1_len++;
		temp=tempArray[j];
		for (m=1;m<5;m++) 
		  conf1[conf1_len][m]=configuration[temp][m];
	      } else {
		conf2_len++;
		temp=tempArray[j];
		for (m=1;m<5;m++) 
		  conf2[conf2_len][m]=configuration[temp][m];
	      }
	    }	      
	    
	    descriptorIndex1=findExcitedClusterIndex(conf1_len, conf1);
	    if (descriptorIndex1 <= 0) continue;
	    
	    descriptorIndex2=findClusterIndex(conf2_len, conf2);
	    if (descriptorIndex2 <= 0) continue;
	    
	    multiplier=-2.0; // factor from H.
	        
	    if (Control.file) {
	      if ((descriptorIndex1 > 0)&&(descriptorIndex2 > 0)){
		int indices[7]={0};
		double dummy;
		indices[0]=config_index;
		indices[1]=term; // Jun08: origin was 1..81, now using mapped term
		indices[2]=descriptorIndex1; // P(i)
		indices[3]=descriptorIndex2; // F(i)
		indices[4]=0;
		indices[5]=0;
		dummy=multiplier*hh[map81[term]].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,3);
	      }
	    }
	    
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}

double term_exc_39(int config_index, int length, int (*string)[5],
                   struct hamiltonian_defn hh[], int term)
{

  /* TERM 39: 

     - Q(i,i) . S(+)(i) . S(+)(j).

  */

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i, j, k, l, m;
  int k1;
  int k2;
  int counter;
  int nn[5];
  int are_they_neighbours;
  int index;
  int maxNumberFlips1;
  int numberFlips1;
  int descriptorIndex;
  double multiplier;
  int (*configuration)[5];
  int (*conf)[5];
  int temp_length;
  double sum;
  sum=0.0000;

  configuration=(int (*)[5])malloc((MAX_LSUBM+3)*5*sizeof(int));
  conf=(int (*)[5])malloc((MAX_LSUBM+3)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  for(k=1;k<MAX_LSUBM+3;k++) 
    for(l=1;l<5;l++) 
      conf[k][l]=0;
      
  for(k1=1;k1<=length;k1++) { // i
    for(k2=1;k2<=length;k2++) { // j
      
      if (k1 != k2) {

	counter=1;
	for(i=1;i<=length;i++) {
	  if (i != k2) { 
	    for(j=1;j<5;j++)
	      conf[counter][j]=configuration[i][j]; // Q(i,i)
	    counter++;
	  }
	}
	
	if (counter > length) {
	  printf("Error in term 53...\nExiting!!\n\n");
	  exit(1);
	}
	
	temp_length=length;
	
	for(m=1;m<5;m++)
	  conf[temp_length][m]=configuration[k1][m]; // `i'
	
	numberFlips1=0;
	
	for(l=1;l<=temp_length;l++) {
	  if (configuration[k1][1]==conf[l][1] &&
	      configuration[k1][2]==conf[l][2] &&
	      configuration[k1][3]==conf[l][3] &&
	      configuration[k1][4]==conf[l][4])
	    numberFlips1++;
	
	}

	are_they_neighbours=0;
	index=configuration[k1][4]; 
	maxNumberFlips1=UnitCell.s2[index];

	for(j=1;j<=hh[map81[term]].k_to_m_number[index];j++) {
	  
	  // nearest neighbour is at x,y,z, cell put information
	  // into configuration and exit.
	  
	  // i --> j or k --> m

	  nn[1]=configuration[k1][1]+
	    hh[map81[term]].k_to_m[index][j][0];
	  nn[2]=configuration[k1][2]+
	    hh[map81[term]].k_to_m[index][j][1];
	  nn[3]=configuration[k1][3]+
	    hh[map81[term]].k_to_m[index][j][2];
	  nn[4]=hh[map81[term]].k_to_m[index][j][3];
	  
	  if(configuration[k2][1] == nn[1] &&
	     configuration[k2][2] == nn[2] &&
	     configuration[k2][3] == nn[3] &&
	     configuration[k2][4] == nn[4])
	    are_they_neighbours=j;      

	  if (are_they_neighbours) break;
	}

	
	if ((maxNumberFlips1 > 1) && 
	    (numberFlips1 <= maxNumberFlips1) &&
	    (are_they_neighbours)) {
	  
	  descriptorIndex=findExcitedClusterIndex(temp_length, conf);
	    
      multiplier=-1.0; //  -1 from H.

	  if (Control.file) {
	    if (descriptorIndex > 0) {
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
              indices[1]=term; // Jun08: origin was 1..81, now using mapped term
	      indices[2]=descriptorIndex; // Q(i,i)
              indices[3]=0;
	      indices[4]=0;
              indices[5]=0;
	      dummy=multiplier*
		hh[map81[term]].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,3);
	    }
	  }

	}
      }
    }
  }
  free(configuration);
  free(conf);
  return sum;
}

// the difficult one: s^- s^- terms 40 to 82

double term_exc_40(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 40: 
     
     4 * P(i) . F(j) . S(z)(i) . S(z)(j)

     one of the most time-consuming terms because you've got to sweep
     through the index `k'

     */
  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);


  int i, j, k, l, m;
  double multiplier;

  int maxX, maxY, maxZ;
  int minX, minY, minZ;

  int x,y,z,cell;
  int maxPointX, maxPointY, maxPointZ;
  int index;

  int numberFlips, maxNumberFlips;

  int (*configuration)[5];

  int (*conf1)[5], (*conf2)[5];
  int conf1_len;
  int conf2_len;
  int counter;
  int spin[MAX_LSUBM+2];
  int neighbour;
  int temp_x;
  int temp_y;
  int temp_z;
  int temp_cell;
  int temp_number_conf1_sites;
  int temp_number_conf2_sites;
  int descriptorIndex1;
  int descriptorIndex2;
  double sum;
  sum=0.0000;

  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<=MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 

  maxPointX=0;
  maxPointY=0;
  maxPointZ=0;
  for(j=1;j<=length;j++) {
    if (maxPointX < abs(string[j][1])) maxPointX=abs(string[j][1]);
    if (maxPointY < abs(string[j][2])) maxPointY=abs(string[j][2]);
    if (maxPointZ < abs(string[j][3])) maxPointZ=abs(string[j][3]);
  }

  if (Params.dim == 1) {
    minX=radiusExc;           minY=0;                minZ=0;
    maxX=radiusExc+maxPointX; maxY=0;                maxZ=0;
  } else if (Params.dim == 2) {
    minX=radiusExc;           minY=radiusExc;           minZ=0;
    maxX=radiusExc+maxPointX; maxY=radiusExc+maxPointY; maxZ=0;
  } else if (Params.dim == 3) {
    minX=radiusExc;           minY=radiusExc;           minZ=radiusExc;
    maxX=radiusExc+maxPointX; maxY=radiusExc+maxPointY; maxZ=radiusExc+maxPointZ;
  } else {
    printf("Really weird error. Dimension (%d)"
     " not recognised...\nExiting||\n\n",Params.dim);
    exit(1);
  }


  for(x=-minX;x<=maxX;x++) {
    for(y=-minY;y<=maxY;y++) {
      for(z=-minZ;z<=maxZ;z++) {
	for(cell=1;cell<=UnitCell.number;cell++) {
	  
          // ToDo: skip x+y+z outside triangle range [0..m]?

	  index=cell;

	  counter=length;
	  
	  for(l=0;l<binary_max[counter];l++) {
	    
	    binary(l,counter,spin);   /* convert l to binary string spin[] */

            // spin[j]: 0 = F(m)  1 = F(k)
            // copy sites for 0-bit to conf2 and sites for 1-bit to conf1
	    conf1_len=0;
	    conf2_len=0;
	    
	    for (j=1;j<=counter;j++) { 
	      if (spin[j]) {
		conf1_len++;
		for (m=1;m<5;m++) 
		  conf1[conf1_len][m]=configuration[j][m]; // p(i)
	      } else {
		conf2_len++;
		for (m=1;m<5;m++) 
		  conf2[conf2_len][m]=configuration[j][m]; // f(j)
	      }
	    }
	    
	    numberFlips=1;
	    
	    for(m=1;m<=conf1_len;m++)
	      if (x==conf1[m][1] &&
		  y==conf1[m][2] &&
		  z==conf1[m][3] &&
		  cell==conf1[m][4])
		numberFlips++;
	    
	    maxNumberFlips = UnitCell.s2[cell];
	    
	    // the number of spin flips that you can have at a given
	    // lattice site is depedent on the quantum spin number
	    // of the site in the UNIT CELL index `length'
	    
	    if (numberFlips > maxNumberFlips) continue; 

            // Took this outside inner loop (Joerg Sep2005)
            temp_number_conf1_sites = conf1_len+1;
            conf1[temp_number_conf1_sites][1]=x;
            conf1[temp_number_conf1_sites][2]=y;
            conf1[temp_number_conf1_sites][3]=z;
            conf1[temp_number_conf1_sites][4]=cell;
            
            descriptorIndex1=findExcitedClusterIndex(temp_number_conf1_sites, conf1);
            if (descriptorIndex1 <= 0) continue;
	      
            for(neighbour=1;
                neighbour<=hh[map81[term]].k_to_m_number[cell];
                neighbour++) {
              
              // nearest neighbour is at x,y,z, cell put information
              // into configuration and exit.
              
              /* k ---> m */
              
              temp_x=x+hh[map81[term]].k_to_m[cell][neighbour][0];
              temp_y=y+hh[map81[term]].k_to_m[cell][neighbour][1];
              temp_z=z+hh[map81[term]].k_to_m[cell][neighbour][2];
              temp_cell=
                       hh[map81[term]].k_to_m[cell][neighbour][3];
              
              numberFlips=1;
              
              for(m=1;m<=conf2_len;m++)
                if (temp_x==conf2[m][1] &&
                    temp_y==conf2[m][2] &&
                    temp_z==conf2[m][3] &&
                    temp_cell==conf2[m][4])
                  numberFlips++;
              
              maxNumberFlips=UnitCell.s2[temp_cell];
                            
              if (numberFlips <= maxNumberFlips) {
                                
                temp_number_conf2_sites = conf2_len+1;
                conf2[temp_number_conf2_sites][1]=temp_x;
                conf2[temp_number_conf2_sites][2]=temp_y;
                conf2[temp_number_conf2_sites][3]=temp_z;
                conf2[temp_number_conf2_sites][4]=temp_cell;
                
                multiplier=UnitCell.s2[cell]*UnitCell.s2[temp_cell]; // 4 * s* s
                
                // coeff from \tilde s^- \tilde s^- 
                
                descriptorIndex2=findClusterIndex(temp_number_conf2_sites, conf2);
                if (descriptorIndex2 <= 0) continue;
                
                if (Control.file) {
                  if ((descriptorIndex1 > 0)&&(descriptorIndex2 > 0)){
                    int indices[7]={0};
                    double dummy;
                    indices[0]=config_index;
                    indices[1]=term; // Jun08: origin was 1..81, now using mapped term
                    indices[2]=descriptorIndex1; // p(i)
                    indices[3]=descriptorIndex2; // f(j)
                    indices[4]=0;
                    indices[5]=0;
                    dummy=multiplier*
                      hh[map81[term]].coefficient[index][neighbour];
                    collect_term(indices,dummy,3);
                  }
                }
                
              }
            }
	  }
	}
      }
    }
  }
  free(configuration);
  free(conf1);
  free(conf2);
  return sum;
}

double term_exc_41(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{
  
  /* TERM 41:
     
     4 * p(i) . G(i,j) . S(+)(i) . S(z)(j)
     
  */

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i,j,k,l,m;
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int (*configuration)[5];
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int numberFlips;
  int (*conf1)[5], (*conf2)[5];
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
  int temp_number_conf2_sites;
  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));
  double sum;
  sum=0.0000;

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<=MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 

  for(k=1;k<=length;k++) {
        
    for(m=1;m<5;m++) {
      conf1[1][m]=configuration[k][m]; // p(k)
      conf2[1][m]=configuration[k][m]; // g(m,k) <-- k here.
    }
      
    counter=1;
    for(m=1;m<=length;m++) {
      if (m != k) {
	tempArray[counter]=m;
	counter++;
      }
    }
    
    counter--;
    
    for(l=0;l<binary_max[counter];l++) {
      
      binary(l,counter,spin);
      conf1_len=1;
      conf2_len=1;
      
      for (j=1;j<=counter;j++) { 
	if (spin[j]) {
	  // F(k)
	  conf1_len++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf1[conf1_len][m]=configuration[temp][m];
	} else {
	  // H(k,m)
	  conf2_len++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf2[conf2_len][m]=configuration[temp][m];
	}
      }
      
      index=conf2[1][4];
      
      for(neighbour=1;neighbour<=hh[map81[term]].k_to_m_number[index];
	  neighbour++) {
	
	// nearest neighbour is at x,y,z, cell put information
	// into configuration and exit.
	
	/* k ---> m */
	
	nn[1]=conf2[1][1]+hh[map81[term]].k_to_m[index][neighbour][0];
	nn[2]=conf2[1][2]+hh[map81[term]].k_to_m[index][neighbour][1];
	nn[3]=conf2[1][3]+hh[map81[term]].k_to_m[index][neighbour][2];
	nn[4]=            hh[map81[term]].k_to_m[index][neighbour][3];
	
	numberFlips=1;
	
	for(m=1;m<=conf2_len;m++)
	  if (nn[1]==conf2[m][1] &&
	      nn[2]==conf2[m][2] &&
	      nn[3]==conf2[m][3] &&
	      nn[4]==conf2[m][4])
	    numberFlips++;
	
	maxNumberFlips=UnitCell.s2[nn[4]];
	
	// the number of spin flips that you can have at a given
	// lattice site is depedent on the quantum spin number
	// of the site in the UNIT CELL index `length'
	
	if (numberFlips <= maxNumberFlips) {
	  
	  temp_number_conf2_sites=conf2_len+1;
	  
	  conf2[temp_number_conf2_sites][1]=nn[1];
	  conf2[temp_number_conf2_sites][2]=nn[2];
	  conf2[temp_number_conf2_sites][3]=nn[3];
	  conf2[temp_number_conf2_sites][4]=nn[4];
	  
	  descriptorIndex1=findExcitedClusterIndex(conf1_len, conf1);
	  if (descriptorIndex1 <= 0) continue;
	  
	  descriptorIndex2=findClusterIndex(temp_number_conf2_sites, conf2);
	  if (descriptorIndex2 <= 0) continue;
	  
	  multiplier=-2*UnitCell.s2[nn[4]]; //  - 4 * s
	      
	  if (Control.file) {
	    if ((descriptorIndex1 > 0)&&(descriptorIndex2 > 0)) {
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=term; // Jun08: origin was 1..81, now using mapped term
	      indices[2]=descriptorIndex1; //p(i)
	      indices[3]=descriptorIndex2; //g(i,j)
	      indices[4]=0;
	      indices[5]=0;
	      dummy=multiplier*
		hh[map81[term]].coefficient[index][neighbour];
            collect_term(indices,dummy,3);
	    }
	  }
	  
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}  

double term_exc_42(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{
  
  /* TERM 42:
     
     2 * p(i) . G(j,j) . S(+)(j) . S(z)(i)
     
  */

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i,j,k,l,m;
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int (*configuration)[5];
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips1;
  int numberFlips1;
  int maxNumberFlips2;
  int numberFlips2;
  int (*conf1)[5], (*conf2)[5];
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
  int temp_number_conf2_sites;
  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));
  double sum;
  sum=0.0000;

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<=MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 

  for(k=1;k<=length;k++) {
        
    for(m=1;m<5;m++) {
      conf1[1][m]=configuration[k][m]; // g(j) <- first one
      conf1[2][m]=configuration[k][m]; // g(j) <- second one
    }
      
    counter=1;
    for(m=1;m<=length;m++) {
      if (m != k) {
	tempArray[counter]=m;
	counter++;
      }
    }
    
    counter--;
    
    for(l=0;l<binary_max[counter];l++) {
      
      binary(l,counter,spin);
      conf1_len=2;
      conf2_len=0;
      
      for (j=1;j<=counter;j++) { 
	if (spin[j]) {
	  // G(j,j)
	  conf1_len++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf1[conf1_len][m]=configuration[temp][m];
	} else {
	  // P(i)
	  conf2_len++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf2[conf2_len][m]=configuration[temp][m];
	}
      }
      
      index=conf2[1][4];
      
      for(neighbour=1;neighbour<=hh[map81[term]].m_to_k_number[index];
	  neighbour++) {
	
	// nearest neighbour is at x,y,z, cell put information
	// into configuration and exit.
	
	/* m ---> k */
	
	// should I use conf1 or conf2 here? 
	
	nn[1]=conf1[1][1]+hh[map81[term]].m_to_k[index][neighbour][0];
	nn[2]=conf1[1][2]+hh[map81[term]].m_to_k[index][neighbour][1];
	nn[3]=conf1[1][3]+hh[map81[term]].m_to_k[index][neighbour][2];
	nn[4]=            hh[map81[term]].m_to_k[index][neighbour][3];
	
	  temp_number_conf2_sites=conf2_len+1;
	  
	  conf2[temp_number_conf2_sites][1]=nn[1]; // P(i) 
	  conf2[temp_number_conf2_sites][2]=nn[2];
	  conf2[temp_number_conf2_sites][3]=nn[3];
	  conf2[temp_number_conf2_sites][4]=nn[4];
	  
	  numberFlips1=0;
	
	for(m=1;m<=temp_number_conf2_sites;m++)
	  if (nn[1]==conf2[m][1] &&
	      nn[2]==conf2[m][2] &&
	      nn[3]==conf2[m][3] &&
	      nn[4]==conf2[m][4])
	    numberFlips1++;
	
	maxNumberFlips1=UnitCell.s2[nn[4]]; 
	
	// the number of spin flips that you can have at a given
	// lattice site is depedent on the quantum spin number
	// of the site in the UNIT CELL index `length'

	  numberFlips2=2;
	for(m=3;m<=conf1_len;m++)
	  if (conf1[1][1]==conf1[m][1] &&
	      conf1[1][2]==conf1[m][2] &&
	      conf1[1][3]==conf1[m][3] &&
	      conf1[1][4]==conf1[m][4])
	    numberFlips2++;
	
	maxNumberFlips2=UnitCell.s2[conf1[1][4]];
	
	if ((numberFlips1 <= maxNumberFlips1) && (numberFlips2 <= maxNumberFlips2)) {

          // g(jj)	  
	  descriptorIndex1=findClusterIndex(conf1_len, conf1);
	  if (descriptorIndex1 <= 0) continue;
	  
	  // p(i)
	  descriptorIndex2=findExcitedClusterIndex(temp_number_conf2_sites, conf2);
	  if (descriptorIndex2 <= 0) continue;
	  
	  multiplier=-UnitCell.s2[nn[4]]; // -2 * s
	      
	  if (Control.file) {
	    if ((descriptorIndex1 > 0)&&(descriptorIndex2 > 0)) {
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=term; // Jun08: origin was 1..81, now using mapped term
	      indices[2]=descriptorIndex2; //p(i)
	      indices[3]=descriptorIndex1; //g(j,j)
	      indices[4]=0;
	      indices[5]=0;
	      dummy=multiplier*
		hh[map81[term]].coefficient[index][neighbour];
            collect_term(indices,dummy,3);
	    }
	  }
	  
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}  

double term_exc_43(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 43: 
     
     2 * P(i) . H(i,j,j) . S(+)(i) . S(+)(j)
     
    */

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int maxNumberFlips1;
  int maxNumberFlips;
  int numberFlips1;
  double sum;
  sum=0.0000;

  int (*configuration)[5];

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {

	maxNumberFlips=UnitCell.s2[configuration[k1][4]];
	
	if (maxNumberFlips > 1) {
	  
	  for(m=1;m<5;m++) {
	    conf1[1][m]=configuration[k1][m]; // H(j,j,i) m--1st one
	    conf1[2][m]=configuration[k1][m]; // H(j,j,i) m--2nd one
	    conf1[3][m]=configuration[k2][m]; // H(j,j,i) k 
	    conf2[1][m]=configuration[k2][m]; // P(k)     k
	  }
	  
	  // loop over nearest neighbours
	  
	  are_they_neighbours=0;
	  index=conf1[1][4];      
	  
	  for(j=1;j<=hh[map81[term]].m_to_k_number[index];j++) {
	    
	    // nearest neighbour is at x,y,z, temp put information
	    // into configuration and exit.
	    
	    /* m ---> k */
	    
	    nn[1]=conf1[1][1]+hh[map81[term]].m_to_k[index][j][0];
	    nn[2]=conf1[1][2]+hh[map81[term]].m_to_k[index][j][1];
	    nn[3]=conf1[1][3]+hh[map81[term]].m_to_k[index][j][2];
	    nn[4]=hh[map81[term]].m_to_k[index][j][3];
	    
	    if(conf2[1][1] == nn[1] &&
	       conf2[1][2] == nn[2] &&
	       conf2[1][3] == nn[3] &&
	       conf2[1][4] == nn[4])
	      are_they_neighbours=j;      
	    if (are_they_neighbours) break;
	  }
	  
	  // conf1[1] and conf2[1] need to be nearest-neighbours 
	  
	  if (are_they_neighbours) {
	    
	    counter=1;
	    for(l=1;l<=length;l++) {
	      if (l != k1 && l != k2 ) {
		tempArray[counter]=l;
		counter++;
	      }
	    }
	    
	    counter--;
	    
	    for(l=0;l<binary_max[counter];l++) {
	      
	      binary(l,counter,spin);
	      conf1_len=3;
	      conf2_len=1;
	      
	      for (j=1;j<=counter;j++) { 
		if (spin[j]) {
		  // H
		  conf1_len++;
		  temp=tempArray[j];
		  for (m=1;m<5;m++) 
		    conf1[conf1_len][m]=configuration[temp][m];
		} else {
		  // H
		  conf2_len++;
		  temp=tempArray[j];
		  for (m=1;m<5;m++) 
		    conf2[conf2_len][m]=configuration[temp][m];
		}
	      }	      
	      
	      // make sure that we don't have too many `k' sites.
	      
	      numberFlips1=0;
	      
	      for(m=1;m<=conf1_len;m++) {
		if (configuration[k1][1]==conf1[m][1] &&
		    configuration[k1][2]==conf1[m][2] &&
		    configuration[k1][3]==conf1[m][3] &&
		    configuration[k1][4]==conf1[m][4])
		  numberFlips1++;
	      }
	      maxNumberFlips1=UnitCell.s2[configuration[k1][4]];
	      
	      if (numberFlips1 <= maxNumberFlips1) {
		
		// H(i,j,j)
		descriptorIndex1=findClusterIndex(conf1_len, conf1);
		if (descriptorIndex1 <= 0) continue;
		
		// P(i)
		descriptorIndex2=findExcitedClusterIndex(conf2_len, conf2);
		if (descriptorIndex2 <= 0) continue;

		multiplier=2.0; // 2 
		
		if (Control.file) {
		  if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)) {
		    int indices[7]={0};
		    double dummy;
		    indices[0]=config_index;
		    indices[1]=term; // Jun08: origin was 1..81, now using mapped term
		    indices[2]=descriptorIndex2; // p(i)
		    indices[3]=descriptorIndex1; // H(i,j,j)
		    indices[4]=0;
		    indices[5]=0;
		    dummy=multiplier*
		      hh[map81[term]].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,3);
		  }
		}

	      }
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}

double term_exc_44(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

   /* TERM 44: 
     
     4 * P(k) . F(k) . F(m) . S(+)(k) . S(z)(m)
     
  */
  
  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int spin[MAX_LSUBM+1];
  int mm,nnn,temp1,temp2,temp3;
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int (*conf3)[5];
  int i, j, k, l, m;
  int nn[5];
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  int descriptorIndex3;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int conf3_len;
  int tempArray[MAX_LSUBM+1];
  int temp_number_conf3_sites;
  int numberFlips;
  int maxNumberFlips;
  int (*configuration)[5];
  double sum;
  sum=0.0000;

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int)); //p(i)
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int)); //f(i)
  conf3=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int)); //f(j)
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
      conf3[k][l]=0;
    }
  } 
 
  for(k=1;k<=length;k++) {
    
    for(m=1;m<5;m++) conf1[1][m]=configuration[k][m]; // P(i)
    for(m=1;m<5;m++) conf2[1][m]=configuration[k][m]; // F(i)
    
    // loop over nearest neighbours
    
    counter=1;
    for(l=1;l<=length;l++) {
      if (l != k) {
	tempArray[counter]=l;
	counter++;
      }
    }
    
    counter--;
    
    if (counter > 30) {
      printf("TERM 33: counter > 30...\nExiting!!\n\n");
      exit(1);
    }
    temp = counter; // split into summands not greater than 10 each
    temp3 = ((temp>10) ? 10 : temp); temp -= temp3;  // 0..10
    temp2 = ((temp>10) ? 10 : temp); temp -= temp2;  // 0..10
    temp1 = ((temp>10) ? 10 : temp); temp -= temp1;  // 0..10
    
    for(l=0;l<trimary_max[temp1];l++) {
      for(mm=0;mm<trimary_max[temp2];mm++) {
	for(nnn=0;nnn<trimary_max[temp3];nnn++) {
	  
	  trimary(l,mm,nnn,counter,spin);
	  conf1_len=1;
	  conf2_len=1;
	  conf3_len=0;
	  
	  for (j=1;j<=counter;j++) { 
	    temp=tempArray[j];
	    if (spin[j]==0) {         // P(k)
	      conf1_len++;
	      for (m=1;m<5;m++) 
		conf1[conf1_len][m]=configuration[temp][m];
	    } else if (spin[j]==1) {  // F(k)
	      conf2_len++;
	      for (m=1;m<5;m++) 
		conf2[conf2_len][m]=configuration[temp][m];
	    } else {                  // F(m)
	      conf3_len++;
	      for (m=1;m<5;m++) 
		conf3[conf3_len][m]=configuration[temp][m];
	    }
	  }	      
	  

          // moved that outside inner loop, Joerg Sep05, speedup
          descriptorIndex1=findExcitedClusterIndex(conf1_len, conf1);
          if (descriptorIndex1 <= 0) continue;

          descriptorIndex2=findClusterIndex(conf2_len, conf2);
          if (descriptorIndex2 <= 0) continue;
	  
	  index=conf1[1][4];      
	  for(j=1;j<=hh[map81[term]].k_to_m_number[index];j++) {
	    
	    // nearest neighbour is at x,y,z, temp put information
	    // into configuration and exit.
	    
	    /* k ---> m */
	    
	    nn[1]=conf1[1][1]+hh[map81[term]].k_to_m[index][j][0];
	    nn[2]=conf1[1][2]+hh[map81[term]].k_to_m[index][j][1];
	    nn[3]=conf1[1][3]+hh[map81[term]].k_to_m[index][j][2];
	    nn[4]=hh[map81[term]].k_to_m[index][j][3];
	    
	    numberFlips=1;
	    
	    for(m=1;m<=conf3_len;m++)
	      if (nn[1]==conf3[m][1] &&
		  nn[2]==conf3[m][2] &&
		  nn[3]==conf3[m][3] &&
		  nn[4]==conf3[m][4])
		numberFlips++;
	    
	    maxNumberFlips=UnitCell.s2[nn[4]];
	        
	    if (numberFlips <= maxNumberFlips) {
	      
	      temp_number_conf3_sites = conf3_len+1;
	      conf3[temp_number_conf3_sites][1]=nn[1];
	      conf3[temp_number_conf3_sites][2]=nn[2];
	      conf3[temp_number_conf3_sites][3]=nn[3];
	      conf3[temp_number_conf3_sites][4]=nn[4];
	      
	      descriptorIndex3=findClusterIndex(temp_number_conf3_sites, conf3);
	      if (descriptorIndex3 <= 0) continue;
	      
	      multiplier=-2*UnitCell.s2[nn[4]]; // -4 s
	      
	      if (Control.file) {
		if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0) &&
		    (descriptorIndex3 > 0)){
		  int indices[7]={0};
		  double dummy;
		  indices[0]=config_index;
		  indices[1]=term; // Jun08: origin was 1..81, now using mapped term
		  indices[2]=descriptorIndex1; //p(i)
		  indices[3]=descriptorIndex2; //f(i)
		  indices[4]=descriptorIndex3; //f(j)
		  indices[5]=0;
		  dummy=multiplier*
		    hh[map81[term]].coefficient[index][j];
                  collect_term(indices,dummy,3);
		}
	      }
		
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(conf3);
  free(configuration);
  return sum;
}

double term_exc_45(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 45: 
     
     4 * P(k) . F(m) . H(k)(m) . S(+)(k) . S(+)(m)
   
    */
  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);


  int spin[MAX_LSUBM+1];
  int mm,nnn,temp1,temp2,temp3;
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int (*conf3)[5];
  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  int descriptorIndex3;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int conf3_len;
  int tempArray[MAX_LSUBM+1];
  double sum;
  sum=0.0000;

  int (*configuration)[5];

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf3=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
      conf3[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) conf1[1][m]=configuration[k1][m]; // P(k)
	for(m=1;m<5;m++) conf2[1][m]=configuration[k2][m]; // F(m)
	for(m=1;m<5;m++) conf3[1][m]=configuration[k1][m]; // G(k,m) <--- k
	
	// loop over nearest neighbours
	
	are_they_neighbours=0;
	index=conf1[1][4];      

	for(j=1;j<=hh[map81[term]].k_to_m_number[index];j++) {
	  
	  // nearest neighbour is at x,y,z, temp put information
	  // into configuration and exit.
	  
	  /* k ---> m */
	  
	  nn[1]=conf1[1][1]+hh[map81[term]].k_to_m[index][j][0];
	  nn[2]=conf1[1][2]+hh[map81[term]].k_to_m[index][j][1];
	  nn[3]=conf1[1][3]+hh[map81[term]].k_to_m[index][j][2];
	  nn[4]=hh[map81[term]].k_to_m[index][j][3];
	  
	  if(conf2[1][1] == nn[1] &&
	     conf2[1][2] == nn[2] &&
	     conf2[1][3] == nn[3] &&
	     conf2[1][4] == nn[4])
	    are_they_neighbours=j;

	  if (are_they_neighbours) break;
	}
	
	// conf1[1] and conf2[1] need to be nearest-neighbours 
	
	if (are_they_neighbours) {
	  
	  for(m=1;m<5;m++) conf3[2][m]=configuration[k2][m]; // G(k,m) <--- m
	  
	  counter=1;
	  for(l=1;l<=length;l++) {
	    if (l != k1 && l != k2 ) {
	      tempArray[counter]=l;
		counter++;
	    }
	  }
	  
	  counter--;
	  
	  if (counter > 30) {
	    printf("TERM 32: counter > 30...\nExiting!!\n\n");
	    exit(1);
	  }
          temp = counter; // split into summands not greater than 10 each
          temp3 = ((temp>10) ? 10 : temp); temp -= temp3;  // 0..10
          temp2 = ((temp>10) ? 10 : temp); temp -= temp2;  // 0..10
          temp1 = ((temp>10) ? 10 : temp); temp -= temp1;  // 0..10
	  
	  for(l=0;l<trimary_max[temp1];l++) {
	    for(mm=0;mm<trimary_max[temp2];mm++) {
	      for(nnn=0;nnn<trimary_max[temp3];nnn++) {
		
		trimary(l,mm,nnn,counter,spin);
		conf1_len=1;
		conf2_len=1;
		conf3_len=2;
		  
		for (j=1;j<=counter;j++) { 
		  temp=tempArray[j];
		  if (spin[j]==0) {          // P(k)
		    conf1_len++;
		    for (m=1;m<5;m++) 
		      conf1[conf1_len][m]=configuration[temp][m];
		  } else if (spin[j]==1) {   // F(m)
		    conf2_len++;
		    for (m=1;m<5;m++) 
		      conf2[conf2_len][m]=configuration[temp][m];
		  } else {                    // G(k,m)
		    conf3_len++;
		    for (m=1;m<5;m++) 
		      conf3[conf3_len][m]=configuration[temp][m];
		  }
		}	      

		descriptorIndex1=findExcitedClusterIndex(conf1_len, conf1);
                if (descriptorIndex1 <= 0) continue;

		descriptorIndex2=findClusterIndex(conf2_len, conf2);
                if (descriptorIndex2 <= 0) continue;

		descriptorIndex3=findClusterIndex(conf3_len, conf3);
                if (descriptorIndex3 <= 0) continue;

		multiplier=4.0; // 4 
				  
		if (Control.file) {
		  if ((descriptorIndex1 > 0)&&(descriptorIndex2 > 0)&&
		      (descriptorIndex3 > 0)){
		    int indices[7]={0};
		    double dummy;
		    indices[0]=config_index;
		    indices[1]=term; // Jun08: origin was 1..81, now using mapped term
		    indices[2]=descriptorIndex1; // p(i)
		    indices[3]=descriptorIndex2; // f(j)
		    indices[4]=descriptorIndex3; // g(i,j)
		    indices[5]=0;
		    dummy=multiplier*
		      hh[map81[term]].coefficient[index][are_they_neighbours];
                    collect_term(indices,dummy,3);
		  }
		}
		  
	      }
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(conf3);
  free(configuration);
  return sum;
}

double term_exc_46(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 46:
     
     2 . p(i) . f(i) . g(j)(j) . S(+)(i) . S(+)(j)
     
  */

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i,j,k,l,m;
  int mm,nnn,temp1,temp2,temp3;
  int k1, k2;
  int descriptorIndex1;
  int descriptorIndex2;
  int descriptorIndex3;
  // double multiplier;
  int (*configuration)[5];
  int index;
  int nn[5];
  int maxNumberFlips;
  int numberFlips;
  int (*conf1)[5], (*conf2)[5], (*conf3)[5];
  int conf1_len;
  int conf2_len;
  int conf3_len;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
  int are_they_neighbours;

  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));
  double sum;
  sum=0.0000;

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf3=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<=MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
      conf3[k][l]=0;
    }
  } 

  for(k1=1;k1<=length;k1++) { 
    for(k2=1;k2<=length;k2++) { 
      
      if (k1 != k2) {
	
	index=configuration[k1][4];
	maxNumberFlips=UnitCell.s2[index];
	
	if (maxNumberFlips > 1) {
	  
	  for(m=1;m<5;m++) {
	    conf1[1][m]=configuration[k1][m]; // g(m,m) <-- 1st m here
	    conf1[2][m]=configuration[k1][m]; // g(m,m) <-- 2nd m here
	    conf2[1][m]=configuration[k2][m]; // P(k)
	    conf3[1][m]=configuration[k2][m]; // F(k)
	  }
	  
	  // loop over nearest neighbours
	  
	  are_they_neighbours=0;
	  index=conf1[1][4];      
	  
	  for(j=1;j<=hh[map81[term]].m_to_k_number[index];j++) {
	    
	    // nearest neighbour is at x,y,z, temp put information
	    // into configuration and exit.
	    
	    /* m ---> k */
	    
	    nn[1]=conf1[1][1]+hh[map81[term]].m_to_k[index][j][0];
	    nn[2]=conf1[1][2]+hh[map81[term]].m_to_k[index][j][1];
	    nn[3]=conf1[1][3]+hh[map81[term]].m_to_k[index][j][2];
	    nn[4]=hh[map81[term]].m_to_k[index][j][3];
	    
	    if(conf2[1][1] == nn[1] &&
	       conf2[1][2] == nn[2] &&
	       conf2[1][3] == nn[3] &&
	       conf2[1][4] == nn[4])
	      are_they_neighbours=j;      

	    if (are_they_neighbours) break;
	  }
	  
	  // conf1[1] and conf2[1] need to be nearest-neighbours 
	  
	  if (are_they_neighbours) {
	    
	    counter=1;
	    for(m=1;m<=length;m++) {
	      if (m != k1 && m != k2) {
		tempArray[counter]=m;
		counter++;
	      }
	    }
	    
	    counter--;

	    if (counter > 30) {
	      printf("TERM 45: counter > 30...\nExiting!!\n\n");
	      exit(1);
	    }
            temp = counter; // split into summands not greater than 10 each
            temp3 = ((temp>10) ? 10 : temp); temp -= temp3;  // 0..10
            temp2 = ((temp>10) ? 10 : temp); temp -= temp2;  // 0..10
            temp1 = ((temp>10) ? 10 : temp); temp -= temp1;  // 0..10
	    
	    for(l=0;l<trimary_max[temp1];l++) {
	      for(mm=0;mm<trimary_max[temp2];mm++) {
		for(nnn=0;nnn<trimary_max[temp3];nnn++) {
		  
		  trimary(l,mm,nnn,counter,spin);
		  
		  conf1_len=2;
		  conf2_len=1;
		  conf3_len=1;

		  for (j=1;j<=counter;j++) { 
		    if (spin[j]==0) {
		      // H(m,m)
		      conf1_len++;
		      temp=tempArray[j];
		      for (m=1;m<5;m++) 
			conf1[conf1_len][m]=configuration[temp][m];
		    } else if (spin[j]==1) {
		      // P(k)
		      conf2_len++;
		      temp=tempArray[j];
		      for (m=1;m<5;m++) 
			conf2[conf2_len][m]=configuration[temp][m];
		    } else {
		      // F(k)
		      conf3_len++;
		      temp=tempArray[j];
		      for (m=1;m<5;m++) 
			conf3[conf3_len][m]=configuration[temp][m];
		    }
		  }
	      
		  numberFlips=0;
	      
		  for(m=1;m<=conf1_len;m++) // k1 is "m" (or "j")
		    if (configuration[k1][1]==conf1[m][1] &&
			configuration[k1][2]==conf1[m][2] &&
			configuration[k1][3]==conf1[m][3] &&
			configuration[k1][4]==conf1[m][4])
		      numberFlips++;
	      
		  maxNumberFlips=UnitCell.s2[configuration[k1][4]];
	      
		  // the number of spin flips that you can have at a given
		  // lattice site is depedent on the quantum spin number
		  // of the site in the UNIT CELL index `length'
	      
		  if (numberFlips <= maxNumberFlips) {
		
		    // H(m,m)
		    descriptorIndex1=findClusterIndex(conf1_len, conf1);
		    if (descriptorIndex1 <= 0) continue;
		    
		    // P(k)
		    descriptorIndex2=findExcitedClusterIndex(conf2_len, conf2);
	            if (descriptorIndex2 <= 0) continue;
		
		    // F(k)
		    descriptorIndex3=findClusterIndex(conf3_len, conf3);
		    if (descriptorIndex3 <= 0) continue;
		
		    if (Control.file) {
		      if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0) &&
			  (descriptorIndex3 > 0)) {
			int indices[7]={0};
			double dummy;
			indices[0]=config_index;
			indices[1]=term; // Jun08: origin was 1..81, now using mapped term
			indices[2]=descriptorIndex2; // p(i)
			indices[3]=descriptorIndex1; // g(j,j)
			indices[4]=descriptorIndex3; // f(i)
			indices[5]=0;
			dummy=2*hh[map81[term]].coefficient[index][are_they_neighbours]; // factor from h is 2 here.
            collect_term(indices,dummy,3);
		      }
		    }

		  }
		}
	      }
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(conf3);
  free(configuration);
  return sum;
}  

double term_exc_47(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 47:
     
     2 * P(k) . F(k) . F(m) . F(m) . S(+)(k) . S(+)(m)
   
    */

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (length < 2) return 0;
  if (Control.output > 10) printf("Term=%d\n",term);

 
  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int (*conf3)[5];
  int (*conf4)[5];
  int i, j, k, l, m, n, p, mm;
  int temp1, temp2, temp3;
  int k1;
  int k2;
  int are_they_neighbours;
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  int descriptorIndex3;
  int descriptorIndex4;
//  double multiplier;
  int conf1_num_sites;
  int conf2_num_sites;
  int conf3_num_sites;
  int conf4_num_sites;
  int tempArray[MAX_LSUBM+1];

  int (*configuration)[5];
  double sum;
  sum=0.0000;
  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf3=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf4=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<=MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
      conf3[k][l]=0;
      conf4[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
        // ToDo: within C++ we dont need the loop-m conf1[1]=config[k1]
	for(m=1;m<5;m++) conf1[1][m]=configuration[k1][m]; // P(k)
	for(m=1;m<5;m++) conf2[1][m]=configuration[k1][m]; // F(k)
	for(m=1;m<5;m++) conf3[1][m]=configuration[k2][m]; // F(m)
	for(m=1;m<5;m++) conf4[1][m]=configuration[k2][m]; // F(m)
	
        index=conf1[1][4];
        are_they_neighbours= test_k_to_m_neighbour(hh,term,conf1[1],conf3[1]);
	// conf1[1] and conf3[1] need to be nearest-neighbours 
	
	if (are_they_neighbours) {
	  
	  counter=1;
	  for(l=1;l<=length;l++) {
	    if (l != k1 && l != k2 ) {
	      tempArray[counter]=l;
	      counter++;
	    }
	  }
	  
	  counter--;
	  
    temp = counter; // split into summands not greater than 10 each
    temp3 = ((temp>10) ? 10 : temp); temp -= temp3;  // 0..10
    temp2 = ((temp>10) ? 10 : temp); temp -= temp2;  // 0..10
    temp1 = ((temp>10) ? 10 : temp); temp -= temp1;  // 0..10
	  
	  for(l=0;l<quartary_max[temp1];l++) {
	    for(m=0;m<quartary_max[temp2];m++) {
	      for(n=0;n<quartary_max[temp3];n++) {
		
		quartary(l,m,n,counter,spin);
		conf1_num_sites=1;
		conf2_num_sites=1;
		conf3_num_sites=1;
		conf4_num_sites=1;
		
		for (p=1;p<=counter;p++) { 
		  if (spin[p]==0) {
		    // F(k)
		    conf1_num_sites++;
		    temp=tempArray[p];
		    for (mm=1;mm<5;mm++) 
		      conf1[conf1_num_sites][mm]=
			configuration[temp][mm];
		  } else if (spin[p]==1) {
		    // F(k)
		    conf2_num_sites++;
		    temp=tempArray[p];
		      for (mm=1;mm<5;mm++) 
			conf2[conf2_num_sites][mm]=
			  configuration[temp][mm];
		  } else if (spin[p]==2) {
		    // F(m)
		    conf3_num_sites++;
		      temp=tempArray[p];
		      for (mm=1;mm<5;mm++) 
			conf3[conf3_num_sites][mm]=
			  configuration[temp][mm];
		  } else {
		    // F(m)
		    conf4_num_sites++;
		    temp=tempArray[p];
		      for (mm=1;mm<5;mm++) 
			conf4[conf4_num_sites][mm]=
			  configuration[temp][mm];
		  }
		}	      
		
		descriptorIndex1=findExcitedClusterIndex(conf1_num_sites, conf1);
                if (descriptorIndex1 <= 0) continue;
		
		descriptorIndex2=findClusterIndex(conf2_num_sites, conf2);
                if (descriptorIndex2 <= 0) continue;
		
		descriptorIndex3=findClusterIndex(conf3_num_sites, conf3);
                if (descriptorIndex3 <= 0) continue;
		
		descriptorIndex4=findClusterIndex(conf4_num_sites, conf4);
                if (descriptorIndex4 <= 0) continue;
		
		if (Control.file) {
                  {
		    static int indices[7]={0};
		    double dummy;
		    indices[0]=config_index;  // original config
		    indices[1]=term; // Jun08: origin was 1..81, now using mapped term
		    indices[2]=descriptorIndex1; // modified config1..4
		    indices[3]=descriptorIndex2;
		    indices[4]=descriptorIndex3;
		    indices[5]=descriptorIndex4;
		    dummy= 
		     2*hh[map81[term]].coefficient[index][are_they_neighbours];
		    collect_term(indices,dummy,3); // factor from H is 2  
		}
		}
		
	    }
	    }
	    }

	}
	}
	}
	}


		

  free(conf1);
  free(conf2);
  free(conf3);
  free(conf4);
  free(configuration);
  return sum;
}


double term_exc_48(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

   /* TERM 48: 
     
     2 * P(k) . F(m) . F(m) . S(+)(m) . S(z)(j)
     
  */
  
  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int spin[MAX_LSUBM+1];
  int mm,nnn,temp1,temp2,temp3;
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int (*conf3)[5];
  int i, j, k, l, m;
  int nn[5];
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  int descriptorIndex3;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int conf3_len;
  int tempArray[MAX_LSUBM+1];
  int temp_number_conf3_sites;
//  int temp_number_conf2_sites;
//  int numberFlips1;
  int maxNumberFlips;
  int numberFlips2;
  int (*configuration)[5];
  double sum;
  sum=0.0000;

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int)); //f(j)
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int)); //f(j)
  conf3=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int)); //p(i)
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
      conf3[k][l]=0;
    }
  } 
 
  for(k=1;k<=length;k++) {
    
    for(m=1;m<5;m++) conf1[1][m]=configuration[k][m]; // f(j)
    for(m=1;m<5;m++) conf2[1][m]=configuration[k][m]; // f(j)
    
    // loop over nearest neighbours
    
    counter=1;
    for(l=1;l<=length;l++) {
      if (l != k) {
	tempArray[counter]=l;
	counter++;
      }
    }
    
    counter--;
    
    if (counter > 30) {
      printf("TERM 33: counter > 30...\nExiting!!\n\n");
      exit(1);
    }
    temp = counter; // split into summands not greater than 10 each
    temp3 = ((temp>10) ? 10 : temp); temp -= temp3;  // 0..10
    temp2 = ((temp>10) ? 10 : temp); temp -= temp2;  // 0..10
    temp1 = ((temp>10) ? 10 : temp); temp -= temp1;  // 0..10
    
    for(l=0;l<trimary_max[temp1];l++) {
      for(mm=0;mm<trimary_max[temp2];mm++) {
	for(nnn=0;nnn<trimary_max[temp3];nnn++) {
	  
	  trimary(l,mm,nnn,counter,spin);
	  conf1_len=1;
	  conf2_len=1;
	  conf3_len=0;
	  
	  for (j=1;j<=counter;j++) { 
	    temp=tempArray[j];
	    if (spin[j]==0) {         // f(j)
	      conf1_len++;
	      for (m=1;m<5;m++) 
		conf1[conf1_len][m]=configuration[temp][m];
	    } else if (spin[j]==1) {  // f(j)
	      conf2_len++;
	      for (m=1;m<5;m++) 
		conf2[conf2_len][m]=configuration[temp][m];
	    } else {                  // p(i)
	      conf3_len++;
	      for (m=1;m<5;m++) 
		conf3[conf3_len][m]=configuration[temp][m];
	    }
	  }	      
	  

          // moved that outside inner loop, Joerg Sep05, speedup
          descriptorIndex1=findClusterIndex(conf1_len, conf1);
          if (descriptorIndex1 <= 0) continue;
	  
          descriptorIndex2=findClusterIndex(conf2_len, conf2);
          if (descriptorIndex2 <= 0) continue;

	  index=conf1[1][4];      
	  for(j=1;j<=hh[map81[term]].m_to_k_number[index];j++) {
	    
	    // nearest neighbour is at x,y,z, temp put information
	    // into configuration and exit.
	    
	    /* m ---> k */
	    
	    nn[1]=conf1[1][1]+hh[map81[term]].m_to_k[index][j][0];
	    nn[2]=conf1[1][2]+hh[map81[term]].m_to_k[index][j][1];
	    nn[3]=conf1[1][3]+hh[map81[term]].m_to_k[index][j][2];
	    nn[4]=hh[map81[term]].m_to_k[index][j][3];
	        
	    maxNumberFlips=UnitCell.s2[nn[4]];
		
	    numberFlips2=1;
	    
	    for(m=1;m<=conf3_len;m++)
	      if (nn[1]==conf3[m][1] &&
		  nn[2]==conf3[m][2] &&
		  nn[3]==conf3[m][3] &&
		  nn[4]==conf3[m][4])
		numberFlips2++;
	        
	    // the number of spin flips that you can have at a given
	    // lattice site is depedent on the quantum spin number
	    // of the site in the UNIT CELL index `length'
	    
	    if (numberFlips2 <= maxNumberFlips) {
	      
	      temp_number_conf3_sites = conf3_len+1;
	      conf3[temp_number_conf3_sites][1]=nn[1];
	      conf3[temp_number_conf3_sites][2]=nn[2];
	      conf3[temp_number_conf3_sites][3]=nn[3];
	      conf3[temp_number_conf3_sites][4]=nn[4];
	      
	      descriptorIndex3=findExcitedClusterIndex(temp_number_conf3_sites, conf3);
	      if (descriptorIndex3 <= 0) continue;
		        
	      multiplier=-1*UnitCell.s2[nn[4]]; // -2s from Hamiltonian
	      
	      
	      if (Control.file) {
		if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0) &&
		    (descriptorIndex3 > 0)){
		  int indices[7]={0};
		  double dummy;
		  indices[0]=config_index;
		  indices[1]=term; // Jun08: origin was 1..81, now using mapped term
		  indices[2]=descriptorIndex3; //p(i)
		  indices[3]=descriptorIndex2; //f(j)
		  indices[4]=descriptorIndex1; //f(j)
		  indices[5]=0;
		  dummy=multiplier*
		    hh[map81[term]].coefficient[index][j];
                  collect_term(indices,dummy,3);
		}
	      }
		
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(conf3);
  free(configuration);
  return sum;
}

double term_exc_49(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 49:
     
     2 * Q(k)(k) . F(j) . S(+)(i) . S(z)(j)
     
  */

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i,j,k,l,m;
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int (*configuration)[5];
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int maxNumberFlips1;
  int maxNumberFlips2;
  int numberFlips1;
  int numberFlips2;
  int (*conf1)[5], (*conf2)[5];
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
  int temp_number_conf2_sites;
  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));
  double sum;
  sum=0.0000;

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<=MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 

  for(k=1;k<=length;k++) { // `m' here. 
    
    index=configuration[k][4];
    maxNumberFlips=UnitCell.s2[index];
    
    if (maxNumberFlips > 1) {
      
      for(m=1;m<5;m++) {
	conf1[1][m]=configuration[k][m]; // Q(k,k) -- 1st k 
	conf1[2][m]=configuration[k][m]; // Q(k,k) -- 2nd k
      }
      
      counter=1;
      for(m=1;m<=length;m++) {
	if (m != k) {
	  tempArray[counter]=m;
	  counter++;
	}
      }
      
      counter--;
      
      for(l=0;l<binary_max[counter];l++) {
	
	binary(l,counter,spin);
	conf1_len=2;
	conf2_len=0;
	
	for (j=1;j<=counter;j++) { 
	  if (spin[j]) {
	    // Q(k,k)
	    conf1_len++;
	    temp=tempArray[j];
	    for (m=1;m<5;m++) 
	      conf1[conf1_len][m]=configuration[temp][m];
	  } else {
	    // F(m)
	    conf2_len++;
	    temp=tempArray[j];
	    for (m=1;m<5;m++) 
	      conf2[conf2_len][m]=configuration[temp][m];
	  }
	}
	
	index=conf1[1][4];
	
	for(neighbour=1;
	    neighbour<=hh[map81[term]].k_to_m_number[index];
	    neighbour++) {
	  
	  // nearest neighbour is at x,y,z, cell put information
	  // into configuration and exit.
	  
	  /* k ---> m */
	  
	  nn[1]=conf1[1][1]+
	    hh[map81[term]].k_to_m[index][neighbour][0];
	  nn[2]=conf1[1][2]+
	    hh[map81[term]].k_to_m[index][neighbour][1];
	  nn[3]=conf1[1][3]+
	    hh[map81[term]].k_to_m[index][neighbour][2];
	  nn[4]=hh[map81[term]].k_to_m[index][neighbour][3];
	  
	  numberFlips1=1;
	
	  for(m=1;m<=conf2_len;m++)
	    if (nn[1]==conf2[m][1] &&
		nn[2]==conf2[m][2] &&
		nn[3]==conf2[m][3] &&
		nn[4]==conf2[m][4])
	      numberFlips1++;
	
	  maxNumberFlips1=UnitCell.s2[nn[4]];
	
	  numberFlips2=0;
	  
	  for(m=1;m<=conf1_len;m++)
	    if (configuration[k][1]==conf1[m][1] &&
		configuration[k][2]==conf1[m][2] &&
		configuration[k][3]==conf1[m][3] &&
		configuration[k][4]==conf1[m][4])
	      numberFlips2++;
	
	  maxNumberFlips2=UnitCell.s2[configuration[k][4]];

	  // the number of spin flips that you can have at a given
	  // lattice site is depedent on the quantum spin number
	  // of the site in the UNIT CELL index `length'
	
	  if (numberFlips1 <= maxNumberFlips1 &&
	      numberFlips2 <= maxNumberFlips2) {
	    
	    temp_number_conf2_sites=conf2_len+1;
	  
	    conf2[temp_number_conf2_sites][1]=nn[1];
	    conf2[temp_number_conf2_sites][2]=nn[2];
	    conf2[temp_number_conf2_sites][3]=nn[3];
	    conf2[temp_number_conf2_sites][4]=nn[4];
	  
	    // Q(k,k)
	    descriptorIndex1=findExcitedClusterIndex(conf1_len, conf1);
	    if (descriptorIndex1 <= 0) continue;
	    
	    // F(m)
	    descriptorIndex2=findClusterIndex(temp_number_conf2_sites, conf2);
	    if (descriptorIndex2 <= 0) continue;

	    multiplier=-UnitCell.s2[nn[4]]; // factor of -2s from H.

	    if (Control.file) {
	      if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)) {
		int indices[7]={0};
		double dummy;
		indices[0]=config_index;
		indices[1]=term; // Jun08: origin was 1..81, now using mapped term
		indices[2]=descriptorIndex1; // Q(i,i)
		indices[3]=descriptorIndex2; // f(j)
		indices[4]=0;
		indices[5]=0;
		dummy=multiplier*
		hh[map81[term]].coefficient[index][neighbour];
            collect_term(indices,dummy,3);
	      }
	    }

	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}  

double term_exc_50(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 50: 
     
     Q(k)(k) . H(m)(m) . S(+)(k) . S(+)(m)
     
    */
  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);


  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  // double multiplier;
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int maxNumberFlips1;
  int maxNumberFlips2;
  int numberFlips1;
  int numberFlips2;
  double sum;
  sum=0.0000;

  int (*configuration)[5];

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {

	maxNumberFlips1=UnitCell.s2[configuration[k1][4]];
	maxNumberFlips2=UnitCell.s2[configuration[k2][4]];
	
	if (maxNumberFlips1 > 1 && maxNumberFlips2 > 1) {
	  
	  for(m=1;m<5;m++) {
	    conf1[1][m]=configuration[k1][m]; // Q(k,k) k--1st one
	    conf1[2][m]=configuration[k1][m]; // Q(k,k) k--2nd one
	    conf2[1][m]=configuration[k2][m]; // H(m,m) m--1st one
	    conf2[2][m]=configuration[k2][m]; // H(m,m) m--2nd one
	  }
	  
	  // loop over nearest neighbours
	  
	  are_they_neighbours=0;
	  index=conf1[1][4];      
	  
	  for(j=1;j<=hh[map81[term]].k_to_m_number[index];j++) {
	    
	    // nearest neighbour is at x,y,z, temp put information
	    // into configuration and exit.
	    
	    /* k ---> m */
	    
	    nn[1]=conf1[1][1]+hh[map81[term]].k_to_m[index][j][0];
	    nn[2]=conf1[1][2]+hh[map81[term]].k_to_m[index][j][1];
	    nn[3]=conf1[1][3]+hh[map81[term]].k_to_m[index][j][2];
	    nn[4]=hh[map81[term]].k_to_m[index][j][3];
	    
	    if(conf2[1][1] == nn[1] &&
	       conf2[1][2] == nn[2] &&
	       conf2[1][3] == nn[3] &&
	       conf2[1][4] == nn[4])
	      are_they_neighbours=j;
      
	    if (are_they_neighbours) break;
	  }
	  
	  // conf1[1] and conf2[1] need to be nearest-neighbours 
	  
	  if (are_they_neighbours) {
	    
	    counter=1;
	    for(l=1;l<=length;l++) {
	      if (l != k1 && l != k2 ) {
		tempArray[counter]=l;
		counter++;
	      }
	    }
	    
	    counter--;
	    
	    for(l=0;l<binary_max[counter];l++) {
	      
	      binary(l,counter,spin);
	      conf1_len=2;
	      conf2_len=2;
	      
	      for (j=1;j<=counter;j++) { 
		if (spin[j]) {
		  // H
		  conf1_len++;
		  temp=tempArray[j];
		  for (m=1;m<5;m++) 
		    conf1[conf1_len][m]=configuration[temp][m];
		} else {
		  // H
		  conf2_len++;
		  temp=tempArray[j];
		  for (m=1;m<5;m++) 
		    conf2[conf2_len][m]=configuration[temp][m];
		}
	      }	      
	      
	      // make sure that we don't have too many `k' sites.
	      
	      numberFlips1=0;
	      
	      for(m=1;m<=conf1_len;m++) {
		if (configuration[k1][1]==conf1[m][1] &&
		    configuration[k1][2]==conf1[m][2] &&
		    configuration[k1][3]==conf1[m][3] &&
		    configuration[k1][4]==conf1[m][4])
		  numberFlips1++;
	      }
	      maxNumberFlips1=UnitCell.s2[configuration[k1][4]];

	      numberFlips2=0;
	      
	      for(m=1;m<=conf2_len;m++) {
		if (configuration[k2][1]==conf2[m][1] &&
		    configuration[k2][2]==conf2[m][2] &&
		    configuration[k2][3]==conf2[m][3] &&
		    configuration[k2][4]==conf2[m][4])
		  numberFlips2++;
	      }
	      maxNumberFlips2=UnitCell.s2[configuration[k2][4]];
	      
	      if (numberFlips1 <= maxNumberFlips1 &&
		  numberFlips2 <= maxNumberFlips2) {
		
		descriptorIndex1=findExcitedClusterIndex(conf1_len, conf1); //Q(ii)
		if (descriptorIndex1 <= 0) continue;
		
		descriptorIndex2=findClusterIndex(conf2_len, conf2); //G(jj)
		if (descriptorIndex2 <= 0) continue;

		if (Control.file) {
		  if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)) {
		    int indices[7]={0};
		    double dummy;
		    indices[0]=config_index;
		    indices[1]=term; // Jun08: origin was 1..81, now using mapped term
		    indices[2]=descriptorIndex1;
		    indices[3]=descriptorIndex2;
		    indices[4]=0;
		    indices[5]=0;
		    dummy=
		      hh[map81[term]].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,3); // multiplier from H is 1.
		  }
		}

	      }
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}

double term_exc_51(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 51:

     Q(k)(k) . F(m) . F(m) . S(+)(k) . S(+)(m)

  */

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i,j,k,l,m;
  int mm,nnn,temp1,temp2,temp3;
  int k1, k2;
  int descriptorIndex1;
  int descriptorIndex2;
  int descriptorIndex3;
  // double multiplier;
  int (*configuration)[5];
  int index;
  int nn[5];
  int maxNumberFlips;
  int numberFlips;
  int (*conf1)[5], (*conf2)[5], (*conf3)[5];
  int conf1_len;
  int conf2_len;
  int conf3_len;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
  int are_they_neighbours;

  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));
  double sum;
  sum=0.0000;

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf3=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<=MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
      conf3[k][l]=0;
    }
  } 
  
  for(k1=1;k1<=length;k1++) { 
    for(k2=1;k2<=length;k2++) { 
      
      if (k1 != k2) {
	
	index=configuration[k1][4];
	maxNumberFlips=UnitCell.s2[index];
	
	if (maxNumberFlips > 1) {
	  
	  for(m=1;m<5;m++) {
	    conf1[1][m]=configuration[k1][m]; // Q(k,k) <-- 1st k here
	    conf1[2][m]=configuration[k1][m]; // Q(k,k) <-- 2nd k here
	    conf2[1][m]=configuration[k2][m]; // F(m)
	    conf3[1][m]=configuration[k2][m]; // F(m)
	  }
	  
	  // loop over nearest neighbours
	  
	  are_they_neighbours=0;
	  index=conf1[1][4];      
	  
	  for(j=1;j<=hh[map81[term]].k_to_m_number[index];j++) {
	    
	    // nearest neighbour is at x,y,z, temp put information
	    // into configuration and exit.
	    
	    /* k ---> m */
	    
	    nn[1]=conf1[1][1]+hh[map81[term]].k_to_m[index][j][0];
	    nn[2]=conf1[1][2]+hh[map81[term]].k_to_m[index][j][1];
	    nn[3]=conf1[1][3]+hh[map81[term]].k_to_m[index][j][2];
	    nn[4]=hh[map81[term]].k_to_m[index][j][3];
	    
	    if(conf2[1][1] == nn[1] &&
	       conf2[1][2] == nn[2] &&
	       conf2[1][3] == nn[3] &&
	       conf2[1][4] == nn[4])
	      are_they_neighbours=j;      

	    if (are_they_neighbours) break;
	  }
	  
	  // conf1[1] and conf2[1] need to be nearest-neighbours 
	  
	  if (are_they_neighbours) {
	    
	    counter=1;
	    for(m=1;m<=length;m++) {
	      if (m != k1 && m != k2) {
		tempArray[counter]=m;
		counter++;
	      }
	    }
	    
	    counter--;
	    
	    if (counter > 30) {
	      printf("TERM 44: counter > 30...\nExiting!!\n\n");
	      exit(1);
	    }
            temp = counter; // split into summands not greater than 10 each
            temp3 = ((temp>10) ? 10 : temp); temp -= temp3;  // 0..10
            temp2 = ((temp>10) ? 10 : temp); temp -= temp2;  // 0..10
            temp1 = ((temp>10) ? 10 : temp); temp -= temp1;  // 0..10
	      
	    for(l=0;l<trimary_max[temp1];l++) {
	      for(mm=0;mm<trimary_max[temp2];mm++) {
		for(nnn=0;nnn<trimary_max[temp3];nnn++) {
		    
		  trimary(l,mm,nnn,counter,spin);
		    
		  conf1_len=2;
		  conf2_len=1;
		  conf3_len=1;
		    
		  for (j=1;j<=counter;j++) { 
		    if (spin[j]==0) {
		      // Q(k,k)
		      conf1_len++;
		      temp=tempArray[j];
		      for (m=1;m<5;m++) 
			conf1[conf1_len][m]=configuration[temp][m];
		    } else if (spin[j]==1) {
		      // F(m)
		      conf2_len++;
		      temp=tempArray[j];
		      for (m=1;m<5;m++) 
			conf2[conf2_len][m]=configuration[temp][m];
		    } else {
		      // F(m)
		      conf3_len++;
		      temp=tempArray[j];
		      for (m=1;m<5;m++) 
			conf3[conf3_len][m]=configuration[temp][m];
		    }
		  }
	      
		  numberFlips=0;
	      
		  for(m=1;m<=conf1_len;m++)
		    if (configuration[k1][1]==conf1[m][1] &&
			configuration[k1][2]==conf1[m][2] &&
			configuration[k1][3]==conf1[m][3] &&
			configuration[k1][4]==conf1[m][4])
		      numberFlips++;
	      
		  maxNumberFlips=UnitCell.s2[configuration[k1][4]];
	      
		  // the number of spin flips that you can have at a given
		  // lattice site is depedent on the quantum spin number
		  // of the site in the UNIT CELL index `length'
	      
		  if (numberFlips <= maxNumberFlips) {
		
		    // Q(k,k)
		    descriptorIndex1=findExcitedClusterIndex(conf1_len, conf1);
		    if (descriptorIndex1 <= 0) continue;
		    
		    // F(m)
		    descriptorIndex2=findClusterIndex(conf2_len, conf2);
	            if (descriptorIndex2 <= 0) continue;
		    
		    // F(m)
		    descriptorIndex3=findClusterIndex(conf3_len, conf3);
		    if (descriptorIndex3 <= 0) continue;
	  
		    if (Control.file) {
		      if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0) &&
			  (descriptorIndex3 > 0)) {
			int indices[7]={0};
			double dummy;
			indices[0]=config_index;
			indices[1]=term; // Jun08: origin was 1..81, now using mapped term
			indices[2]=descriptorIndex1;
			indices[3]=descriptorIndex2;
			indices[4]=descriptorIndex3;
			indices[5]=0;
			dummy=
			  hh[map81[term]].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,3); // factor from H is 1.
		      }
		    }

		  }
		}
	      }
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(conf3);
  free(configuration);
  return sum;
}  

double term_exc_52(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 52: 
     
     4 * Q(k)(m) . S(z)(k) . S(z)(m)

     one of the most time-consuming terms because you've got to sweep
     through the index `k'

     */

  int i, j, l;
  int descriptorIndex;
//  double multiplier;
  double temp_multiplier;
  double sum;
  sum=0.0000;

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int minX, minY, minZ;
  int maxX, maxY, maxZ;

  int x,y,z,cell;
  int index;

  int numberFlips, maxNumberFlips;

  int (*configuration)[5];

  // term27 generates clusters of length+1+1, so we can skip some work
  // for length+1+1 > Params.n we always have an invalid index
  if (length+1+1 > Params.n) return 0;

  configuration=(int (*)[5])malloc((MAX_LSUBM+4)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  int maxPointX;
  int maxPointY;
  int maxPointZ;

  maxPointX=0;
  maxPointY=0;
  maxPointZ=0;
  for(j=1;j<=length;j++) {
    if (maxPointX < abs(string[j][1])) maxPointX=abs(string[j][1]);
    if (maxPointY < abs(string[j][2])) maxPointY=abs(string[j][2]);
    if (maxPointZ < abs(string[j][3])) maxPointZ=abs(string[j][3]);
  }

  if (Params.dim == 1) {
    minX=radiusExc;           minY=0;                minZ=0;
    maxX=radiusExc+maxPointX; maxY=0;                maxZ=0;
  } else if (Params.dim == 2) {
    minX=radiusExc;           minY=radiusExc;           minZ=0;
    maxX=radiusExc+maxPointX; maxY=radiusExc+maxPointY; maxZ=0;
  } else if (Params.dim == 3) {
    minX=radiusExc;           minY=radiusExc;           minZ=radiusExc;
    maxX=radiusExc+maxPointX; maxY=radiusExc+maxPointY; maxZ=radiusExc+maxPointZ;
  } else {
    printf("Really weird error. Dimension (%d)"
     " not recognised...\nExiting||\n\n",Params.dim);
    exit(1);
  }

  for(x=-minX;x<=maxX;x++) { // ToDo: better (faster) x = maxX-m ... m ???
    for(y=-minY;y<=maxY;y++) {
      for(z=-minZ;z<=maxZ;z++) {
	for(cell=1;cell<=UnitCell.number;cell++) {

          // Joerg ToDo: lot of x,y,z pairs lying outside the bounding triangle
          //      and scipt if descriptor not exist, can be done earlier?
          //      same for term30
	  
	  configuration[length+1][1]=x;
	  configuration[length+1][2]=y;
	  configuration[length+1][3]=z;
	  configuration[length+1][4]=cell;
	  
	  index=cell;
	  numberFlips=1;
	  
	  for(l=1;l<=length;l++)
	    if (configuration[length+1][1]==configuration[l][1] &&
		configuration[length+1][2]==configuration[l][2] &&
		configuration[length+1][3]==configuration[l][3] &&
		configuration[length+1][4]==configuration[l][4])
	      numberFlips++;
	  
	  maxNumberFlips=UnitCell.s2[configuration[length+1][4]];
	  
	  // the number of spin flips that you can have at a given
	  // lattice site is depedent on the quantum spin number
	  // of the site in the UNIT CELL index `length'
	  
	  if (numberFlips <= maxNumberFlips) {
	    
	    for(j=1;j<=hh[map81[term]].k_to_m_number[index];j++) {
	      
	      // nearest neighbour is at x,y,z, temp put information
	      // into configuration and exit.
	      
	      /* k --> m */
	      
	      configuration[length+1+1][1]=configuration[length+1][1]
		+hh[map81[term]].k_to_m[index][j][0];
	      configuration[length+1+1][2]=configuration[length+1][2]
		+hh[map81[term]].k_to_m[index][j][1];
	      configuration[length+1+1][3]=configuration[length+1][3]
		+hh[map81[term]].k_to_m[index][j][2];
	      configuration[length+1+1][4]=
		hh[map81[term]].k_to_m[index][j][3];
	      
	      numberFlips=1;
	      
	      for(l=1;l<=length+1;l++)
		if (configuration[length+1+1][1]==configuration[l][1] &&
		    configuration[length+1+1][2]==configuration[l][2] &&
		    configuration[length+1+1][3]==configuration[l][3] &&
		    configuration[length+1+1][4]==configuration[l][4])
		  numberFlips++;
	      
	      maxNumberFlips=UnitCell.s2[configuration[length+1+1][4]];
	      
	      // the number of spin flips that you can have at a given
	      // lattice site is depedent on the quantum spin number
	      // of the site in the UNIT CELL index `length'
	      
	      if (numberFlips <= maxNumberFlips) {
		
		temp_multiplier=UnitCell.s2[configuration[length+1][4]]*
		  UnitCell.s2[configuration[length+1+1][4]]; // 4 s*s from Hamiltonian
		
		descriptorIndex=findExcitedClusterIndex((length+1+1), configuration);
		
		if (Control.file) {
		  if (descriptorIndex > 0){
		    int indices[7]={0};
		    double dummy;
		    indices[0]=config_index;
		    indices[1]=term; // Jun08: origin was 1..81, now using mapped term
		    indices[2]=descriptorIndex; // q(i,j)
		    indices[3]=0;
		    indices[4]=0;
		    indices[5]=0;
		    dummy=temp_multiplier*
		      hh[map81[term]].coefficient[index][j];
            collect_term(indices,dummy,3);
		  }
		}
		
	      }
	    }
	  }
	}
      }
    }
  }
  free(configuration);
  return sum;
}

double term_exc_53(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 53:
     
     4 * Q(k)(m) . F(m) . S(+)(m) . S(z)(k)
     
  */

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i,j,k,l,m;
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int (*configuration)[5];
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int numberFlips;
  int (*conf1)[5], (*conf2)[5];
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
  int temp_number_conf2_sites;
  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));
  double sum;

  sum=0.0000;

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<=MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  }

  for(k=1;k<=length;k++) {
        
    for(m=1;m<5;m++) {
      conf1[1][m]=configuration[k][m]; // F(m)
      conf2[1][m]=configuration[k][m]; // Q(m,k) <-- m here.
    }
      
    counter=1;
    for(m=1;m<=length;m++) {
      if (m != k) {
	tempArray[counter]=m;
	counter++;
      }
    }    
    counter--;
    
    for(l=0;l<binary_max[counter];l++) {
      
      binary(l,counter,spin);
      conf1_len=1;
      conf2_len=1;
      
      for (j=1;j<=counter;j++) { 
	if (spin[j]) {
	  // F(m)
	  conf1_len++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf1[conf1_len][m]=configuration[temp][m];
	} else {
	  // Q(k,m)
	  conf2_len++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf2[conf2_len][m]=configuration[temp][m];
	}
      }
      
      index=conf2[1][4];
 
      for(neighbour=1;neighbour<=hh[map81[term]].m_to_k_number[index];
	  neighbour++) {
	
	// nearest neighbour is at x,y,z, cell put information
	// into configuration and exit.
	
	/* m ---> k */
	
	nn[1]=conf2[1][1]+hh[map81[term]].m_to_k[index][neighbour][0];
	nn[2]=conf2[1][2]+hh[map81[term]].m_to_k[index][neighbour][1];
	nn[3]=conf2[1][3]+hh[map81[term]].m_to_k[index][neighbour][2];
	nn[4]=hh[map81[term]].m_to_k[index][neighbour][3];
	
	numberFlips=1;
	
	for(m=1;m<=conf2_len;m++)
	  if (nn[1]==conf2[m][1] &&
	      nn[2]==conf2[m][2] &&
	      nn[3]==conf2[m][3] &&
	      nn[4]==conf2[m][4])
	    numberFlips++;
	
	maxNumberFlips=UnitCell.s2[nn[4]];
	
	// the number of spin flips that you can have at a given
	// lattice site is depedent on the quantum spin number
	// of the site in the UNIT CELL index `length'
	
	if (numberFlips <= maxNumberFlips) {
	  
	  temp_number_conf2_sites=conf2_len+1;
	  
	  conf2[temp_number_conf2_sites][1]=nn[1];
	  conf2[temp_number_conf2_sites][2]=nn[2];
	  conf2[temp_number_conf2_sites][3]=nn[3];
	  conf2[temp_number_conf2_sites][4]=nn[4];
	  
	  descriptorIndex1=findClusterIndex(conf1_len, conf1);
	  if (descriptorIndex1 <= 0) continue;
	 
	  descriptorIndex2=findExcitedClusterIndex(temp_number_conf2_sites, conf2);
	  if (descriptorIndex2 <= 0) continue;
					  
	  multiplier=-2*UnitCell.s2[nn[4]]; // -4 * s factor from H. 
	  
	  if (Control.file) {
	    if ((descriptorIndex1 > 0)&&(descriptorIndex2 > 0)){
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=term; // Jun08: origin was 1..81, now using mapped term
	      indices[2]=descriptorIndex2; // Q(i,j)
	      indices[3]=descriptorIndex1; // F(j)
	      indices[4]=0;
	      indices[5]=0;
	      dummy=multiplier*
		hh[map81[term]].coefficient[index][neighbour];
            collect_term(indices,dummy,3);
	    }
	  }
	  
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}  

double term_exc_54(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 54:
     
     4 * Q(i,j) . F(i) . S(+)(i) . S(z)(j)
     
  */

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i,j,k,l,m;
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int (*configuration)[5];
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int numberFlips;
  int (*conf1)[5], (*conf2)[5];
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
  int temp_number_conf2_sites;
  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));
  double sum;

  sum=0.0000;

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<=MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  }

  for(k=1;k<=length;k++) {
        
    for(m=1;m<5;m++) {
      conf1[1][m]=configuration[k][m]; // F(i)
      conf2[1][m]=configuration[k][m]; // Q(i,j) <-- i here.
    }
      
    counter=1;
    for(m=1;m<=length;m++) {
      if (m != k) {
	tempArray[counter]=m;
	counter++;
      }
    }    
    counter--;
    
    for(l=0;l<binary_max[counter];l++) {
      
      binary(l,counter,spin);
      conf1_len=1;
      conf2_len=1;
      
      for (j=1;j<=counter;j++) { 
	if (spin[j]) {
	  // F(i)
	  conf1_len++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf1[conf1_len][m]=configuration[temp][m];
	} else {
	  // Q(i,j)
	  conf2_len++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf2[conf2_len][m]=configuration[temp][m];
	}
      }
      
      index=conf2[1][4];
 
      for(neighbour=1;neighbour<=hh[map81[term]].k_to_m_number[index];
	  neighbour++) {
	
	// nearest neighbour is at x,y,z, cell put information
	// into configuration and exit.
	
	/* k ---> m */
	
	nn[1]=conf2[1][1]+hh[map81[term]].k_to_m[index][neighbour][0];
	nn[2]=conf2[1][2]+hh[map81[term]].k_to_m[index][neighbour][1];
	nn[3]=conf2[1][3]+hh[map81[term]].k_to_m[index][neighbour][2];
	nn[4]=hh[map81[term]].k_to_m[index][neighbour][3];
	
	numberFlips=1;
	
	for(m=1;m<=conf2_len;m++)
	  if (nn[1]==conf2[m][1] &&
	      nn[2]==conf2[m][2] &&
	      nn[3]==conf2[m][3] &&
	      nn[4]==conf2[m][4])
	    numberFlips++;
	
	maxNumberFlips=UnitCell.s2[nn[4]];
	
	// the number of spin flips that you can have at a given
	// lattice site is depedent on the quantum spin number
	// of the site in the UNIT CELL index `length'
	
	if (numberFlips <= maxNumberFlips) {
	  
	  temp_number_conf2_sites=conf2_len+1;
	  
	  conf2[temp_number_conf2_sites][1]=nn[1];
	  conf2[temp_number_conf2_sites][2]=nn[2];
	  conf2[temp_number_conf2_sites][3]=nn[3];
	  conf2[temp_number_conf2_sites][4]=nn[4];
	  
	  descriptorIndex1=findClusterIndex(conf1_len, conf1);
	  if (descriptorIndex1 <= 0) continue;
	  
	  descriptorIndex2=findExcitedClusterIndex(temp_number_conf2_sites, conf2);
	  if (descriptorIndex2 <= 0) continue;
					  
	  multiplier=-2*UnitCell.s2[nn[4]]; // factor of -4 s here from H.
	  
	  if (Control.file) {
	    if ((descriptorIndex1 > 0)&&(descriptorIndex2 > 0)){
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=term; // Jun08: origin was 1..81, now using mapped term
	      indices[2]=descriptorIndex2; // Q(i,j)
	      indices[3]=descriptorIndex1;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=multiplier*
		hh[map81[term]].coefficient[index][neighbour];
            collect_term(indices,dummy,3);
	    }
	  }
	  
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}  

double term_exc_55(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{
  /* TERM 55: 

     4 * Q(k)(m) . H(k)(m) . S(+)(k) . S(+)(m).
     
  */

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  double sum;
  sum=0.0000;

  int (*configuration)[5];

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) 
	  conf1[1][m]=configuration[k1][m]; // Q(k)

	for(m=1;m<5;m++) 
	  conf1[2][m]=configuration[k2][m]; // Q(m)
	
	for(m=1;m<5;m++) 
	  conf2[1][m]=configuration[k1][m]; // G(k)
	
	for(m=1;m<5;m++) 
	  conf2[2][m]=configuration[k2][m]; // G(m)
	
	// loop over nearest neighbours
	
	are_they_neighbours=0;
	index=conf1[1][4];      
	
	for(j=1;j<=hh[map81[term]].k_to_m_number[index];j++) {
	  
	  // nearest neighbour is at x,y,z, temp put information
	  // into configuration and exit.
	  
	  nn[1]=conf1[1][1]+hh[map81[term]].k_to_m[index][j][0];
	  nn[2]=conf1[1][2]+hh[map81[term]].k_to_m[index][j][1];
	  nn[3]=conf1[1][3]+hh[map81[term]].k_to_m[index][j][2];
	  nn[4]=hh[map81[term]].k_to_m[index][j][3];
	  
	  if(conf1[2][1] == nn[1] &&
	     conf1[2][2] == nn[2] &&
	     conf1[2][3] == nn[3] &&
	     conf1[2][4] == nn[4])
	    are_they_neighbours=j;      

	  if (are_they_neighbours) break;
	}
	
	// conf1[1] and conf2[1] need to be nearest-neighbours 
	
	if (are_they_neighbours) {
	  
	  counter=1;
	  for(l=1;l<=length;l++) {
	    if (l != k1 && l != k2 ) {
	      tempArray[counter]=l;
	      counter++;
	    }
	  }
	  counter--;
	  
	  for(l=0;l<binary_max[counter];l++) {
	    
	    binary(l,counter,spin);
	    conf1_len=2;
	    conf2_len=2;
	    
	    for (j=1;j<=counter;j++) { 
	      if (spin[j]) {
		conf1_len++;
		temp=tempArray[j];
		for (m=1;m<5;m++) 
		  conf1[conf1_len][m]=configuration[temp][m];
	      } else {
		conf2_len++;
		temp=tempArray[j];
		for (m=1;m<5;m++) 
		  conf2[conf2_len][m]=configuration[temp][m];
	      }
	    }	      
	    
	    descriptorIndex1=findExcitedClusterIndex(conf1_len, conf1);
	    if (descriptorIndex1 <= 0) continue;
	    
	    descriptorIndex2=findClusterIndex(conf2_len, conf2);
	    if (descriptorIndex2 <= 0) continue;
   
	    multiplier=4.0; // factor of 4 from H.
	    
	    if (Control.file) {
	      if ((descriptorIndex1 > 0)&&(descriptorIndex2 > 0)){
		int indices[7]={0};
		double dummy;
		indices[0]=config_index;
		indices[1]=term; // Jun08: origin was 1..81, now using mapped term
		indices[2]=descriptorIndex1;
		indices[3]=descriptorIndex2;
		indices[4]=0;
		indices[5]=0;
		dummy=multiplier*
		  hh[map81[term]].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,3);
	      }
	    }
	    
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}

double term_exc_56(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 56: 
     
     4 * Q(k,m) . F(k) . F(m) . S(+)(k) . S(+)(m)
   
    */
	
  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);


  int spin[MAX_LSUBM+1];
  int mm,nnn,temp1,temp2,temp3;
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int (*conf3)[5];
  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  int descriptorIndex3;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int conf3_len;
  int tempArray[MAX_LSUBM+1];
  double sum;
  sum=0.0000;

  int (*configuration)[5];

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf3=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
      conf3[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) conf1[1][m]=configuration[k1][m]; // F(k)
	for(m=1;m<5;m++) conf2[1][m]=configuration[k2][m]; // F(m)
	for(m=1;m<5;m++) conf3[1][m]=configuration[k1][m]; // Q(k,m) <--- k
	
	// loop over nearest neighbours
	
	are_they_neighbours=0;
	index=conf1[1][4];      

	for(j=1;j<=hh[map81[term]].k_to_m_number[index];j++) {
	  
	  // nearest neighbour is at x,y,z, temp put information
	  // into configuration and exit.
	  
	  /* k ---> m */
	  
	  nn[1]=conf1[1][1]+hh[map81[term]].k_to_m[index][j][0];
	  nn[2]=conf1[1][2]+hh[map81[term]].k_to_m[index][j][1];
	  nn[3]=conf1[1][3]+hh[map81[term]].k_to_m[index][j][2];
	  nn[4]=hh[map81[term]].k_to_m[index][j][3];
	  
	  if(conf2[1][1] == nn[1] &&
	     conf2[1][2] == nn[2] &&
	     conf2[1][3] == nn[3] &&
	     conf2[1][4] == nn[4])
	    are_they_neighbours=j;

	  if (are_they_neighbours) break;
	}
	
	// conf1[1] and conf2[1] need to be nearest-neighbours 
	
	if (are_they_neighbours) {
	  
	  for(m=1;m<5;m++) conf3[2][m]=configuration[k2][m]; // Q(k,m) <--- m
	  
	  counter=1;
	  for(l=1;l<=length;l++) {
	    if (l != k1 && l != k2 ) {
	      tempArray[counter]=l;
		counter++;
	    }
	  }
	  
	  counter--;
	  
	  if (counter > 30) {
	    printf("TERM 32: counter > 30...\nExiting!!\n\n");
	    exit(1);
	  }
          temp = counter; // split into summands not greater than 10 each
          temp3 = ((temp>10) ? 10 : temp); temp -= temp3;  // 0..10
          temp2 = ((temp>10) ? 10 : temp); temp -= temp2;  // 0..10
          temp1 = ((temp>10) ? 10 : temp); temp -= temp1;  // 0..10
	  
	  for(l=0;l<trimary_max[temp1];l++) {
	    for(mm=0;mm<trimary_max[temp2];mm++) {
	      for(nnn=0;nnn<trimary_max[temp3];nnn++) {
		
		trimary(l,mm,nnn,counter,spin);
		conf1_len=1;
		conf2_len=1;
		conf3_len=2;
		  
		for (j=1;j<=counter;j++) { 
		  temp=tempArray[j];
		  if (spin[j]==0) {          // F(k)
		    conf1_len++;
		    for (m=1;m<5;m++) 
		      conf1[conf1_len][m]=configuration[temp][m];
		  } else if (spin[j]==1) {   // F(m)
		    conf2_len++;
		    for (m=1;m<5;m++) 
		      conf2[conf2_len][m]=configuration[temp][m];
		  } else {                    // Q(k,m)
		    conf3_len++;
		    for (m=1;m<5;m++) 
		      conf3[conf3_len][m]=configuration[temp][m];
		  }
		}	      
		
		descriptorIndex1=findClusterIndex(conf1_len, conf1);
                if (descriptorIndex1 <= 0) continue;

		descriptorIndex2=findClusterIndex(conf2_len, conf2);
                if (descriptorIndex2 <= 0) continue;

		descriptorIndex3=findExcitedClusterIndex(conf3_len, conf3);
                if (descriptorIndex3 <= 0) continue;
		
		multiplier=4.0; // factor of 4 from H.
				  
		if (Control.file) {
		  if ((descriptorIndex1 > 0)&&(descriptorIndex2 > 0)&&
		      (descriptorIndex3 > 0)){
		    int indices[7]={0};
		    double dummy;
		    indices[0]=config_index;
		    indices[1]=term; // Jun08: origin was 1..81, now using mapped term
		    indices[2]=descriptorIndex3; // q(i,j)
		    indices[3]=descriptorIndex2; // f(i)
		    indices[4]=descriptorIndex1; // f(j)
		    indices[5]=0;
		    dummy=multiplier*
		      hh[map81[term]].coefficient[index][are_they_neighbours];
                    collect_term(indices,dummy,3);
		  }
		}
		  
	      }
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(conf3);
  free(configuration);
  return sum;
}

double term_exc_57(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 57:
     
     2 * L(k)(k)(m) . S(+)(k) . S(z)(m)
     
  */

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i,j,k,l,m;
  int descriptorIndex;
  int (*configuration)[5];
  int (*conf)[5];
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int maxNumberFlips1;
  int maxNumberFlips2;
  int numberFlips1;
  int numberFlips2;
  int temp_length;
  double temp_multiplier;
  double sum;
  sum=0.0000;

  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));
  conf=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++) 
      configuration[i][j]=string[i][j];
  
  for(i=1;i<=MAX_LSUBM;i++)
    for(j=1;j<5;j++) 
      conf[i][j]=0;

  for(k=1;k<=length;k++) {
    
    index=configuration[k][4];
    maxNumberFlips=UnitCell.s2[index];

    if (maxNumberFlips > 1) {
      
      for(j=1;j<=length;j++)
		for(m=1;m<5;m++)
			conf[j][m]=configuration[j][m]; // L(k,k,m)
      
      temp_length=length+1;
      
      // we repeat the `k' 
      
      for(m=1;m<5;m++)
	conf[temp_length][m]=configuration[k][m];

      index=configuration[k][4];
      
      for(neighbour=1;neighbour<=hh[map81[term]].k_to_m_number[index];
	  neighbour++) {
	
	// nearest neighbour is at x,y,z, cell put information
	// into configuration and exit.
	
	/* k ---> m */
	
	nn[1]=configuration[k][1]+
	  hh[map81[term]].k_to_m[index][neighbour][0];
	nn[2]=configuration[k][2]+
	  hh[map81[term]].k_to_m[index][neighbour][1];
	nn[3]=configuration[k][3]+
	  hh[map81[term]].k_to_m[index][neighbour][2];
	nn[4]=hh[map81[term]].k_to_m[index][neighbour][3];
	
	numberFlips1=1;
	
	for(l=1;l<=temp_length;l++)
	  if (nn[1]==conf[l][1] &&
	      nn[2]==conf[l][2] &&
	      nn[3]==conf[l][3] &&
	      nn[4]==conf[l][4])
	    numberFlips1++;
	
	numberFlips2=0;
	
	for(l=1;l<=temp_length;l++)
	  if (configuration[k][1]==conf[l][1] &&
	      configuration[k][2]==conf[l][2] &&
	      configuration[k][3]==conf[l][3] &&
	      configuration[k][4]==conf[l][4])
	    numberFlips2++;
	
	maxNumberFlips1=UnitCell.s2[nn[4]];
	maxNumberFlips2=UnitCell.s2[index];

	// the number of spin flips that you can have at a given
	// lattice site is depedent on the quantum spin number
	// of the site in the UNIT CELL index `length'
	
	if (numberFlips1 <= maxNumberFlips1 && 
	    numberFlips2 <= maxNumberFlips2)  {
	  
	  temp_length=length+2;
	  
	  conf[temp_length][1]=nn[1]; // `m' is the nearest neighbour.
	  conf[temp_length][2]=nn[2];
	  conf[temp_length][3]=nn[3];
	  conf[temp_length][4]=nn[4];
	  
	  descriptorIndex=findExcitedClusterIndex(temp_length,conf);
	  
	  temp_multiplier=-UnitCell.s2[nn[4]]; // -2 * s from Hamiltonian .
	  
	  // coefficient from \tilde s^- \tilde s^- expansion
	    
	  if (Control.file) {
	    if (descriptorIndex > 0) {	
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=term; // Jun08: origin was 1..81, now using mapped term
	      indices[2]=descriptorIndex; // L(i,i,j)
	      indices[3]=0;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=temp_multiplier*
		hh[map81[term]].coefficient[index][neighbour];
            collect_term(indices,dummy,3);
	    }
	  }

	}
      }
    }
  }
  free(configuration);
  free(conf);
  return sum;
}

double term_exc_58(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 58: 
     
     2 . L(k)(k)(m) . F(m) . S(+)(k) . S(+)(m)
     
    */

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int maxNumberFlips1;
  int maxNumberFlips;
  int numberFlips1;
  double sum;
  sum=0.0000;

  int (*configuration)[5];

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {

	maxNumberFlips=UnitCell.s2[configuration[k1][4]];
	
	if (maxNumberFlips > 1) {	  

	  for(m=1;m<5;m++) {
	    conf1[1][m]=configuration[k1][m]; // L(k,k,m) k--1st one
	    conf1[2][m]=configuration[k1][m]; // L(k,k,m) k--2nd one
	    conf1[3][m]=configuration[k2][m]; // L(k,k,m) m 
	    conf2[1][m]=configuration[k2][m]; // F(m)     m
	  }
	  
	  // loop over nearest neighbours
	  
	  are_they_neighbours=0;
	  index=conf1[1][4];      
	  
	  for(j=1;j<=hh[map81[term]].k_to_m_number[index];j++) {
	    
	    // nearest neighbour is at x,y,z, temp put information
	    // into configuration and exit.
	    
	    /* k ---> m */
	    
	    nn[1]=conf1[1][1]+hh[map81[term]].k_to_m[index][j][0];
	    nn[2]=conf1[1][2]+hh[map81[term]].k_to_m[index][j][1];
	    nn[3]=conf1[1][3]+hh[map81[term]].k_to_m[index][j][2];
	    nn[4]=hh[map81[term]].k_to_m[index][j][3];
	    
	    if(conf2[1][1] == nn[1] &&
	       conf2[1][2] == nn[2] &&
	       conf2[1][3] == nn[3] &&
	       conf2[1][4] == nn[4])
	      are_they_neighbours=j;      
	    if (are_they_neighbours) break;
	  }
	  
	  // conf1[1] and conf2[1] need to be nearest-neighbours 
	  
	  if (are_they_neighbours) {
	    
	    counter=1;
	    for(l=1;l<=length;l++) {
	      if (l != k1 && l != k2 ) {
		tempArray[counter]=l;
		counter++;
	      }
	    }
	    
	    counter--;
	    
	    for(l=0;l<binary_max[counter];l++) {
	      
	      binary(l,counter,spin);
	      conf1_len=3;
	      conf2_len=1;
	      
	      for (j=1;j<=counter;j++) { 
		if (spin[j]) {
		  // H
		  conf1_len++;
		  temp=tempArray[j];
		  for (m=1;m<5;m++) 
		    conf1[conf1_len][m]=configuration[temp][m];
		} else {
		  // H
		  conf2_len++;
		  temp=tempArray[j];
		  for (m=1;m<5;m++) 
		    conf2[conf2_len][m]=configuration[temp][m];
		}
	      }	      
	      
	      // make sure that we don't have too many `k' sites.
	      
	      numberFlips1=0;
	      
	      for(m=1;m<=conf1_len;m++) {
		if (configuration[k1][1]==conf1[m][1] &&
		    configuration[k1][2]==conf1[m][2] &&
		    configuration[k1][3]==conf1[m][3] &&
		    configuration[k1][4]==conf1[m][4])
		  numberFlips1++;
	      }

	      maxNumberFlips1=UnitCell.s2[configuration[k1][4]];
	      
	      if (numberFlips1 <= maxNumberFlips1) {
		
		descriptorIndex1=findExcitedClusterIndex(conf1_len, conf1); // L(i,i,j)
		if (descriptorIndex1 <= 0) continue;
		
		descriptorIndex2=findClusterIndex(conf2_len, conf2);	 // F(j)	
		if (descriptorIndex2 <= 0) continue;

		multiplier=2.0; // from Hamiltonian factor is 2 
		
		if (Control.file) {
		  if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)){
		    int indices[7]={0};
		    double dummy;
		    indices[0]=config_index;
		    indices[1]=term; // Jun08: origin was 1..81, now using mapped term
		    indices[2]=descriptorIndex1; // L(i,i,j)
		    indices[3]=descriptorIndex2; // f(j)
		    indices[4]=0;
		    indices[5]=0;
		    dummy=multiplier*
		      hh[map81[term]].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,3);
		  }
		}

	      }
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}

  
double term_exc_59(int config_index, int length, int (*string)[5],
                   struct hamiltonian_defn hh[], int term)
{

  /* TERM 59:
     
     2 * L(i)(j)(j) . S(+)(j) . S(z)(i)
     
  */

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i,j,k,l,m;
  int descriptorIndex;
  // double multiplier;
  int (*configuration)[5];
  int (*conf)[5];
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int maxNumberFlips1;
  int maxNumberFlips2;
  int numberFlips1;
  int numberFlips2;
  int temp_length;
  double temp_multiplier;
  double sum;
  sum=0.0000;

  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));
  conf=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++) 
      configuration[i][j]=string[i][j];
  
  for(i=1;i<=MAX_LSUBM;i++)
    for(j=1;j<5;j++) 
      conf[i][j]=0;

  for(k=1;k<=length;k++) { // the variable "k" contains the index m here.
    
    index=configuration[k][4];
    maxNumberFlips=UnitCell.s2[index];

    if (maxNumberFlips > 1) {
      
      for(j=1;j<=length;j++)
	for(m=1;m<5;m++)
	  conf[j][m]=configuration[j][m]; // L(k,m,m)
      
      temp_length=length+1;
      
      // we repeat the `m' 
      
      for(m=1;m<5;m++)
	conf[temp_length][m]=configuration[k][m]; // again variable "k" is index "m"
            
      for(neighbour=1;neighbour<=hh[map81[term]].m_to_k_number[index];
	  neighbour++) {
	
	// nearest neighbour is at x,y,z, cell put information
	// into configuration and exit.
	
	/* m ---> k */
	
	nn[1]=configuration[k][1]+
	  hh[map81[term]].m_to_k[index][neighbour][0];
	nn[2]=configuration[k][2]+
	  hh[map81[term]].m_to_k[index][neighbour][1];
	nn[3]=configuration[k][3]+
	  hh[map81[term]].m_to_k[index][neighbour][2];
	nn[4]=hh[map81[term]].m_to_k[index][neighbour][3];
	
	numberFlips1=1;
	
	for(l=1;l<=temp_length;l++)
	  if (nn[1]==conf[l][1] &&
	      nn[2]==conf[l][2] &&
	      nn[3]==conf[l][3] &&
	      nn[4]==conf[l][4])
	    numberFlips1++;
	
	numberFlips2=0;
	
	for(l=1;l<=temp_length;l++)
	  if (configuration[k][1]==conf[l][1] &&
	      configuration[k][2]==conf[l][2] &&
	      configuration[k][3]==conf[l][3] &&
	      configuration[k][4]==conf[l][4])
	    numberFlips2++;
	
	maxNumberFlips1=UnitCell.s2[nn[4]];
	maxNumberFlips2=UnitCell.s2[index];
	
	// the number of spin flips that you can have at a given
	// lattice site is depedent on the quantum spin number
	// of the site in the UNIT CELL index `length'
	
	if (numberFlips1 <= maxNumberFlips1 &&
	    numberFlips2 <= maxNumberFlips2) {	
	  
	  temp_length=length+2;
	  
	  conf[temp_length][1]=nn[1];
	  conf[temp_length][2]=nn[2];
	  conf[temp_length][3]=nn[3];
	  conf[temp_length][4]=nn[4];
	  
	  descriptorIndex=findExcitedClusterIndex(temp_length,conf);
	  
	  temp_multiplier=-1*UnitCell.s2[nn[4]]; // -2 * s from Hamiltonian.
	    
	  if (Control.file) {
	    if (descriptorIndex > 0) {
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=term; // Jun08: origin was 1..81, now using mapped term
	      indices[2]=descriptorIndex;
              indices[3]=0;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=temp_multiplier*
		hh[map81[term]].coefficient[index][neighbour];
            collect_term(indices,dummy,3);
	    }
	  }
	}
      }
    }
  }
  free(configuration);
  free(conf);
  return sum;
}

double term_exc_60(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 60: 
     
     2 * L(m)(m)(k) . F(k) . S(+)(k) . S(+)(m)
     
    */

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int maxNumberFlips1;
  int maxNumberFlips;
  int numberFlips1;
  double sum;
  sum=0.0000;

  int (*configuration)[5];

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {

	maxNumberFlips=UnitCell.s2[configuration[k1][4]];
	
	if (maxNumberFlips > 1) {
	  
	  for(m=1;m<5;m++) {
	    conf1[1][m]=configuration[k1][m]; // L(m,m,k) m--1st one
	    conf1[2][m]=configuration[k1][m]; // L(m,m,k) m--2nd one
	    conf1[3][m]=configuration[k2][m]; // L(m,m,k) k 
	    conf2[1][m]=configuration[k2][m]; // F(k)     k
	  }
	  
	  // loop over nearest neighbours
	  
	  are_they_neighbours=0;
	  index=conf1[1][4];      
	  
	  for(j=1;j<=hh[map81[term]].m_to_k_number[index];j++) {
	    
	    // nearest neighbour is at x,y,z, temp put information
	    // into configuration and exit.
	    
	    /* m ---> k */
	    
	    nn[1]=conf1[1][1]+hh[map81[term]].m_to_k[index][j][0];
	    nn[2]=conf1[1][2]+hh[map81[term]].m_to_k[index][j][1];
	    nn[3]=conf1[1][3]+hh[map81[term]].m_to_k[index][j][2];
	    nn[4]=hh[map81[term]].m_to_k[index][j][3];
	    
	    if(conf2[1][1] == nn[1] &&
	       conf2[1][2] == nn[2] &&
	       conf2[1][3] == nn[3] &&
	       conf2[1][4] == nn[4])
	      are_they_neighbours=j;      
	    if (are_they_neighbours) break;
	  }
	  
	  // conf1[1] and conf2[1] need to be nearest-neighbours 
	  
	  if (are_they_neighbours) {
	    
	    counter=1;
	    for(l=1;l<=length;l++) {
	      if (l != k1 && l != k2 ) {
		tempArray[counter]=l;
		counter++;
	      }
	    }
	    
	    counter--;
	    
	    for(l=0;l<binary_max[counter];l++) {
	      
	      binary(l,counter,spin);
	      conf1_len=3;
	      conf2_len=1;
	      
	      for (j=1;j<=counter;j++) { 
		if (spin[j]) {
		  // H
		  conf1_len++;
		  temp=tempArray[j];
		  for (m=1;m<5;m++) 
		    conf1[conf1_len][m]=configuration[temp][m];
		} else {
		  // H
		  conf2_len++;
		  temp=tempArray[j];
		  for (m=1;m<5;m++) 
		    conf2[conf2_len][m]=configuration[temp][m];
		}
	      }	      
	      
	      // make sure that we don't have too many `k' sites.
	      
	      numberFlips1=0;
	      
	      for(m=1;m<=conf1_len;m++) {
		if (configuration[k1][1]==conf1[m][1] &&
		    configuration[k1][2]==conf1[m][2] &&
		    configuration[k1][3]==conf1[m][3] &&
		    configuration[k1][4]==conf1[m][4])
		  numberFlips1++;
	      }
	      maxNumberFlips1=UnitCell.s2[configuration[k1][4]];
	      
	      if (numberFlips1 <= maxNumberFlips1) {
		
		descriptorIndex1=findExcitedClusterIndex(conf1_len, conf1);
		if (descriptorIndex1 <= 0) continue;
		
		descriptorIndex2=findClusterIndex(conf2_len, conf2);
		if (descriptorIndex2 <= 0) continue;

		multiplier=2.0; // factor of 2 from Hamiltonian.
		
		if (Control.file) {
		  if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)) {
		    int indices[7]={0};
		    double dummy;
		    indices[0]=config_index;
		    indices[1]=term; // Jun08: origin was 1..81, now using mapped term
		    indices[2]=descriptorIndex1; // L(i,j,j)
		    indices[3]=descriptorIndex2; // F(i)
		    indices[4]=0;
		    indices[5]=0;
		    dummy=multiplier*
		      hh[map81[term]].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,3);
			  }
			}
	      }
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}

double term_exc_61(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 61: 

     M(k)(k)(m)(m) . S(+)(k) . S(+)(m).

  */

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index, index1, index2;
  int maxNumberFlips1, maxNumberFlips2;
  int numberFlips1, numberFlips2;
  int descriptorIndex;
  // double multiplier;
  int (*configuration)[5];
  int (*conf)[5];
  int temp_length;
  double sum;
  sum=0.0000;

  configuration=(int (*)[5])malloc((MAX_LSUBM+3)*5*sizeof(int));
  conf=         (int (*)[5])malloc((MAX_LSUBM+3)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  for(k=1;k<MAX_LSUBM+3;k++) {
    for(l=1;l<5;l++) {
      conf[k][l]=0;
    }
  } 
  
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      
      if (k1 != k2) {

	for(i=1;i<=length;i++)
	  for(j=1;j<5;j++)
	    conf[i][j]=configuration[i][j];
	
	// a check here...
	
	// now add in extra `k' and `m' sites for M(k,k,m,m)
	
	temp_length=length+1;
	
	for(m=1;m<5;m++)
	  conf[temp_length][m]=configuration[k1][m]; // `k'
	
	temp_length=length+2;
	
	for(m=1;m<5;m++)
	  conf[temp_length][m]=configuration[k2][m]; // `m'
	
	numberFlips1=1;
	numberFlips2=1;
	
	for(l=1;l<=temp_length-2;l++) {
	  if (configuration[k1][1]==conf[l][1] &&
	      configuration[k1][2]==conf[l][2] &&
	      configuration[k1][3]==conf[l][3] &&
	      configuration[k1][4]==conf[l][4])
	    numberFlips1++;
	
	  if (configuration[k2][1]==conf[l][1] &&
	      configuration[k2][2]==conf[l][2] &&
	      configuration[k2][3]==conf[l][3] &&
	      configuration[k2][4]==conf[l][4])
	    numberFlips2++;
	}

	index1=configuration[k1][4];
	maxNumberFlips1=UnitCell.s2[index1];
	
	index2=configuration[k2][4];
	maxNumberFlips2=UnitCell.s2[index2];

	are_they_neighbours=0;
	index=configuration[k1][4]; 
	
	for(j=1;j<=hh[map81[term]].k_to_m_number[index];j++) {
	  
	  // nearest neighbour is at x,y,z, cell put information
	  // into configuration and exit.
	  
	  nn[1]=configuration[k1][1]+
	    hh[map81[term]].k_to_m[index][j][0];
	  nn[2]=configuration[k1][2]+
	    hh[map81[term]].k_to_m[index][j][1];
	  nn[3]=configuration[k1][3]+
	    hh[map81[term]].k_to_m[index][j][2];
	  nn[4]=hh[map81[term]].k_to_m[index][j][3];
	  
	  if(configuration[k2][1] == nn[1] &&
	     configuration[k2][2] == nn[2] &&
	     configuration[k2][3] == nn[3] &&
	     configuration[k2][4] == nn[4])
	    are_they_neighbours=j;      

	  if (are_they_neighbours) break;
	}
	
	if ((maxNumberFlips1 > 1) && 
	    (maxNumberFlips2 > 1) &&
	    (numberFlips1 <= maxNumberFlips1) &&
	    (numberFlips2 <= maxNumberFlips2) &&
	    (are_they_neighbours)) {
	  
	  descriptorIndex=findExcitedClusterIndex(temp_length, conf);
	    
	  if (Control.file) {
	    if (descriptorIndex > 0) {
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=term; // Jun08: origin was 1..81, now using mapped term
	      indices[2]=descriptorIndex;
	      indices[3]=0;
	      indices[4]=0;
              indices[5]=0;
	      dummy=
		hh[map81[term]].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,3); // factor of 1 from Hamiltonian.
	    }
	  }

	}
      }
    }
  }
  free(configuration);
  free(conf);
  return sum;
}

double term_exc_62(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{
  
  /* TERM 62:
     
     4 * P(j) . G(i,j) . S(+)(j) . S(z)(i)
     
  */

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i,j,k,l,m;
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int (*configuration)[5];
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int numberFlips;
  int (*conf1)[5], (*conf2)[5];
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
  int temp_number_conf2_sites;
  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));
  double sum;
  sum=0.0000;

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<=MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 

  for(k=1;k<=length;k++) {
        
    for(m=1;m<5;m++) {
      conf1[1][m]=configuration[k][m]; // p(m)
      conf2[1][m]=configuration[k][m]; // g(m,k) <-- m here.
    }
      
    counter=1;
    for(m=1;m<=length;m++) {
      if (m != k) {
	tempArray[counter]=m;
	counter++;
      }
    }
    
    counter--;
    
    for(l=0;l<binary_max[counter];l++) {
      
      binary(l,counter,spin);
      conf1_len=1;
      conf2_len=1;
      
      for (j=1;j<=counter;j++) { 
	if (spin[j]) {
	  // F
	  conf1_len++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf1[conf1_len][m]=configuration[temp][m];
	} else {
	  // H
	  conf2_len++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf2[conf2_len][m]=configuration[temp][m];
	}
      }
      
      index=conf2[1][4];
      
      for(neighbour=1;neighbour<=hh[map81[term]].m_to_k_number[index];
	  neighbour++) {
	
	// nearest neighbour is at x,y,z, cell put information
	// into configuration and exit.
	
	/* m ---> k */
	
	nn[1]=conf2[1][1]+hh[map81[term]].m_to_k[index][neighbour][0];
	nn[2]=conf2[1][2]+hh[map81[term]].m_to_k[index][neighbour][1];
	nn[3]=conf2[1][3]+hh[map81[term]].m_to_k[index][neighbour][2];
	nn[4]=            hh[map81[term]].m_to_k[index][neighbour][3];
	
	numberFlips=1;
	
	for(m=1;m<=conf2_len;m++)
	  if (nn[1]==conf2[m][1] &&
	      nn[2]==conf2[m][2] &&
	      nn[3]==conf2[m][3] &&
	      nn[4]==conf2[m][4])
	    numberFlips++;
	
	maxNumberFlips=UnitCell.s2[nn[4]];
		
	if (numberFlips <= maxNumberFlips) {
	  
	  temp_number_conf2_sites=conf2_len+1;
	  
	  conf2[temp_number_conf2_sites][1]=nn[1];
	  conf2[temp_number_conf2_sites][2]=nn[2];
	  conf2[temp_number_conf2_sites][3]=nn[3];
	  conf2[temp_number_conf2_sites][4]=nn[4];
	  
	  descriptorIndex1=findExcitedClusterIndex(conf1_len, conf1); // p(i)
	  if (descriptorIndex1 <= 0) continue;
	  
	  descriptorIndex2=findClusterIndex(temp_number_conf2_sites, conf2);
	  if (descriptorIndex2 <= 0) continue;
	  
	  multiplier=-2*UnitCell.s2[nn[4]]; // -4 * s from Hamiltonian.
	      
	  if (Control.file) {
	    if ((descriptorIndex1 > 0)&&(descriptorIndex2 > 0)) {
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=term; // Jun08: origin was 1..81, now using mapped term
	      indices[2]=descriptorIndex1; //p(j)
	      indices[3]=descriptorIndex2; //g(i,j)
	      indices[4]=0;
	      indices[5]=0;
	      dummy=multiplier*
		hh[map81[term]].coefficient[index][neighbour];
            collect_term(indices,dummy,3);
	    }
	  }
	  
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}  

double term_exc_63(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 63: 
     
     2 * P(j) . L(i,i,j) . S(+)(i) . S(+)(j)
     
    */

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int maxNumberFlips1;
  int maxNumberFlips;
  int numberFlips1;
  double sum;
  sum=0.0000;

  int (*configuration)[5];

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {

	maxNumberFlips=UnitCell.s2[configuration[k1][4]];
	
	if (maxNumberFlips > 1) {
	  
	  for(m=1;m<5;m++) {
	    conf1[1][m]=configuration[k1][m]; // L(i,i,j) i--1st one
	    conf1[2][m]=configuration[k1][m]; // L(i,i,j) i--2nd one
	    conf1[3][m]=configuration[k2][m]; // L(i,i,j) j 
	    conf2[1][m]=configuration[k2][m]; // P(j)     j
	  }
	  
	  // loop over nearest neighbours
	  
	  are_they_neighbours=0;
	  index=conf1[1][4];      
	  
	  for(j=1;j<=hh[map81[term]].k_to_m_number[index];j++) {
	    
	    // nearest neighbour is at x,y,z, temp put information
	    // into configuration and exit.
	    
	    /* k ---> m */
	    
	    nn[1]=conf1[1][1]+hh[map81[term]].k_to_m[index][j][0];
	    nn[2]=conf1[1][2]+hh[map81[term]].k_to_m[index][j][1];
	    nn[3]=conf1[1][3]+hh[map81[term]].k_to_m[index][j][2];
	    nn[4]=hh[map81[term]].k_to_m[index][j][3];
	    
	    if(conf2[1][1] == nn[1] &&
	       conf2[1][2] == nn[2] &&
	       conf2[1][3] == nn[3] &&
	       conf2[1][4] == nn[4])
	      are_they_neighbours=j;      
	    if (are_they_neighbours) break;
	  }
	  
	  // conf1[1] and conf2[1] need to be nearest-neighbours 
	  
	  if (are_they_neighbours) {
	    
	    counter=1;
	    for(l=1;l<=length;l++) {
	      if (l != k1 && l != k2 ) {
		tempArray[counter]=l;
		counter++;
	      }
	    }
	    
	    counter--;
	    
	    for(l=0;l<binary_max[counter];l++) {
	      
	      binary(l,counter,spin);
	      conf1_len=3;
	      conf2_len=1;
	      
	      for (j=1;j<=counter;j++) { 
		if (spin[j]) {
		  // H
		  conf1_len++;
		  temp=tempArray[j];
		  for (m=1;m<5;m++) 
		    conf1[conf1_len][m]=configuration[temp][m];
		} else {
		  // H
		  conf2_len++;
		  temp=tempArray[j];
		  for (m=1;m<5;m++) 
		    conf2[conf2_len][m]=configuration[temp][m];
		}
	      }	      
	      
	      // make sure that we don't have too many `k' sites.
	      
	      numberFlips1=0;
	      
	      for(m=1;m<=conf1_len;m++) {
		if (configuration[k1][1]==conf1[m][1] &&
		    configuration[k1][2]==conf1[m][2] &&
		    configuration[k1][3]==conf1[m][3] &&
		    configuration[k1][4]==conf1[m][4])
		  numberFlips1++;
	      }
	      maxNumberFlips1=UnitCell.s2[configuration[k1][4]];
	      
	      if (numberFlips1 <= maxNumberFlips1) {
		
		descriptorIndex1=findClusterIndex(conf1_len, conf1);
		if (descriptorIndex1 <= 0) continue;
		
		descriptorIndex2=findExcitedClusterIndex(conf2_len, conf2);
		if (descriptorIndex2 <= 0) continue;

		multiplier=2.0; // factor of 2 from H.

                if (Control.file) {
		  if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)) {
		    int indices[7]={0};
		    double dummy;
		    indices[0]=config_index;
		    indices[1]=term; // Jun08: origin was 1..81, now using mapped term
		    indices[2]=descriptorIndex2; // p(i)
		    indices[3]=descriptorIndex1; // H(i,j,j)
		    indices[4]=0;
		    indices[5]=0;
		    dummy=multiplier*
		      hh[map81[term]].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,3);
		  }
		}

	      }
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}

double term_exc_64(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 64: 
     
     4 * P(j) . F(i) . S(z)(i) . S(z)(j)

     one of the most time-consuming terms because you've got to sweep
     through the index `k'

     */
  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);


  int i, j, k, l, m;
  double multiplier;

  int maxX, maxY, maxZ;
  int minX, minY, minZ;

  int x,y,z,cell;
  int maxPointX, maxPointY, maxPointZ;
  int index;

  int numberFlips, maxNumberFlips;

  int (*configuration)[5];

  int (*conf1)[5], (*conf2)[5];
  int conf1_len;
  int conf2_len;
  int counter;
  int spin[MAX_LSUBM+2];
  int neighbour;
  int temp_x;
  int temp_y;
  int temp_z;
  int temp_cell;
  int temp_number_conf1_sites;
  int temp_number_conf2_sites;
  int descriptorIndex1;
  int descriptorIndex2;
  double sum;
  sum=0.0000;

  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<=MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 

  maxPointX=0;
  maxPointY=0;
  maxPointZ=0;
  for(j=1;j<=length;j++) {
    if (maxPointX < abs(string[j][1])) maxPointX=abs(string[j][1]);
    if (maxPointY < abs(string[j][2])) maxPointY=abs(string[j][2]);
    if (maxPointZ < abs(string[j][3])) maxPointZ=abs(string[j][3]);
  }

  if (Params.dim == 1) {
    minX=radiusExc;           minY=0;                minZ=0;
    maxX=radiusExc+maxPointX; maxY=0;                maxZ=0;
  } else if (Params.dim == 2) {
    minX=radiusExc;           minY=radiusExc;           minZ=0;
    maxX=radiusExc+maxPointX; maxY=radiusExc+maxPointY; maxZ=0;
  } else if (Params.dim == 3) {
    minX=radiusExc;           minY=radiusExc;           minZ=radiusExc;
    maxX=radiusExc+maxPointX; maxY=radiusExc+maxPointY; maxZ=radiusExc+maxPointZ;
  } else {
    printf("Really weird error. Dimension (%d)"
     " not recognised...\nExiting||\n\n",Params.dim);
    exit(1);
  }


  for(x=-minX;x<=maxX;x++) {
    for(y=-minY;y<=maxY;y++) {
      for(z=-minZ;z<=maxZ;z++) {
	for(cell=1;cell<=UnitCell.number;cell++) {
	  
          // ToDo: skip x+y+z outside triangle range [0..m]?

	  index=cell;

	  counter=length;
	  
	  for(l=0;l<binary_max[counter];l++) {
	    
	    binary(l,counter,spin);   /* convert l to binary string spin[] */

            // spin[j]: 0 = F(m)  1 = F(k)
            // copy sites for 0-bit to conf2 and sites for 1-bit to conf1
	    conf1_len=0;
	    conf2_len=0;
	    
	    for (j=1;j<=counter;j++) { 
	      if (spin[j]) {
		conf1_len++;
		for (m=1;m<5;m++) 
		  conf1[conf1_len][m]=configuration[j][m]; // p(i)
	      } else {
		conf2_len++;
		for (m=1;m<5;m++) 
		  conf2[conf2_len][m]=configuration[j][m]; // f(j)
	      }
	    }
	    
	    numberFlips=1;
	    
	    for(m=1;m<=conf1_len;m++)
	      if (x==conf1[m][1] &&
		  y==conf1[m][2] &&
		  z==conf1[m][3] &&
		  cell==conf1[m][4])
		numberFlips++;
	    
	    maxNumberFlips = UnitCell.s2[cell];
	    
	    // the number of spin flips that you can have at a given
	    // lattice site is depedent on the quantum spin number
	    // of the site in the UNIT CELL index `length'
	    
	    if (numberFlips > maxNumberFlips) continue; 

            // Took this outside inner loop (Joerg Sep2005)
            temp_number_conf1_sites = conf1_len+1;
            conf1[temp_number_conf1_sites][1]=x;
            conf1[temp_number_conf1_sites][2]=y;
            conf1[temp_number_conf1_sites][3]=z;
            conf1[temp_number_conf1_sites][4]=cell;
            
            descriptorIndex1=findExcitedClusterIndex(temp_number_conf1_sites, conf1);
            if (descriptorIndex1 <= 0) continue;
	      
            for(neighbour=1;
                neighbour<=hh[map81[term]].m_to_k_number[cell];
                neighbour++) {
              
              // nearest neighbour is at x,y,z, cell put information
              // into configuration and exit.
              
              /* m ---> k */
              
              temp_x=x+hh[map81[term]].m_to_k[cell][neighbour][0];
              temp_y=y+hh[map81[term]].m_to_k[cell][neighbour][1];
              temp_z=z+hh[map81[term]].m_to_k[cell][neighbour][2];
              temp_cell=
                       hh[map81[term]].m_to_k[cell][neighbour][3];
              
              numberFlips=1;
              
              for(m=1;m<=conf2_len;m++)
                if (temp_x==conf2[m][1] &&
                    temp_y==conf2[m][2] &&
                    temp_z==conf2[m][3] &&
                    temp_cell==conf2[m][4])
                  numberFlips++;
              
              maxNumberFlips=UnitCell.s2[temp_cell];
              
              // the number of spin flips that you can have at a given
              // lattice site is depedent on the quantum spin number
              // of the site in the UNIT CELL index `length'
              
              if (numberFlips <= maxNumberFlips) {
                
                temp_number_conf2_sites = conf2_len+1;
                conf2[temp_number_conf2_sites][1]=temp_x;
                conf2[temp_number_conf2_sites][2]=temp_y;
                conf2[temp_number_conf2_sites][3]=temp_z;
                conf2[temp_number_conf2_sites][4]=temp_cell;
                
                multiplier=UnitCell.s2[cell]*UnitCell.s2[temp_cell]; // factor of 4 * s * s
                
                // coeff from \tilde s^- \tilde s^- 
                
                descriptorIndex2=findClusterIndex(temp_number_conf2_sites, conf2);
                if (descriptorIndex2 <= 0) continue;
                
                if (Control.file) {
                  if ((descriptorIndex1 > 0)&&(descriptorIndex2 > 0)){
                    int indices[7]={0};
                    double dummy;
                    indices[0]=config_index;
                    indices[1]=term; // Jun08: origin was 1..81, now using mapped term
                    indices[2]=descriptorIndex1; // p(j)
                    indices[3]=descriptorIndex2; // f(i)
                    indices[4]=0;
                    indices[5]=0;
                    dummy=multiplier*
                      hh[map81[term]].coefficient[index][neighbour];
                    collect_term(indices,dummy,3);
                  }
                }
                
              }
            }
	  }
	}
      }
    }
  }
  free(configuration);
  free(conf1);
  free(conf2);
  return sum;
}

double term_exc_65(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

   /* TERM 65: 
     
     4 * P(j) . F(i) . F(j) . S(+)(j) . S(z)(i)
     
  */
  
  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int spin[MAX_LSUBM+1];
  int mm,nnn,temp1,temp2,temp3;
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int (*conf3)[5];
  int i, j, k, l, m;
  int nn[5];
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  int descriptorIndex3;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int conf3_len;
  int tempArray[MAX_LSUBM+1];
  int temp_number_conf3_sites;
  int numberFlips;
  int maxNumberFlips;
  int (*configuration)[5];
  double sum;
  sum=0.0000;

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int)); //p(j)
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int)); //f(j)
  conf3=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int)); //f(i)
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
      conf3[k][l]=0;
    }
  } 
 
  for(k=1;k<=length;k++) {
    
    for(m=1;m<5;m++) conf1[1][m]=configuration[k][m]; // P(j)
    for(m=1;m<5;m++) conf2[1][m]=configuration[k][m]; // F(j)
    
    // loop over nearest neighbours
    
    counter=1;
    for(l=1;l<=length;l++) {
      if (l != k) {
	tempArray[counter]=l;
	counter++;
      }
    }
    
    counter--;
    
    if (counter > 30) {
      printf("TERM 33: counter > 30...\nExiting!!\n\n");
      exit(1);
    }
    temp = counter; // split into summands not greater than 10 each
    temp3 = ((temp>10) ? 10 : temp); temp -= temp3;  // 0..10
    temp2 = ((temp>10) ? 10 : temp); temp -= temp2;  // 0..10
    temp1 = ((temp>10) ? 10 : temp); temp -= temp1;  // 0..10
    
    for(l=0;l<trimary_max[temp1];l++) {
      for(mm=0;mm<trimary_max[temp2];mm++) {
	for(nnn=0;nnn<trimary_max[temp3];nnn++) {
	  
	  trimary(l,mm,nnn,counter,spin);
	  conf1_len=1;
	  conf2_len=1;
	  conf3_len=0;
	  
	  for (j=1;j<=counter;j++) { 
	    temp=tempArray[j];
	    if (spin[j]==0) {         // P(j)
	      conf1_len++;
	      for (m=1;m<5;m++) 
		conf1[conf1_len][m]=configuration[temp][m];
	    } else if (spin[j]==1) {  // F(j)
	      conf2_len++;
	      for (m=1;m<5;m++) 
		conf2[conf2_len][m]=configuration[temp][m];
	    } else {                  // F(i)
	      conf3_len++;
	      for (m=1;m<5;m++) 
		conf3[conf3_len][m]=configuration[temp][m];
	    }
	  }	      
	  

          descriptorIndex1=findExcitedClusterIndex(conf1_len, conf1);
          if (descriptorIndex1 <= 0) continue;

          descriptorIndex2=findClusterIndex(conf2_len, conf2);
          if (descriptorIndex2 <= 0) continue;
	  
	  index=conf1[1][4];      
	  for(j=1;j<=hh[map81[term]].m_to_k_number[index];j++) {
	    
	    // nearest neighbour is at x,y,z, temp put information
	    // into configuration and exit.
	    
	    /* m ---> k */
	    
	    nn[1]=conf1[1][1]+hh[map81[term]].m_to_k[index][j][0];
	    nn[2]=conf1[1][2]+hh[map81[term]].m_to_k[index][j][1];
	    nn[3]=conf1[1][3]+hh[map81[term]].m_to_k[index][j][2];
	    nn[4]=hh[map81[term]].m_to_k[index][j][3];
	    
	    numberFlips=1;
	    
	    for(m=1;m<=conf3_len;m++)
	      if (nn[1]==conf3[m][1] &&
		  nn[2]==conf3[m][2] &&
		  nn[3]==conf3[m][3] &&
		  nn[4]==conf3[m][4])
		numberFlips++;
	    
	    maxNumberFlips=UnitCell.s2[nn[4]];
	    
	    // the number of spin flips that you can have at a given
	    // lattice site is depedent on the quantum spin number
	    // of the site in the UNIT CELL index `length'
	    
	    if (numberFlips <= maxNumberFlips) {
	      
	      temp_number_conf3_sites = conf3_len+1;
	      conf3[temp_number_conf3_sites][1]=nn[1];
	      conf3[temp_number_conf3_sites][2]=nn[2];
	      conf3[temp_number_conf3_sites][3]=nn[3];
	      conf3[temp_number_conf3_sites][4]=nn[4];
	      
	      descriptorIndex3=findClusterIndex(temp_number_conf3_sites, conf3);
	      if (descriptorIndex3 <= 0) continue;
	      
	      multiplier=-2*UnitCell.s2[nn[4]]; // -4 * s
	      
	      
	      if (Control.file) {
		if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0) &&
		    (descriptorIndex3 > 0)){
		  int indices[7]={0};
		  double dummy;
		  indices[0]=config_index;
		  indices[1]=term; // Jun08: origin was 1..81, now using mapped term
		  indices[2]=descriptorIndex1; //p(j)
		  indices[3]=descriptorIndex2; //f(j)
		  indices[4]=descriptorIndex3; //f(i)
		  indices[5]=0;
		  dummy=multiplier*
		    hh[map81[term]].coefficient[index][j];
                  collect_term(indices,dummy,3);
		}
	      }
		
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(conf3);
  free(configuration);
  return sum;
}

double term_exc_66(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{
  
  /* TERM 66:
     
     2 * p(j) . G(i,i) . S(+)(i) . S(z)(j)
     
  */

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i,j,k,l,m;
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int (*configuration)[5];
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips1;
  int numberFlips1;
  int maxNumberFlips2;
  int numberFlips2;
  int (*conf1)[5], (*conf2)[5];
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
  int temp_number_conf2_sites;
  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));
  double sum;
  sum=0.0000;

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<=MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 

  for(k=1;k<=length;k++) {
        
    for(m=1;m<5;m++) {
      conf1[1][m]=configuration[k][m]; // g(i) <- first one
      conf1[2][m]=configuration[k][m]; // g(i) <- second one
    }
      
    counter=1;
    for(m=1;m<=length;m++) {
      if (m != k) {
	tempArray[counter]=m;
	counter++;
      }
    }
    
    counter--;
    
    for(l=0;l<binary_max[counter];l++) {
      
      binary(l,counter,spin);
      conf1_len=2;
      conf2_len=0;
      
      for (j=1;j<=counter;j++) { 
	if (spin[j]) {
	  // G(i,i)
	  conf1_len++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf1[conf1_len][m]=configuration[temp][m];
	} else {
	  // P(j)
	  conf2_len++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf2[conf2_len][m]=configuration[temp][m];
	}
      }
      
      index=conf2[1][4];
      
      for(neighbour=1;neighbour<=hh[map81[term]].k_to_m_number[index];
	  neighbour++) {
	
	// nearest neighbour is at x,y,z, cell put information
	// into configuration and exit.
	
	/* k ---> m */
	
	nn[1]=conf1[1][1]+hh[map81[term]].k_to_m[index][neighbour][0];
	nn[2]=conf1[1][2]+hh[map81[term]].k_to_m[index][neighbour][1];
	nn[3]=conf1[1][3]+hh[map81[term]].k_to_m[index][neighbour][2];
	nn[4]=            hh[map81[term]].k_to_m[index][neighbour][3];
	
	  temp_number_conf2_sites=conf2_len+1;
	  
	  conf2[temp_number_conf2_sites][1]=nn[1]; // P(j) 
	  conf2[temp_number_conf2_sites][2]=nn[2];
	  conf2[temp_number_conf2_sites][3]=nn[3];
	  conf2[temp_number_conf2_sites][4]=nn[4];
	  
	  numberFlips1=0;
	
	for(m=1;m<=temp_number_conf2_sites;m++)
	  if (nn[1]==conf2[m][1] &&
	      nn[2]==conf2[m][2] &&
	      nn[3]==conf2[m][3] &&
	      nn[4]==conf2[m][4])
	    numberFlips1++;
	
	maxNumberFlips1=UnitCell.s2[nn[4]];
	
	// the number of spin flips that you can have at a given
	// lattice site is depedent on the quantum spin number
	// of the site in the UNIT CELL index `length'

	  numberFlips2=2;
	for(m=3;m<=conf1_len;m++)
	  if (conf1[1][1]==conf1[m][1] &&
	      conf1[1][2]==conf1[m][2] &&
	      conf1[1][3]==conf1[m][3] &&
	      conf1[1][4]==conf1[m][4])
	    numberFlips2++;
	
	maxNumberFlips2=UnitCell.s2[conf1[1][4]];
	
	if ((numberFlips1 <= maxNumberFlips1) && (numberFlips2 <= maxNumberFlips2)) {

	  // g(i,i)
	  descriptorIndex1=findClusterIndex(conf1_len, conf1);
	  if (descriptorIndex1 <= 0) continue;
	  
          // p(j)
	  descriptorIndex2=findExcitedClusterIndex(temp_number_conf2_sites, conf2);
          if (descriptorIndex2 <= 0) continue;
	  
	  multiplier=(-2)*0.5*UnitCell.s2[nn[4]]; // - 2 * s
	      
	  if (Control.file) {
	    if ((descriptorIndex1 > 0)&&(descriptorIndex2 > 0)) {
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=term; // Jun08: origin was 1..81, now using mapped term
	      indices[2]=descriptorIndex2; //p(j)
	      indices[3]=descriptorIndex1; //g(i,i)
	      indices[4]=0;
	      indices[5]=0;
	      dummy=multiplier*
		hh[map81[term]].coefficient[index][neighbour];
            collect_term(indices,dummy,3);
	    }
	  }
	  
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}  

double term_exc_67(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 67:
     
     2 . p(j) . f(j) . g(i)(i) . S(+)(i) . S(+)(j)
     
  */

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i,j,k,l,m;
  int mm,nnn,temp1,temp2,temp3;
  int k1, k2;
  int descriptorIndex1;
  int descriptorIndex2;
  int descriptorIndex3;
  // double multiplier;
  int (*configuration)[5];
  int index;
  int nn[5];
  int maxNumberFlips;
  int numberFlips;
  int (*conf1)[5], (*conf2)[5], (*conf3)[5];
  int conf1_len;
  int conf2_len;
  int conf3_len;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
  int are_they_neighbours;

  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));
  double sum;
  sum=0.0000;

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf3=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<=MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
      conf3[k][l]=0;
    }
  } 

  for(k1=1;k1<=length;k1++) { 
    for(k2=1;k2<=length;k2++) { 
      
      if (k1 != k2) {
	
	index=configuration[k1][4];
	maxNumberFlips=UnitCell.s2[index];
	
	if (maxNumberFlips > 1) {
	  
	  for(m=1;m<5;m++) {
	    conf1[1][m]=configuration[k1][m]; // g(i,i) <-- 1st i here
	    conf1[2][m]=configuration[k1][m]; // g(i,i) <-- 2nd i here
	    conf2[1][m]=configuration[k2][m]; // P(j)
	    conf3[1][m]=configuration[k2][m]; // F(j)
	  }
	  
	  // loop over nearest neighbours
	  
	  are_they_neighbours=0;
	  index=conf1[1][4];      
	  
	  for(j=1;j<=hh[map81[term]].k_to_m_number[index];j++) {
	    
	    // nearest neighbour is at x,y,z, temp put information
	    // into configuration and exit.
	    
	    /* k ---> m */
	    
	    nn[1]=conf1[1][1]+hh[map81[term]].k_to_m[index][j][0];
	    nn[2]=conf1[1][2]+hh[map81[term]].k_to_m[index][j][1];
	    nn[3]=conf1[1][3]+hh[map81[term]].k_to_m[index][j][2];
	    nn[4]=hh[map81[term]].k_to_m[index][j][3];
	    
	    if(conf2[1][1] == nn[1] &&
	       conf2[1][2] == nn[2] &&
	       conf2[1][3] == nn[3] &&
	       conf2[1][4] == nn[4])
	      are_they_neighbours=j;      

	    if (are_they_neighbours) break;
	  }
	  
	  // conf1[1] and conf2[1] need to be nearest-neighbours 
	  
	  if (are_they_neighbours) {
	    
	    counter=1;
	    for(m=1;m<=length;m++) {
	      if (m != k1 && m != k2) {
		tempArray[counter]=m;
		counter++;
	      }
	    }
	    
	    counter--;

	    if (counter > 30) {
	      printf("TERM 45: counter > 30...\nExiting!!\n\n");
	      exit(1);
	    }
            temp = counter; // split into summands not greater than 10 each
            temp3 = ((temp>10) ? 10 : temp); temp -= temp3;  // 0..10
            temp2 = ((temp>10) ? 10 : temp); temp -= temp2;  // 0..10
            temp1 = ((temp>10) ? 10 : temp); temp -= temp1;  // 0..10
	    
	    for(l=0;l<trimary_max[temp1];l++) {
	      for(mm=0;mm<trimary_max[temp2];mm++) {
		for(nnn=0;nnn<trimary_max[temp3];nnn++) {
		  
		  trimary(l,mm,nnn,counter,spin);
		  
		  conf1_len=2;
		  conf2_len=1;
		  conf3_len=1;

		  for (j=1;j<=counter;j++) { 
		    if (spin[j]==0) {
		      // H(m,m)
		      conf1_len++;
		      temp=tempArray[j];
		      for (m=1;m<5;m++) 
			conf1[conf1_len][m]=configuration[temp][m];
		    } else if (spin[j]==1) {
		      // F(k)
		      conf2_len++;
		      temp=tempArray[j];
		      for (m=1;m<5;m++) 
			conf2[conf2_len][m]=configuration[temp][m];
		    } else {
		      // F(k)
		      conf3_len++;
		      temp=tempArray[j];
		      for (m=1;m<5;m++) 
			conf3[conf3_len][m]=configuration[temp][m];
		    }
		  }
	      
		  numberFlips=0;
	      
		  for(m=1;m<=conf1_len;m++)
		    if (configuration[k1][1]==conf1[m][1] &&
			configuration[k1][2]==conf1[m][2] &&
			configuration[k1][3]==conf1[m][3] &&
			configuration[k1][4]==conf1[m][4])
		      numberFlips++;
	      
		  maxNumberFlips=UnitCell.s2[configuration[k1][4]];
	      
		  // the number of spin flips that you can have at a given
		  // lattice site is depedent on the quantum spin number
		  // of the site in the UNIT CELL index `length'
	      
		  if (numberFlips <= maxNumberFlips) {
		
		    // H(m,m)
		    descriptorIndex1=findClusterIndex(conf1_len, conf1);
		    if (descriptorIndex1 <= 0) continue;
		    
		    // P(k)
		    descriptorIndex2=findExcitedClusterIndex(conf2_len, conf2);
	            if (descriptorIndex2 <= 0) continue;
		
		    // F(k)
		    descriptorIndex3=findClusterIndex(conf3_len, conf3);
		    if (descriptorIndex3 <= 0) continue;
		
		    if (Control.file) {
		      if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0) &&
			  (descriptorIndex3 > 0)) {
			int indices[7]={0};
			double dummy;
			indices[0]=config_index;
			indices[1]=term; // Jun08: origin was 1..81, now using mapped term
			indices[2]=descriptorIndex2; // p(i)
			indices[3]=descriptorIndex1; // g(j,j)
			indices[4]=descriptorIndex3; // f(i)
			indices[5]=0;
			dummy=2*hh[map81[term]].coefficient[index][are_they_neighbours]; // 2 factor from Hamiltonian 
            collect_term(indices,dummy,3);
		      }
		    }

		  }
		}
	      }
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(conf3);
  free(configuration);
  return sum;
}  

double term_exc_68(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

   /* TERM 68: 
     
     2 * P(m) . F(k) . F(k) . S(+)(k) . S(z)(m)
     
  */
  
  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int spin[MAX_LSUBM+1];
  int mm,nnn,temp1,temp2,temp3;
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int (*conf3)[5];
  int i, j, k, l, m;
  int nn[5];
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  int descriptorIndex3;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int conf3_len;
  int tempArray[MAX_LSUBM+1];
  int temp_number_conf3_sites;
//  int temp_number_conf2_sites;
//  int numberFlips1;
  int maxNumberFlips;
  int numberFlips2;
  int (*configuration)[5];
  double sum;
  sum=0.0000;

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int)); //f(i)
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int)); //f(i)
  conf3=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int)); //p(j)
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
      conf3[k][l]=0;
    }
  } 
 
  for(k=1;k<=length;k++) {
    
    for(m=1;m<5;m++) conf1[1][m]=configuration[k][m]; // f(i)
    for(m=1;m<5;m++) conf2[1][m]=configuration[k][m]; // f(i)
    
    // loop over nearest neighbours
    
    counter=1;
    for(l=1;l<=length;l++) {
      if (l != k) {
	tempArray[counter]=l;
	counter++;
      }
    }
    
    counter--;
    
    if (counter > 30) {
      printf("TERM 33: counter > 30...\nExiting!!\n\n");
      exit(1);
    }
    temp = counter; // split into summands not greater than 10 each
    temp3 = ((temp>10) ? 10 : temp); temp -= temp3;  // 0..10
    temp2 = ((temp>10) ? 10 : temp); temp -= temp2;  // 0..10
    temp1 = ((temp>10) ? 10 : temp); temp -= temp1;  // 0..10
    
    for(l=0;l<trimary_max[temp1];l++) {
      for(mm=0;mm<trimary_max[temp2];mm++) {
	for(nnn=0;nnn<trimary_max[temp3];nnn++) {
	  
	  trimary(l,mm,nnn,counter,spin);
	  conf1_len=1;
	  conf2_len=1;
	  conf3_len=0;
	  
	  for (j=1;j<=counter;j++) { 
	    temp=tempArray[j];
	    if (spin[j]==0) {         // f(i)
	      conf1_len++;
	      for (m=1;m<5;m++) 
		conf1[conf1_len][m]=configuration[temp][m];
	    } else if (spin[j]==1) {  // f(i)
	      conf2_len++;
	      for (m=1;m<5;m++) 
		conf2[conf2_len][m]=configuration[temp][m];
	    } else {                  // p(j)
	      conf3_len++;
	      for (m=1;m<5;m++) 
		conf3[conf3_len][m]=configuration[temp][m];
	    }
	  }	      
	  

          // moved that outside inner loop, Joerg Sep05, speedup
          descriptorIndex1=findClusterIndex(conf1_len, conf1);
          if (descriptorIndex1 <= 0) continue;
	  
          descriptorIndex2=findClusterIndex( conf2_len, conf2);
          if (descriptorIndex2 <= 0) continue;

	  index=conf1[1][4];      
	  for(j=1;j<=hh[map81[term]].k_to_m_number[index];j++) {
	    
	    // nearest neighbour is at x,y,z, temp put information
	    // into configuration and exit.
	    
	    /* k ---> m */
	    
	    nn[1]=conf1[1][1]+hh[map81[term]].k_to_m[index][j][0];
	    nn[2]=conf1[1][2]+hh[map81[term]].k_to_m[index][j][1];
	    nn[3]=conf1[1][3]+hh[map81[term]].k_to_m[index][j][2];
	    nn[4]=hh[map81[term]].k_to_m[index][j][3];

	    maxNumberFlips=UnitCell.s2[nn[4]];

	    numberFlips2=1;

	    for(m=1;m<=conf3_len;m++)
	      if (nn[1]==conf3[m][1] &&
		  nn[2]==conf3[m][2] &&
		  nn[3]==conf3[m][3] &&
		  nn[4]==conf3[m][4])
		numberFlips2++;
	        
	    // the number of spin flips that you can have at a given
	    // lattice site is depedent on the quantum spin number
	    // of the site in the UNIT CELL index `length'
	    
	    if (numberFlips2 <= maxNumberFlips) {
	      
	      temp_number_conf3_sites = conf3_len+1;
	      conf3[temp_number_conf3_sites][1]=nn[1];
	      conf3[temp_number_conf3_sites][2]=nn[2];
	      conf3[temp_number_conf3_sites][3]=nn[3];
	      conf3[temp_number_conf3_sites][4]=nn[4];
	      
	      descriptorIndex3=findExcitedClusterIndex(temp_number_conf3_sites, conf3);
	      if (descriptorIndex3 <= 0) continue;
		        
	      multiplier=-1*UnitCell.s2[nn[4]]; // - 2 * s from Hamiltonian.
	      
	      if (Control.file) {
		if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0) &&
		    (descriptorIndex3 > 0)){
		  int indices[7]={0};
		  double dummy;
		  indices[0]=config_index;
		  indices[1]=term; // Jun08: origin was 1..81, now using mapped term
		  indices[2]=descriptorIndex3; //p(j)
		  indices[3]=descriptorIndex2; //f(i)
		  indices[4]=descriptorIndex1; //f(i)
		  indices[5]=0;
		  dummy=multiplier*
		    hh[map81[term]].coefficient[index][j];
                  collect_term(indices,dummy,3);
		}
	      }
		
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(conf3);
  free(configuration);
  return sum;
}



double term_exc_69(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 69:
     
     2 * P(m) . F(m) . F(k) . F(k) . S(+)(k) . S(+)(m)

     this is same as term_exc_47 but k <=> m
   
    */

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (length < 2) return 0;
  if (Control.output > 10) printf("Term=%d\n",term);

 
  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int (*conf3)[5];
  int (*conf4)[5];
  int i, j, k, l, m, n, p, mm;
  int temp1, temp2, temp3;
  int k1;
  int k2;
  int are_they_neighbours;
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  int descriptorIndex3;
  int descriptorIndex4;
//  double multiplier;
  int conf1_num_sites;
  int conf2_num_sites;
  int conf3_num_sites;
  int conf4_num_sites;
  int tempArray[MAX_LSUBM+1];

  int (*configuration)[5];
  double sum;
  sum=0.0000;
  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf3=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf4=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<=MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
      conf3[k][l]=0;
      conf4[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) conf1[1][m]=configuration[k1][m]; // P(m)
	for(m=1;m<5;m++) conf2[1][m]=configuration[k1][m]; // F(m)
	for(m=1;m<5;m++) conf3[1][m]=configuration[k2][m]; // F(k)
	for(m=1;m<5;m++) conf4[1][m]=configuration[k2][m]; // F(k)
	
        index=conf1[1][4];
        are_they_neighbours= test_m_to_k_neighbour(hh,term,conf1[1],conf3[1]); 
	// conf1[1] and conf3[1] need to be nearest-neighbours 
	
	if (are_they_neighbours) {
	  
	  counter=1;
	  for(l=1;l<=length;l++) {
	    if (l != k1 && l != k2 ) {
	      tempArray[counter]=l;
	      counter++;
	    }
	  }
	  
	  counter--;
	  
    temp = counter; // split into summands not greater than 10 each
    temp3 = ((temp>10) ? 10 : temp); temp -= temp3;  // 0..10
    temp2 = ((temp>10) ? 10 : temp); temp -= temp2;  // 0..10
    temp1 = ((temp>10) ? 10 : temp); temp -= temp1;  // 0..10
	  
	  for(l=0;l<quartary_max[temp1];l++) {
	    for(m=0;m<quartary_max[temp2];m++) {
	      for(n=0;n<quartary_max[temp3];n++) {
		
		quartary(l,m,n,counter,spin);
		conf1_num_sites=1;
		conf2_num_sites=1;
		conf3_num_sites=1;
		conf4_num_sites=1;
		
		for (p=1;p<=counter;p++) { 
		  if (spin[p]==0) {
		    // F(k)
		    conf1_num_sites++;
		    temp=tempArray[p];
		    for (mm=1;mm<5;mm++) 
		      conf1[conf1_num_sites][mm]=
			configuration[temp][mm];
		  } else if (spin[p]==1) {
		    // F(k)
		    conf2_num_sites++;
		    temp=tempArray[p];
		      for (mm=1;mm<5;mm++) 
			conf2[conf2_num_sites][mm]=
			  configuration[temp][mm];
		  } else if (spin[p]==2) {
		    // F(m)
		    conf3_num_sites++;
		      temp=tempArray[p];
		      for (mm=1;mm<5;mm++) 
			conf3[conf3_num_sites][mm]=
			  configuration[temp][mm];
		  } else {
		    // F(m)
		    conf4_num_sites++;
		    temp=tempArray[p];
		      for (mm=1;mm<5;mm++) 
			conf4[conf4_num_sites][mm]=
			  configuration[temp][mm];
		  }
		}	      
		
		descriptorIndex1=findExcitedClusterIndex(conf1_num_sites, conf1);
                if (descriptorIndex1 <= 0) continue;
		
		descriptorIndex2=findClusterIndex(conf2_num_sites, conf2);
                if (descriptorIndex2 <= 0) continue;
		
		descriptorIndex3=findClusterIndex(conf3_num_sites, conf3);
                if (descriptorIndex3 <= 0) continue;
		
		descriptorIndex4=findClusterIndex(conf4_num_sites, conf4);
                if (descriptorIndex4 <= 0) continue;
		
		if (Control.file) {
                {
		    static int indices[7]={0};
		    double dummy;
		    indices[0]=config_index;  // original config
		    indices[1]=term; // Jun08: origin was 1..81, now using mapped term
		    indices[2]=descriptorIndex1; // modified config1..4
		    indices[3]=descriptorIndex2;
		    indices[4]=descriptorIndex3;
		    indices[5]=descriptorIndex4;
		    dummy= 
		     2*hh[map81[term]].coefficient[index][are_they_neighbours];
				collect_term(indices,dummy,3); // factor of 2 from H.
		}
		}
		
	    }
	    }
	    }

	}
	}
	}
	}


		

  free(conf1);
  free(conf2);
  free(conf3);
  free(conf4);
  free(configuration);
  return sum;
}


double term_exc_70(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 70: 
     
     4 * P(j) . F(i) . G(k)(m) . S(+)(k) . S(+)(m)
   
    */
  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);


  int spin[MAX_LSUBM+1];
  int mm,nnn,temp1,temp2,temp3;
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int (*conf3)[5];
  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  int descriptorIndex3;
  double multiplier;
  int conf1_len;
  int conf2_len;
  int conf3_len;
  int tempArray[MAX_LSUBM+1];
  double sum;
  sum=0.0000;

  int (*configuration)[5];

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf3=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
      conf3[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) conf1[1][m]=configuration[k1][m]; // P(j)
	for(m=1;m<5;m++) conf2[1][m]=configuration[k2][m]; // F(i)
	for(m=1;m<5;m++) conf3[1][m]=configuration[k1][m]; // H(i,j) <--- j
	
	// loop over nearest neighbours
	
	are_they_neighbours=0;
	index=conf1[1][4];      

	for(j=1;j<=hh[map81[term]].m_to_k_number[index];j++) {
	  
	  // nearest neighbour is at x,y,z, temp put information
	  // into configuration and exit.
	  
	  /* m ---> k */
	  
	  nn[1]=conf1[1][1]+hh[map81[term]].m_to_k[index][j][0];
	  nn[2]=conf1[1][2]+hh[map81[term]].m_to_k[index][j][1];
	  nn[3]=conf1[1][3]+hh[map81[term]].m_to_k[index][j][2];
	  nn[4]=hh[map81[term]].m_to_k[index][j][3];
	  
	  if(conf2[1][1] == nn[1] &&
	     conf2[1][2] == nn[2] &&
	     conf2[1][3] == nn[3] &&
	     conf2[1][4] == nn[4])
	    are_they_neighbours=j;

	  if (are_they_neighbours) break;
	}
	
	// conf1[1] and conf2[1] need to be nearest-neighbours 
	
	if (are_they_neighbours) {
	  
	  for(m=1;m<5;m++) conf3[2][m]=configuration[k2][m]; // G(i,j) <--- i
	  
	  counter=1;
	  for(l=1;l<=length;l++) {
	    if (l != k1 && l != k2 ) {
	      tempArray[counter]=l;
		counter++;
	    }
	  }
	  
	  counter--;
	  
	  if (counter > 30) {
	    printf("TERM 32: counter > 30...\nExiting!!\n\n");
	    exit(1);
	  }
          temp = counter; // split into summands not greater than 10 each
          temp3 = ((temp>10) ? 10 : temp); temp -= temp3;  // 0..10
          temp2 = ((temp>10) ? 10 : temp); temp -= temp2;  // 0..10
          temp1 = ((temp>10) ? 10 : temp); temp -= temp1;  // 0..10
	  
	  for(l=0;l<trimary_max[temp1];l++) {
	    for(mm=0;mm<trimary_max[temp2];mm++) {
	      for(nnn=0;nnn<trimary_max[temp3];nnn++) {
		
		trimary(l,mm,nnn,counter,spin);
		conf1_len=1;
		conf2_len=1;
		conf3_len=2;
		  
		for (j=1;j<=counter;j++) { 
		  temp=tempArray[j];
		  if (spin[j]==0) {          // p(k)
		    conf1_len++;
		    for (m=1;m<5;m++) 
		      conf1[conf1_len][m]=configuration[temp][m];
		  } else if (spin[j]==1) {   // F(m)
		    conf2_len++;
		    for (m=1;m<5;m++) 
		      conf2[conf2_len][m]=configuration[temp][m];
		  } else {                    // G(k,m)
		    conf3_len++;
		    for (m=1;m<5;m++) 
		      conf3[conf3_len][m]=configuration[temp][m];
		  }
		}	      
		
		descriptorIndex1=findExcitedClusterIndex(conf1_len, conf1);
                if (descriptorIndex1 <= 0) continue;

		descriptorIndex2=findClusterIndex(conf2_len, conf2);
                if (descriptorIndex2 <= 0) continue;

		descriptorIndex3=findClusterIndex(conf3_len, conf3);
                if (descriptorIndex3 <= 0) continue;

		multiplier=4.0; // factor from 4 from Hamiltonian.
				  
		if (Control.file) {
		  if ((descriptorIndex1 > 0)&&(descriptorIndex2 > 0)&&
		      (descriptorIndex3 > 0)){
		    int indices[7]={0};
		    double dummy;
		    indices[0]=config_index;
		    indices[1]=term; // Jun08: origin was 1..81, now using mapped term
		    indices[2]=descriptorIndex1; // p(j)
		    indices[3]=descriptorIndex2; // f(i)
		    indices[4]=descriptorIndex3; // g(i,j)
		    indices[5]=0;
		    dummy=multiplier*
		      hh[map81[term]].coefficient[index][are_they_neighbours];
                    collect_term(indices,dummy,3);
		  }
		}
		  
	      }
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(conf3);
  free(configuration);
  return sum;
}

double term_exc_71(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 71:
     
     2 * Q(j,j) . F(i) . S(+)(j) . S(z)(i)
     
  */

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i,j,k,l,m;
  int descriptorIndex1;
  int descriptorIndex2;
  double multiplier;
  int (*configuration)[5];
  int index;
  int nn[5];
  int neighbour;
  int maxNumberFlips;
  int maxNumberFlips1;
  int maxNumberFlips2;
  int numberFlips1;
  int numberFlips2;
  int (*conf1)[5], (*conf2)[5];
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
  int temp_number_conf2_sites;
  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));
  double sum;
  sum=0.0000;

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<=MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 

  for(k=1;k<=length;k++) { // `j' here. 
    
    index=configuration[k][4];
    maxNumberFlips=UnitCell.s2[index];
    
    if (maxNumberFlips > 1) {
      
      for(m=1;m<5;m++) {
	conf1[1][m]=configuration[k][m]; // Q(j,j) -- 1st k 
	conf1[2][m]=configuration[k][m]; // Q(j,j) -- 2nd k
      }
      
      counter=1;
      for(m=1;m<=length;m++) {
	if (m != k) {
	  tempArray[counter]=m;
	  counter++;
	}
      }
      
      counter--;
      
      for(l=0;l<binary_max[counter];l++) {
	
	binary(l,counter,spin);
	conf1_len=2;
	conf2_len=0;
	
	for (j=1;j<=counter;j++) { 
	  if (spin[j]) {
	    // Q(j,i)
	    conf1_len++;
	    temp=tempArray[j];
	    for (m=1;m<5;m++) 
	      conf1[conf1_len][m]=configuration[temp][m];
	  } else {
	    // F(i)
	    conf2_len++;
	    temp=tempArray[j];
	    for (m=1;m<5;m++) 
	      conf2[conf2_len][m]=configuration[temp][m];
	  }
	}
	
	index=conf1[1][4];
	
	for(neighbour=1;
	    neighbour<=hh[map81[term]].m_to_k_number[index];
	    neighbour++) {
	  
	  // nearest neighbour is at x,y,z, cell put information
	  // into configuration and exit.
	  
	  /* m ---> k */
	  
	  nn[1]=conf1[1][1]+
	    hh[map81[term]].m_to_k[index][neighbour][0];
	  nn[2]=conf1[1][2]+
	    hh[map81[term]].m_to_k[index][neighbour][1];
	  nn[3]=conf1[1][3]+
	    hh[map81[term]].m_to_k[index][neighbour][2];
	  nn[4]=hh[map81[term]].m_to_k[index][neighbour][3];
	  
	  numberFlips1=1;
	
	  for(m=1;m<=conf2_len;m++)
	    if (nn[1]==conf2[m][1] &&
		nn[2]==conf2[m][2] &&
		nn[3]==conf2[m][3] &&
		nn[4]==conf2[m][4])
	      numberFlips1++;
	
	  maxNumberFlips1=UnitCell.s2[nn[4]];
	
	  numberFlips2=0;
	  
	  for(m=1;m<=conf1_len;m++)
	    if (configuration[k][1]==conf1[m][1] &&
		configuration[k][2]==conf1[m][2] &&
		configuration[k][3]==conf1[m][3] &&
		configuration[k][4]==conf1[m][4])
	      numberFlips2++;
	
	  maxNumberFlips2=UnitCell.s2[configuration[k][4]];

	  // the number of spin flips that you can have at a given
	  // lattice site is depedent on the quantum spin number
	  // of the site in the UNIT CELL index `length'
	
	  if (numberFlips1 <= maxNumberFlips1 &&
	      numberFlips2 <= maxNumberFlips2) {
	    
	    temp_number_conf2_sites=conf2_len+1;
	  
	    conf2[temp_number_conf2_sites][1]=nn[1];
	    conf2[temp_number_conf2_sites][2]=nn[2];
	    conf2[temp_number_conf2_sites][3]=nn[3];
	    conf2[temp_number_conf2_sites][4]=nn[4];
	  
	    // Q(j,j)
	    descriptorIndex1=findExcitedClusterIndex(conf1_len, conf1);
            if (descriptorIndex1 <= 0) continue;
	    
	    // F(i)
	    descriptorIndex2=findClusterIndex(temp_number_conf2_sites, conf2);
	    if (descriptorIndex2 <= 0) continue;

	    multiplier=(-2)*0.5*UnitCell.s2[nn[4]]; // -2 * s from Hamiltonian
	  
	    if (Control.file) {
	      if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)) {
		int indices[7]={0};
		double dummy;
		indices[0]=config_index;
		indices[1]=term; // Jun08: origin was 1..81, now using mapped term
		indices[2]=descriptorIndex1; // Q(j,j)
		indices[3]=descriptorIndex2; // f(i)
		indices[4]=0;
		indices[5]=0;
		dummy=multiplier*
		hh[map81[term]].coefficient[index][neighbour];
            collect_term(indices,dummy,3);
	      }
	    }

	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}  

double term_exc_72(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 72: 
     
     Q(j)(j) . H(i)(i) . S(+)(i) . S(+)(j)
     
    */
  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);


  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int i, j, k, l, m;
  int k1;
  int k2;
  int nn[5];
  int are_they_neighbours;
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  // double multiplier;
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int maxNumberFlips1;
  int maxNumberFlips2;
  int numberFlips1;
  int numberFlips2;
  double sum;
  sum=0.0000;

  int (*configuration)[5];

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {

	maxNumberFlips1=UnitCell.s2[configuration[k1][4]];
	maxNumberFlips2=UnitCell.s2[configuration[k2][4]];
	
	if (maxNumberFlips1 > 1 && maxNumberFlips2 > 1) {
	  
	  for(m=1;m<5;m++) {
	    conf1[1][m]=configuration[k1][m]; // Q(j,j) j--1st one
	    conf1[2][m]=configuration[k1][m]; // Q(j,j) j--2nd one
	    conf2[1][m]=configuration[k2][m]; // H(i,i) i--1st one
	    conf2[2][m]=configuration[k2][m]; // H(i,i) i--2nd one
	  }
	  
	  // loop over nearest neighbours
	  
	  are_they_neighbours=0;
	  index=conf1[1][4];      
	  
	  for(j=1;j<=hh[map81[term]].m_to_k_number[index];j++) {
	    
	    // nearest neighbour is at x,y,z, temp put information
	    // into configuration and exit.
	    
	    /* m ---> k */
	    
	    nn[1]=conf1[1][1]+hh[map81[term]].m_to_k[index][j][0];
	    nn[2]=conf1[1][2]+hh[map81[term]].m_to_k[index][j][1];
	    nn[3]=conf1[1][3]+hh[map81[term]].m_to_k[index][j][2];
	    nn[4]=hh[map81[term]].m_to_k[index][j][3];
	    
	    if(conf2[1][1] == nn[1] &&
	       conf2[1][2] == nn[2] &&
	       conf2[1][3] == nn[3] &&
	       conf2[1][4] == nn[4])
	      are_they_neighbours=j;
      
	    if (are_they_neighbours) break;
	  }
	  
	  // conf1[1] and conf2[1] need to be nearest-neighbours 
	  
	  if (are_they_neighbours) {
	    
	    counter=1;
	    for(l=1;l<=length;l++) {
	      if (l != k1 && l != k2 ) {
		tempArray[counter]=l;
		counter++;
	      }
	    }
	    
	    counter--;
	    
	    for(l=0;l<binary_max[counter];l++) {
	      
	      binary(l,counter,spin);
	      conf1_len=2;
	      conf2_len=2;
	      
	      for (j=1;j<=counter;j++) { 
		if (spin[j]) {
		  // H
		  conf1_len++;
		  temp=tempArray[j];
		  for (m=1;m<5;m++) 
		    conf1[conf1_len][m]=configuration[temp][m];
		} else {
		  // H
		  conf2_len++;
		  temp=tempArray[j];
		  for (m=1;m<5;m++) 
		    conf2[conf2_len][m]=configuration[temp][m];
		}
	      }	      
	      
	      // make sure that we don't have too many `k' sites.
	      
	      numberFlips1=0;
	      
	      for(m=1;m<=conf1_len;m++) {
		if (configuration[k1][1]==conf1[m][1] &&
		    configuration[k1][2]==conf1[m][2] &&
		    configuration[k1][3]==conf1[m][3] &&
		    configuration[k1][4]==conf1[m][4])
		  numberFlips1++;
	      }
	      maxNumberFlips1=UnitCell.s2[configuration[k1][4]];

	      numberFlips2=0;
	      
	      for(m=1;m<=conf2_len;m++) {
		if (configuration[k2][1]==conf2[m][1] &&
		    configuration[k2][2]==conf2[m][2] &&
		    configuration[k2][3]==conf2[m][3] &&
		    configuration[k2][4]==conf2[m][4])
		  numberFlips2++;
	      }
	      maxNumberFlips2=UnitCell.s2[configuration[k2][4]];
	      
	      if (numberFlips1 <= maxNumberFlips1 &&
		  numberFlips2 <= maxNumberFlips2) {
		
		descriptorIndex1=findExcitedClusterIndex(conf1_len, conf1);
		if (descriptorIndex1 <= 0) continue;
		
		descriptorIndex2=findClusterIndex(conf2_len, conf2);
		if (descriptorIndex2 <= 0) continue;

		if (Control.file) {
		  if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)) {
		    int indices[7]={0};
		    double dummy;
		    indices[0]=config_index;
		    indices[1]=term; // Jun08: origin was 1..81, now using mapped term
		    indices[2]=descriptorIndex1; // q(j,j)
		    indices[3]=descriptorIndex2; // g(i,i)
		    indices[4]=0;
		    indices[5]=0;
		    dummy=
		      hh[map81[term]].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,3); // factor of 1 from Hamiltonian.
		  }
		}

	      }
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}

double term_exc_73(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 73:
     
     Q(j)(j) . F(i) . F(i) . S(+)(k) . S(+)(m)
     
  */

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i,j,k,l,m;
  int mm,nnn,temp1,temp2,temp3;
  int k1, k2;
  int descriptorIndex1;
  int descriptorIndex2;
  int descriptorIndex3;
  // double multiplier;
  int (*configuration)[5];
  int index;
  int nn[5];
  int maxNumberFlips;
  int numberFlips;
  int (*conf1)[5], (*conf2)[5], (*conf3)[5];
  int conf1_len;
  int conf2_len;
  int conf3_len;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
  int are_they_neighbours;

  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));
  double sum;
  sum=0.0000;

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf3=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<=MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
      conf3[k][l]=0;
    }
  } 
  
  for(k1=1;k1<=length;k1++) { 
    for(k2=1;k2<=length;k2++) { 
      
      if (k1 != k2) {
	
	index=configuration[k1][4];
	maxNumberFlips=UnitCell.s2[index];
	
	if (maxNumberFlips > 1) {
	  
	  for(m=1;m<5;m++) {
	    conf1[1][m]=configuration[k1][m]; // Q(j,j) <-- 1st j here
	    conf1[2][m]=configuration[k1][m]; // Q(j,j) <-- 2nd j here
	    conf2[1][m]=configuration[k2][m]; // F(i)
	    conf3[1][m]=configuration[k2][m]; // F(i)
	  }
	  
	  // loop over nearest neighbours
	  
	  are_they_neighbours=0;
	  index=conf1[1][4];      
	  
	  for(j=1;j<=hh[map81[term]].m_to_k_number[index];j++) {
	    
	    // nearest neighbour is at x,y,z, temp put information
	    // into configuration and exit.
	    
	    /* m ---> k */
	    
	    nn[1]=conf1[1][1]+hh[map81[term]].m_to_k[index][j][0];
	    nn[2]=conf1[1][2]+hh[map81[term]].m_to_k[index][j][1];
	    nn[3]=conf1[1][3]+hh[map81[term]].m_to_k[index][j][2];
	    nn[4]=hh[map81[term]].m_to_k[index][j][3];
	    
	    if(conf2[1][1] == nn[1] &&
	       conf2[1][2] == nn[2] &&
	       conf2[1][3] == nn[3] &&
	       conf2[1][4] == nn[4])
	      are_they_neighbours=j;      

	    if (are_they_neighbours) break;
	  }
	  
	  // conf1[1] and conf2[1] need to be nearest-neighbours 
	  
	  if (are_they_neighbours) {
	    
	    counter=1;
	    for(m=1;m<=length;m++) {
	      if (m != k1 && m != k2) {
		tempArray[counter]=m;
		counter++;
	      }
	    }
	    
	    counter--;
	    
	    if (counter > 30) {
	      printf("TERM 44: counter > 30...\nExiting!!\n\n");
	      exit(1);
	    }
            temp = counter; // split into summands not greater than 10 each
            temp3 = ((temp>10) ? 10 : temp); temp -= temp3;  // 0..10
            temp2 = ((temp>10) ? 10 : temp); temp -= temp2;  // 0..10
            temp1 = ((temp>10) ? 10 : temp); temp -= temp1;  // 0..10
	      
	    for(l=0;l<trimary_max[temp1];l++) {
	      for(mm=0;mm<trimary_max[temp2];mm++) {
		for(nnn=0;nnn<trimary_max[temp3];nnn++) {
		    
		  trimary(l,mm,nnn,counter,spin);
		    
		  conf1_len=2;
		  conf2_len=1;
		  conf3_len=1;
		    
		  for (j=1;j<=counter;j++) { 
		    if (spin[j]==0) {
		      // H(k,k)
		      conf1_len++;
		      temp=tempArray[j];
		      for (m=1;m<5;m++) 
			conf1[conf1_len][m]=configuration[temp][m];
		    } else if (spin[j]==1) {
		      // F(m)
		      conf2_len++;
		      temp=tempArray[j];
		      for (m=1;m<5;m++) 
			conf2[conf2_len][m]=configuration[temp][m];
		    } else {
		      // F(m)
		      conf3_len++;
		      temp=tempArray[j];
		      for (m=1;m<5;m++) 
			conf3[conf3_len][m]=configuration[temp][m];
		    }
		  }
	      
		  numberFlips=0;
	      
		  for(m=1;m<=conf1_len;m++)
		    if (configuration[k1][1]==conf1[m][1] &&
			configuration[k1][2]==conf1[m][2] &&
			configuration[k1][3]==conf1[m][3] &&
			configuration[k1][4]==conf1[m][4])
		      numberFlips++;
	      
		  maxNumberFlips=UnitCell.s2[configuration[k1][4]];
	      
		  // the number of spin flips that you can have at a given
		  // lattice site is depedent on the quantum spin number
		  // of the site in the UNIT CELL index `length'
	      
		  if (numberFlips <= maxNumberFlips) {
		
		    // Q(k,k)
		    descriptorIndex1=findExcitedClusterIndex(conf1_len, conf1);
		    if (descriptorIndex1 <= 0) continue;
		    
		    // F(m)
		    descriptorIndex2=findClusterIndex(conf2_len, conf2);
	            if (descriptorIndex2 <= 0) continue;
		    
		    // F(m)
		    descriptorIndex3=findClusterIndex(conf3_len, conf3);
		    if (descriptorIndex3 <= 0) continue;
	  
		    if (Control.file) {
		      if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0) &&
			  (descriptorIndex3 > 0)) {
			int indices[7]={0};
			double dummy;
			indices[0]=config_index;
			indices[1]=term; // Jun08: origin was 1..81, now using mapped term
			indices[2]=descriptorIndex1; // q(j,j)
			indices[3]=descriptorIndex2; // f(i)
			indices[4]=descriptorIndex3; // f(i)
			indices[5]=0;
			dummy=
			  hh[map81[term]].coefficient[index][are_they_neighbours];
            collect_term(indices,dummy,3); // factor of 1 from Hamiltonian.
		      }
		    }

		  }
		}
	      }
	    }
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(conf3);
  free(configuration);
  return sum;
}  

// phew. Now do s^z term 74 only!

double term_exc_74(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{
  
  /* 
     TERM 74: 
     
     P(k) . S(+)(k) 

     Wed Jul 10 BST 2002

  */
  
  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i,j,k;
  int (*configuration)[5];
  int index;
  double sum;
  sum=0.0000;
  
  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];
  
  // get the multiplying coefficient correct first!

  for(k=1;k<=length;k++) {
    
#if DEBUG
    int descriptorIndex;
    descriptorIndex=findExcitedClusterIndex(length,configuration);
    if (descriptorIndex != config_index) {
      printf("Term 74 Error: descriptorIndex (%d"
	") != config_index (%d)...\nExiting!!\n\n",
        descriptorIndex,config_index);
      exit(1);
    }
#endif

    index=configuration[k][4];      
        
    if (Control.file) {
      int indices[7]={0};
      double dummy;
      indices[0]=config_index;
      indices[1]=term; // Jun08: origin was 1..81, now using mapped term
      indices[2]=config_index; // p(i)
      indices[3]=0;
      indices[4]=0;
      indices[5]=0;
      dummy=hh[map81[term]].coefficient[index][1];
      collect_term(indices,dummy,3);
    }
  
  }
  
  free(configuration);
  return sum;
}

double term_exc_75(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 75: 
     
     -2 * P(k) . S(z)(k) 
  
  */
  
  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);
  // Oct05: configs with length+1 does not generate terms, so we skip it
  if (length+1>Params.n) return 0;

  int i, j, l;
  int descriptorIndex;
  // double multiplier;
  // double temp_multiplier;
  double sum;
  sum=0.0000;
  
  int minX, minY, minZ;  
  int maxX, maxY, maxZ;
  
  int x,y,z,cell;
  int index;
  
  int numberFlips, maxNumberFlips;
  
  int (*configuration)[5];
  
  configuration=(int (*)[5])malloc((MAX_LSUBM+4)*5*sizeof(int));
  
  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];
  
  int maxPointX;
  int maxPointY;
  int maxPointZ;
  
  maxPointX=0;
  maxPointY=0;
  maxPointZ=0;
  for(j=1;j<=length;j++) {
    if (maxPointX < abs(string[j][1])) maxPointX=abs(string[j][1]);
    if (maxPointY < abs(string[j][2])) maxPointY=abs(string[j][2]);
    if (maxPointZ < abs(string[j][3])) maxPointZ=abs(string[j][3]);
  }
  
  if (Params.dim == 1) {
    minX=radiusExc;           minY=0;                minZ=0;
    maxX=radiusExc+maxPointX; maxY=0;                maxZ=0;
  } else if (Params.dim == 2) {
    minX=radiusExc;           minY=radiusExc;           minZ=0;
    maxX=radiusExc+maxPointX; maxY=radiusExc+maxPointY; maxZ=0;
  } else if (Params.dim == 3) {
    minX=radiusExc;           minY=radiusExc;           minZ=radiusExc;
    maxX=radiusExc+maxPointX; maxY=radiusExc+maxPointY; maxZ=radiusExc+maxPointZ;
  } else {
    printf("Really weird error. Dimension (%d)"
     " not recognised...\nExiting||\n\n",Params.dim);
    exit(1);
  }

  
  for(x=-minX;x<=maxX;x++) {
    for(y=-minY;y<=maxY;y++) {
      for(z=-minZ;z<=maxZ;z++) {
	for(cell=1;cell<=UnitCell.number;cell++) {
	  
	  // order=length+1;   // this is the `k'    
	  
	  configuration[length+1][1]=x;
	  configuration[length+1][2]=y;
	  configuration[length+1][3]=z;
	  configuration[length+1][4]=cell;
	  
	  index=cell;
	  numberFlips=1;
	  
	  // check that we don't have too many spins at site `k'.
	  
	  for(l=1;l<=length;l++)
	    if (configuration[length+1][1]==configuration[l][1] &&
		configuration[length+1][2]==configuration[l][2] &&
		configuration[length+1][3]==configuration[l][3] &&
		configuration[length+1][4]==configuration[l][4])
	      numberFlips++;
	  
	  maxNumberFlips=UnitCell.s2[configuration[length+1][4]];
	  
	  // the number of spin flips that you can have at a given
	  // lattice site is depedent on the quantum spin number
	  // of the site in the UNIT CELL index `length'
	  
	  if (numberFlips <= maxNumberFlips) {
	    
	    // temp_multiplier=0.5*UnitCell.s2[configuration[length+1][4]]*
	    //  multiplier;
	    
	    descriptorIndex=findExcitedClusterIndex(length+1, configuration);
	        
	    if (Control.file) {
	      if (descriptorIndex > 0){
		int indices[7]={0};
		double dummy;
		indices[0]=config_index;
		indices[1]=term; // Jun08: origin was 1..81, now using mapped term
		indices[2]=descriptorIndex; // p(i)
		indices[3]=0;
		indices[4]=0;
		indices[5]=0;
		dummy=(-2.0)*(-0.5*UnitCell.s2[configuration[length+1][4]])*
		  hh[map81[term]].coefficient[index][1]; // factor from H is 2 s
		collect_term(indices,dummy,3);
	      }
	    }
	    
	  }
	}
      }
    }    
  }
  return sum;  
}

double term_exc_76(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 76: 
     
     - 2 P(k) . F(k) . S(+)(k) 

  */

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i,j,k,l,m;
  int descriptorIndex1;
  int descriptorIndex2;
  int (*configuration)[5];
  int index;
  int (*conf1)[5], (*conf2)[5];
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  int counter;
  int spin[MAX_LSUBM+2];
  int temp;
  double sum;
  sum=0.0000;

  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));
  
  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];
  
  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 

  for(k=1;k<=length;k++) {
        
    for(m=1;m<5;m++)
      conf1[1][m]=configuration[k][m]; // P(k)
    
    for(m=1;m<5;m++)
      conf2[1][m]=configuration[k][m]; // F(k)

    counter=1;
    for(m=1;m<=length;m++) {
      if (m != k) {
	tempArray[counter]=m;
	counter++;
      }
    }
    
    counter--;
    
    for(l=0;l<binary_max[counter];l++) {
      
      binary(l,counter,spin);
      conf1_len=1;
      conf2_len=1;
      
      for (j=1;j<=counter;j++) { 
	if (spin[j]) {
	  // F(m) 1st one
	  conf1_len++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf1[conf1_len][m]=configuration[temp][m];
	} else {
	  // F(m) 2nd one
	  conf2_len++;
	  temp=tempArray[j];
	  for (m=1;m<5;m++) 
	    conf2[conf2_len][m]=configuration[temp][m];
	}
      }
      
      index=conf1[1][4];	

      descriptorIndex1=findExcitedClusterIndex(conf1_len, conf1);
      if (descriptorIndex1 <= 0) continue;
      
      descriptorIndex2=findClusterIndex(conf2_len, conf2);
      if (descriptorIndex2 <= 0) continue;
				      
      if (Control.file) {
	if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)){
	  int indices[7]={0};
	  double dummy;
	  indices[0]=config_index;
	  indices[1]=term; // Jun08: origin was 1..81, now using mapped term
	  indices[2]=descriptorIndex1; // p(i)
	  indices[3]=descriptorIndex2; // f(i)
	  indices[4]=0;
	  indices[5]=0;
	  dummy=(-2.0)*hh[map81[term]].coefficient[index][1]; // factor of -2 from H.
	  collect_term(indices,dummy,3);
	}
      }
      
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}


double term_exc_77(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 77:
     
     - Q(k)(k) . S(+)(k) 
  
  */
  
  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i,j,k,l,m;
  int descriptorIndex;
  // double multiplier;
  int (*configuration)[5];
  int (*conf)[5];
  int index;
  int maxNumberFlips;
  int maxNumberFlips2;
  int numberFlips2;
  int temp_length;
  double sum;
  sum=0.0000;
  
  configuration=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));
  conf=(int (*)[5])malloc((MAX_LSUBM+2)*5*sizeof(int));
  
  for(i=1;i<=length;i++)
    for(j=1;j<5;j++) 
      configuration[i][j]=string[i][j];
  
  for(i=1;i<=MAX_LSUBM;i++)
    for(j=1;j<5;j++) 
      conf[i][j]=0;
  
  // multiplier=1.0;
  
  for(k=1;k<=length;k++) {
    
    index=configuration[k][4];
    maxNumberFlips=UnitCell.s2[index];
    
    if (maxNumberFlips > 1) {
      
      for(j=1;j<=length;j++)
	for(m=1;m<5;m++)
	  conf[j][m]=configuration[j][m]; // Q(k,k) <- 1st k
      
      temp_length=length+1;
      
      // we repeat the `k' 
      
      for(m=1;m<5;m++)
	conf[temp_length][m]=configuration[k][m]; // Q(k,k) <- 2nd k
      
      index=configuration[k][4];
      
      numberFlips2=0;
      
      for(l=1;l<=temp_length;l++)
	if (configuration[k][1]==conf[l][1] &&
	    configuration[k][2]==conf[l][2] &&
	    configuration[k][3]==conf[l][3] &&
	    configuration[k][4]==conf[l][4])
	  numberFlips2++;
      
      maxNumberFlips2=UnitCell.s2[index];
      
      if (numberFlips2 <= maxNumberFlips2)  {
	
	descriptorIndex=findExcitedClusterIndex(temp_length,conf);
	
//	if (descriptorIndex > 0)
//	  sum+=(-1.0)*CCMcoeffs[descriptorIndex]*
//	    hh[map81[term]].coefficient[index][1];	
	
	if (Control.file) {
	  if (descriptorIndex > 0) {
	    int indices[7]={0};
	    double dummy;
	    indices[0]=config_index;
	    indices[1]=term; // Jun08: origin was 1..81, now using mapped term
	    indices[2]=descriptorIndex; // q(i,i)
	    indices[3]=0;
	    indices[4]=0;
	    indices[5]=0;
	    dummy=(-1.0)*hh[map81[term]].coefficient[index][1]; // factor of -1
            collect_term(indices,dummy,3);
	  }
	}
	
      }
    }
  }
  free(configuration);
  free(conf);
  return sum;
}


// now (s^z)^2 terms 78 to 81

// (s^z) ^2 now

double term_exc_78(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{
  
  /* 
     TERM 78: 
     
     2 P(k) . S(+)(k) . S(z)(k)

  */
  
  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i,j,k;
  int descriptorIndex;
  double multiplier;
  int (*configuration)[5];
  int index;
  double sum;
  sum=0.0000;
  
  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  
  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];
  
  // get the multiplying coefficient correct first!

  for(k=1;k<=length;k++) {
    
    descriptorIndex=findClusterIndex(length,configuration);
    
    if (descriptorIndex != config_index) {
      printf("Term 58 Error: descriptorIndex (%d"
	") != config_index (%d)...\nExiting!!\n\n",
        descriptorIndex, config_index);
      exit(1);
    }
    
    index=configuration[k][4];      
    
    // note that no *CCMcoeffs[config_index] to find in sum
    // because it's the lhs.
    
//    sum+=multiplier*(-0.5*UnitCell.s2[index])*
//      hh[map81[term]].coefficient[index][1];
    
    // -0.5*UnitCell.s2[new_index] -- coefficient from \tilde s^z \tilde s^z
    
    if (Control.file) {
      int indices[7]={0};
      double dummy;
      indices[0]=config_index;
      indices[1]=term; // Jun08: origin was 1..81, now using mapped term
      indices[2]=config_index; // p(i)
      indices[3]=0;
      indices[4]=0;
      indices[5]=0;
      dummy=-1*UnitCell.s2[index]* 
	hh[map81[term]].coefficient[index][1];
            collect_term(indices,dummy,3); // factor -2s from H.
    }
    
  }
  free(configuration);
  return sum;
}

double term_exc_79(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 79: 
     
     P(k) . S(+)(k) . S(+)(k).
     
  */


  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int i, j, l;
  int k1;
  int k2;
  int are_they_the_same_lattice_site;
  int index;
  int descriptorIndex;
  double multiplier;
  int (*configuration)[5];
  int (*conf)[5];
  int tempConfig[MAX_LSUBM+2];
  int counter;
  double sum;
  sum=0.0000;

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++) {
      configuration[i][j]=string[i][j];
      conf[i][j]=0;
    }
  
  multiplier=1.0;
  
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      
      if (k1 != k2) {
	
	for(l=1;l<5;l++) { conf[1][l]=configuration[k1][l]; } // P(m)
	for(l=1;l<5;l++) { tempConfig[l]=configuration[k2][l]; } // s(k)^+
	
	are_they_the_same_lattice_site=0;
	index=conf[1][4];
	
	if(tempConfig[1] == conf[1][1] &&
	   tempConfig[2] == conf[1][2] &&
	   tempConfig[3] == conf[1][3] &&
	   tempConfig[4] == conf[1][4])
	  are_they_the_same_lattice_site=1;      
	  
	// now load the rest into conf
	
	counter=1;
	for(i=1;i<=length;i++) {
	  if ((i != k2) && (i != k1)) {
	    counter++;
	    for(l=1;l<5;l++)
	      conf[counter][l]=configuration[i][l];
	  }
	}
	
	if (counter != (length-1)) {
	  printf("TERM 59: Error counter != (length-1)...\nExiting\n\n");
	  exit(1);
	}
	
	if (are_they_the_same_lattice_site) {
	  
	  descriptorIndex=findExcitedClusterIndex(length-1,conf);
	  
	  if (Control.file) {
	    if (descriptorIndex > 0){
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=term; // Jun08: origin was 1..81, now using mapped term
	      indices[2]=descriptorIndex; // p(i)
	      indices[3]=0;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=hh[map81[term]].coefficient[index][1]; // factor of 1 from H.
            collect_term(indices,dummy,3);
	    }
	  }
	  
	}
      }
    }
  }
  free(configuration);
  free(conf);
  return sum;
}

double term_exc_80(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 80: 
     
     Q{k,k} . S(+)(k) . S(+)(k).
     
  */

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);


  int i, j, l;
  int k1;
  int k2;
  int are_they_the_same_lattice_site;
  int index;
  int descriptorIndex;
  double multiplier;
  int (*configuration)[5];
  int (*conf)[5];
  int counter;
  double sum;
  sum=0.0000;

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++) {
      configuration[i][j]=string[i][j];
      conf[i][j]=0;
    }
  
  multiplier=1.0;
  
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      
      if (k1 != k2) {
	
	for(l=1;l<5;l++) { conf[1][l]=configuration[k1][l]; } // F(m)
	for(l=1;l<5;l++) { conf[2][l]=configuration[k2][l]; } // s(k)^+
	
	are_they_the_same_lattice_site=0;
	index=conf[1][4];
	
	if(conf[2][1] == conf[1][1] &&
	   conf[2][2] == conf[1][2] &&
	   conf[2][3] == conf[1][3] &&
	   conf[2][4] == conf[1][4])
	  are_they_the_same_lattice_site=1;      
	  
	// now load the rest into conf
	
	counter=2;
	for(i=1;i<=length;i++) {
	  if ((i != k2) && (i != k1)) {
	    counter++;
	    for(l=1;l<5;l++)
	      conf[counter][l]=configuration[i][l];
	  }
	}
	
	if (counter != length) {
	  printf("TERM 60: Error counter != length...\nExiting\n\n");
	  exit(1);
	}
	
	if (are_they_the_same_lattice_site) {
	  
	  descriptorIndex=findExcitedClusterIndex(length,conf);
	    
	  if (Control.file) {
	    if (descriptorIndex > 0){
	      int indices[7]={0};
	      double dummy;
	      indices[0]=config_index;
	      indices[1]=term; // Jun08: origin was 1..81, now using mapped term
	      indices[2]=descriptorIndex; // q(i,i)
	      indices[3]=0;
	      indices[4]=0;
	      indices[5]=0;
	      dummy=hh[map81[term]].coefficient[index][1]; // factor of +1 from H. 
            collect_term(indices,dummy,3);
	    }
	  }
	  
	}
      }
    }
  }
  free(configuration);
  free(conf);
  return sum;
}

double term_exc_81(int config_index, int length, int (*string)[5], struct hamiltonian_defn hh[], int term)
{

  /* TERM 81: 

     2 . P(k) . F(k) . S(+)(k) . S(+)(k). 
     
  */

  if (!hh[map81[term]].evaluate) return 0; /* skip that term */
  if (Control.output > 10) printf("Term=%d\n",term);

  int spin[MAX_LSUBM+1];
  int counter;
  int temp;
  int (*conf1)[5];
  int (*conf2)[5];
  int i, j, k, l, m;
  int k1;
  int k2;
  int are_they_the_same_lattice_site;
  int index;
  int descriptorIndex1;
  int descriptorIndex2;
  int conf1_len;
  int conf2_len;
  int tempArray[MAX_LSUBM+1];
  double sum;
  int (*configuration)[5];

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  sum=0.0000;

  for(i=1;i<=length;i++)
    for(j=1;j<5;j++)
      configuration[i][j]=string[i][j];

  conf1=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  conf2=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
 
  for(k=1;k<MAX_LSUBM;k++) {
    for(l=1;l<5;l++) {
      conf1[k][l]=0;
      conf2[k][l]=0;
    }
  } 
 
  for(k1=1;k1<=length;k1++) {
    for(k2=1;k2<=length;k2++) {
      if (k2 != k1) {
	
	for(m=1;m<5;m++) 
	  conf1[1][m]=configuration[k1][m]; 
	
	for(m=1;m<5;m++) 
	  conf2[1][m]=configuration[k2][m];
	
	// loop over nearest neighbours
	
	are_they_the_same_lattice_site=0;
	index=conf1[1][4];

	if(conf2[1][1] == conf1[1][1] &&
	   conf2[1][2] == conf1[1][2] &&
	   conf2[1][3] == conf1[1][3] &&
	   conf2[1][4] == conf1[1][4])
	  are_they_the_same_lattice_site=1;      
	
	// conf1[1] and conf2[1] need to be nearest-neighbours 
	
	if (are_they_the_same_lattice_site) {
	  
	  counter=1;
	  for(l=1;l<=length;l++) {
	    if (l != k1 && l != k2 ) {
	      tempArray[counter]=l;
	      counter++;
	    }
	  }	  
	  counter--;
	  
	  for(l=0;l<binary_max[counter];l++) {
	    
	    binary(l,counter,spin);
	    conf1_len=1;
	    conf2_len=1;
	    
	    for (j=1;j<=counter;j++) {
	      if (spin[j]) {
		conf1_len++;
		temp=tempArray[j];
		for (m=1;m<5;m++)
		  conf1[conf1_len][m]=configuration[temp][m];
	      } else {
		conf2_len++;
		temp=tempArray[j];
		for (m=1;m<5;m++) 
		  conf2[conf2_len][m]=configuration[temp][m];
	      }
	    }	      
	    
	    descriptorIndex1=findExcitedClusterIndex(conf1_len, conf1);
	    if (descriptorIndex1 <= 0) continue;
	    
	    descriptorIndex2=findClusterIndex(conf2_len, conf2);
	    if (descriptorIndex2 <= 0) continue;
	    
	    // now find the multiplying cofficient.
	        
//	    if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0))
//	      sum+=CCMcoeffs[descriptorIndex1]
//		  *CCMcoeffs[descriptorIndex2]*
//		hh[map81[term]].coefficient[index][1];
	    
	    if (Control.file) {
	      if ((descriptorIndex1 > 0) && (descriptorIndex2 > 0)) {
		int indices[7]={0};
		double dummy;
		indices[0]=config_index;
		indices[1]=term; // Jun08: origin was 1..81, now using mapped term
		indices[2]=descriptorIndex1; // p(i)
		indices[3]=descriptorIndex2; // f(i)
		indices[4]=0;
		indices[5]=0;
		dummy=2*hh[map81[term]].coefficient[index][1]; // factor of 2 from H.
		collect_term(indices,dummy,3);
	      }
	    }
	    
	  }
	}
      }
    }
  }
  free(conf1);
  free(conf2);
  free(configuration);
  return sum;
}


void obtainExcitations(void)
{

  // A new routine to find the CCM excited-state equations.
  // DJJF June 2008.
  // here we have a different meaning of indices[]:
  //   x[2]    <= excitation coefficients
  //   x[3..5] <= g-s ket-state coefficients (NO PARTICULAR ORDER)
// The key to understanding why there is only one excitation coefficient is in
// the e^-S [H,X] e^S part of the equations -> X is linear in the equations,
// whereas e^S means that there are 1 + S + 1/2! S^2 + ... terms in the
// equations and so there can be higher order terms for the g-s coefficients.
// (DJJF Jun08)


  int    i,j,bonds;
  int    length;
  int    (*configuration)[5];
  int    dummy;
  int flag_subset;
//  double energy;
  double lhs, rhs;
//  char output[60];
  time_t t1, t2;
  struct hamiltonian_defn *hh;  /* [MAX_TERMS13] */
 
//   Joerg Jun08: much better done by collect routine setting id=3 
//  char name[256];
//  strcpy(name,Control.name); 
//  strcat(name,".exc_terms");
//  fp_exc=fopen(name,"w");
//  if (!fp_exc) {
//    printf("Cannot Open %s...\nExiting!!\n\n", name);
//    exit(1);
//  }


#define MAX_out 10

  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  flag_subset=1;  // do subset-Check before findClusterIndex Apr08 JS

  if (!myid) printf("\n");
  
  binary_max[0]=1;     dummy=2;   // 2^0 .. 2^30  # 30bit
  for(i=1;i<=30;i++) { binary_max[i]=dummy;   dummy=2*dummy; }

  trimary_max[0]=1;    dummy=3;   // 3^0 .. 3^10  # 16bit, i<=MAX_LSUBM
  for(i=1;i<=10;i++) { trimary_max[i]=dummy;  dummy=3*dummy; }

  quartary_max[0]=1;   dummy=4;   // 4^0 .. 4^10  # 20bit
  for(i=1;i<=10;i++) { quartary_max[i]=dummy; dummy=4*dummy; }

  radiusExc=0;
  {
    int length;
    int (*tempConfig)[5]=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
    for(i=1;i<=NumberExcitedConfigurations;i++) {
      length=findClusterFromDescriptor(excitedSpinConfigurations[i].descriptor, tempConfig);
      for(j=1;j<=length;j++) {
        if (abs(tempConfig[j][1]) > radiusExc) radiusExc=abs(tempConfig[j][1]);
        if (abs(tempConfig[j][2]) > radiusExc) radiusExc=abs(tempConfig[j][2]);
        if (abs(tempConfig[j][3]) > radiusExc) radiusExc=abs(tempConfig[j][3]);
      }
    }
    radiusExc+=2; // OK? DJJF Dec 08
    if (!myid)
      printf("\nmaximum cell radiusExc of all clusters = %d\n\n",radiusExc);
    free(tempConfig);
  }

  // find CCM equations.

  if (!myid){
    printf("==========================================================\n\n");
    printf("OBTAIN CCM EXCITED-STATE EQUATIONS: START ALGORITHM.\n\n");
    printf("==========================================================\n\n");
  }

  t1=time(NULL);
  
  for(bonds=1;bonds<=Params.bonds;bonds++) {

    // n_out=0;  // reset for each bond    

    for(i=1;i<=NumberExcitedConfigurations;i++) { // total number of clusters

      if ((i%npe)==myid) {  // split loop to MPI threads
        
        static unsigned long nt_print=1; // used to output all 4^n

	length = findClusterFromDescriptor(
             excitedSpinConfigurations[i].descriptor, configuration); 

        if (Control.output>2 && myid==0 && i<100) {
    printf("TARGET %6d = ", i);								   
	  for(int l=1;l<=length;l++) 
	     printf("(%d,%d,%d:%d) ",
	     configuration[l][1],configuration[l][2],
	     configuration[l][3],configuration[l][4]);
	  printf("\n");
        }
	
        // we want to output not to much because not every user knows
        //    what he does ;)
        // ToDo: output only for numterms=1,4,16,64,256,1024,4096,...
	//   printf("obtain: Bond %d : Cluster<= %5d\n",bonds,i);
 if (Control.output > 0 && myid == 0
         && nterms_out1 > nt_print) {  // we only output every 4^n
          t2=time(NULL)-t1;
          if (nt_print<<1) nt_print<<=1;  
          if (t2>27)   
          printf("obtain: Bond %d : Cluster<= %5d"
                 " : terms+= %8ld %5ld/s %5.1fm\n",
                 bonds, i, nterms_out1, ((t2>0)?nterms_out1/t2:0), t2/60. );
          fflush(stdout); 
          if (i==MAX_out)
            printf("obtain: output is reduced after %d terms\n",MAX_out);
        }

        hh=Hamiltonian[bonds];  // abbreviation

	// s^z s^z : s all s = t1
	
        // last argument is the porojection from terms 1..81 to 1..63
        //    could be replaced by mapping to 1..13 and "debugstring f(z+-ij)"
	lhs += term_exc_1(i,length,configuration,hh,1);  // map62[2]=1
        lhs += term_exc_2(i,length,configuration,hh,2);  // use 2 for term in all these.
        lhs += term_exc_3(i,length,configuration,hh,3);
        lhs += term_exc_4(i,length,configuration,hh,4);
        lhs += term_exc_5(i,length,configuration,hh,5);
	
	// s^z s^+ all s = t2
	
	rhs += term_exc_6(i,length,configuration,hh,6);  // map62[6]=2
      
	// s^z s^- all s = t3
	
        rhs += term_exc_7(i,length,configuration,hh,7); // map62[8]=3
        rhs += term_exc_8(i,length,configuration,hh,8); // use 8 as this is gives correct
        rhs += term_exc_9(i,length,configuration,hh,9); // Hamiltonian coefficient in ccm eqns
        rhs += term_exc_10(i,length,configuration,hh,10); 
        rhs += term_exc_11(i,length,configuration,hh,11); 
        rhs += term_exc_12(i,length,configuration,hh,12); 
        rhs += term_exc_13(i,length,configuration,hh,13);  
        rhs += term_exc_14(i,length,configuration,hh,14);  
        rhs += term_exc_15(i,length,configuration,hh,15);  
        rhs += term_exc_16(i,length,configuration,hh,16);  
	rhs += term_exc_17(i,length,configuration,hh,17);  
        rhs += term_exc_18(i,length,configuration,hh,18);  
        rhs += term_exc_19(i,length,configuration,hh,19);  

	// s^+ s^z all s = t4
	
	rhs += term_exc_20(i,length,configuration,hh,20); // map62[14]=4

	// s^- s^z all s
      
	rhs += term_exc_21(i,length,configuration,hh,21); // map62[16]=5
	rhs += term_exc_22(i,length,configuration,hh,22);
	rhs += term_exc_23(i,length,configuration,hh,23);
	rhs += term_exc_24(i,length,configuration,hh,24);
	rhs += term_exc_25(i,length,configuration,hh,25);
	rhs += term_exc_26(i,length,configuration,hh,26);
	rhs += term_exc_27(i,length,configuration,hh,27);
	rhs += term_exc_28(i,length,configuration,hh,28);
	rhs += term_exc_29(i,length,configuration,hh,29);
	rhs += term_exc_30(i,length,configuration,hh,30);
	rhs += term_exc_31(i,length,configuration,hh,31);
	rhs += term_exc_32(i,length,configuration,hh,32);
	rhs += term_exc_33(i,length,configuration,hh,33);

      // s^+ s^- all s

  rhs += term_exc_34(i,length,configuration,hh,34); // map62[22]=6
  rhs += term_exc_35(i,length,configuration,hh,35);
  rhs += term_exc_36(i,length,configuration,hh,36);

      // s^- s^+ all s

	rhs += term_exc_37(i,length,configuration,hh,37); // map62[24]=7
	rhs += term_exc_38(i,length,configuration,hh,38);
	rhs += term_exc_39(i,length,configuration,hh,39);
      
      // s^+ s^+ all s = t8 <- no terms for excitations as [s^+ s^+,X]=0
      
      // s^- s^- all s

        rhs += term_exc_40(i,length,configuration,hh,40); // map62[27]=9
        rhs += term_exc_41(i,length,configuration,hh,41); 
        rhs += term_exc_42(i,length,configuration,hh,42); 
        rhs += term_exc_43(i,length,configuration,hh,43); 
        rhs += term_exc_44(i,length,configuration,hh,44); 
        rhs += term_exc_45(i,length,configuration,hh,45); 
        rhs += term_exc_46(i,length,configuration,hh,46); 
        rhs += term_exc_47(i,length,configuration,hh,47); // call quartary slow!
        rhs += term_exc_48(i,length,configuration,hh,48); 
        rhs += term_exc_49(i,length,configuration,hh,49); 
        rhs += term_exc_50(i,length,configuration,hh,50); 
        rhs += term_exc_51(i,length,configuration,hh,51); 
        rhs += term_exc_52(i,length,configuration,hh,52); 
        rhs += term_exc_53(i,length,configuration,hh,53); 
        rhs += term_exc_54(i,length,configuration,hh,54); 
        rhs += term_exc_55(i,length,configuration,hh,55); 
        rhs += term_exc_56(i,length,configuration,hh,56); 
        rhs += term_exc_57(i,length,configuration,hh,57); 
        rhs += term_exc_58(i,length,configuration,hh,58); 
        rhs += term_exc_59(i,length,configuration,hh,59); 
        rhs += term_exc_60(i,length,configuration,hh,60); 
        rhs += term_exc_61(i,length,configuration,hh,61); 
        rhs += term_exc_62(i,length,configuration,hh,62); 
        rhs += term_exc_63(i,length,configuration,hh,63); 
        rhs += term_exc_64(i,length,configuration,hh,64); 
        rhs += term_exc_65(i,length,configuration,hh,65); 
        rhs += term_exc_66(i,length,configuration,hh,66); 
        rhs += term_exc_67(i,length,configuration,hh,67); 
        rhs += term_exc_68(i,length,configuration,hh,68); 
        rhs += term_exc_69(i,length,configuration,hh,69); // call quartary slow!
        rhs += term_exc_70(i,length,configuration,hh,70); 
        rhs += term_exc_71(i,length,configuration,hh,71); 
        rhs += term_exc_72(i,length,configuration,hh,72); 
        rhs += term_exc_73(i,length,configuration,hh,73); 

	// s^z all s

        rhs += term_exc_74(i,length,configuration,hh,74); // map62[54]=10

	// s^- all s

        rhs += term_exc_75(i,length,configuration,hh,75); // map62[55]=11
	rhs += term_exc_76(i,length,configuration,hh,76); 
	rhs += term_exc_77(i,length,configuration,hh,77); 

	// (s^z)^2 all s

        rhs += term_exc_78(i,length,configuration,hh,78); // map62[58]=12
	rhs += term_exc_79(i,length,configuration,hh,79); 
	rhs += term_exc_80(i,length,configuration,hh,80); 
	rhs += term_exc_81(i,length,configuration,hh,81); 
	
     }
    }
    if (Control.file) {
      int indices[7]={0};
      indices[0]=STOP_CFG;indices[1]=STOP_TERM;
      indices[2]=STOP_CFG;indices[3]=STOP_CFG;
      indices[4]=STOP_CFG;indices[5]=STOP_CFG;
      collect_term(indices,0.0,3); // excitations
    }
    if (Control.output>0 && myid==0) {
      printf("END bonds %d ==================================\n\n",bonds);
    }
  }
  
  if (myid==0) {
    t2=time(NULL)-t1;
    printf(" obtainExcitatitons.t= %.2f\n", t2/60.);
    printf("\nOBTAIN CCM EXCITED-STATE EQUATIONS: SUCCESSFULLY COMPLETED!\n\n");
  }

//  fclose(fp_exc); 

  free(configuration); // memleak until v0606

}
